/************************************************************************
 * fte_fifo.c
 *
 * Author: Klemen Vodopivec (Cosylab)
 * Copyright (C) 2010-2015 ITER Organization
 *
 * This file is subject to the term and conditions of the GNU General Public
 * License. See the file COPYING in the main directory of the Linux
 * distribution for more details.
 ************************************************************************/

#include <linux/list.h>
#include <linux/slab.h>
#include <linux/time.h>
#include <linux/delay.h>
#include "fte_fifo.h"

struct fte {
	u64 time;
	u64 term;
	u64 level;
	u64 linestate;

	struct list_head list_node;
};

struct fte_fifo {
	size_t capacity;
	size_t size;

	u32 min_time_delta;
	u64 current_linestate;

	struct list_head queued_ftes;
	struct list_head pool;
	struct fte fte_buffer[0];
};

#define fte_fifo_merge_event(_itr_, _bitmask_, _level_) ({		\
	struct fte *curr_fte = list_entry(_itr_, struct fte, list_node);\
	curr_fte->term |= _bitmask_;					\
	curr_fte->level &= ~_bitmask_;					\
	curr_fte->level |= _level_;					\
	curr_fte->linestate &= ~_bitmask_;				\
	curr_fte->linestate |= _level_;					\
	_itr_; })

#define fte_fifo_assign(_fte_, _time_, _bitmask_, _level_, _prev_linestate_) ({\
	_fte_->time = _time_;						\
	_fte_->term = _bitmask_;					\
	_fte_->level = _level_;						\
	_fte_->linestate = _prev_linestate_;				\
	_fte_->linestate &= ~_bitmask_;					\
	_fte_->linestate |= _level_;					\
	&_fte_->list_node; })

#define fte_fifo_min_time_delta_err(_prev_time_, _time_, _next_time_) ({\
	bool err = false;						\
	if (_prev_time_ && _prev_time_ + fifo->min_time_delta > _time_) {\
		pr_debug("FTE %llu too close to previous %llu, need %u\n",\
			_time_, _prev_time_, fifo->min_time_delta);	\
		err = true;						\
	} else if (_next_time_ && _time_ + fifo->min_time_delta >	\
		   _next_time_) {					\
		pr_debug("FTE %llu too close to subsequent %llu, need %u\n",\
				_time_, _next_time_, fifo->min_time_delta); \
		err = true;						\
	}								\
	err; })

#define fte_fifo_insert_before(_itr_, _time_, _bitmask_, _level_,	\
				_prev_linestate_) ({			\
	int _err = 0;							\
	if (list_empty(&fifo->pool)) {					\
		pr_debug("=> FIFO full\n");				\
		_err = -ENOMEM;						\
	} else {							\
		struct fte *new_fte = list_first_entry(&fifo->pool, struct fte,\
							list_node);	\
		list_move_tail(&new_fte->list_node, _itr_);		\
		fte_fifo_assign(new_fte, _time_, _bitmask_, _level_,	\
				_prev_linestate_);			\
	}								\
	_err; })

static void _fte_fifo_dbg_print(struct fte_fifo *fifo)
{
	struct fte *curr_fte;
	int count;

	pr_debug("%s()\n", __func__);
	pr_debug("fte_fifo: %p\n", fifo);
	pr_debug("          capacity: %zu\n", fifo->capacity);
	pr_debug("              size: %zu\n", fifo->size);
	pr_debug("    min_time_delta: %u\n", fifo->min_time_delta);
	pr_debug(" current_linestate: 0x%016llx\n", fifo->current_linestate);
	pr_debug("       queued_ftes:\n");
	count = 0;
	list_for_each_entry(curr_fte, &fifo->queued_ftes, list_node) {
		pr_debug("  [%04d]: %p, time: %10llus %9lluns, term: 0x%016llx level: 0x%016llx line state: 0x%016llx\n",
			count, &curr_fte->list_node,
			curr_fte->time / NSEC_PER_SEC,
			curr_fte->time % NSEC_PER_SEC,
			curr_fte->term, curr_fte->level, curr_fte->linestate);
		count++;
	}
}

static void _fte_fifo_reset(struct fte_fifo *fifo)
{
	int i;

	pr_debug("%s()\n", __func__);
	fifo->size = 0;

	INIT_LIST_HEAD(&fifo->queued_ftes);
	INIT_LIST_HEAD(&fifo->pool);

	for (i = 0; i < fifo->capacity; i++)
		list_add_tail(&fifo->fte_buffer[i].list_node, &fifo->pool);
}

struct fte_fifo *fte_fifo_alloc(const size_t capacity)
{
	struct fte_fifo *fifo;

	pr_debug("%s(capacity: %zu)\n", __func__, capacity);

	fifo = kzalloc(sizeof(struct fte_fifo) + sizeof(struct fte) * capacity,
		       GFP_KERNEL);
	if (!fifo)
		return NULL;

	fifo->capacity = capacity;

	_fte_fifo_reset(fifo);
	fifo->current_linestate = 0;

	_fte_fifo_dbg_print(fifo);
	pr_debug("=> fifo = %p\n", fifo);

	return fifo;
}

void fte_fifo_destroy(struct fte_fifo *fifo)
{
	pr_debug("%s(fifo: %p)\n", __func__, fifo);
	kfree(fifo);
}

size_t fte_fifo_capacity(struct fte_fifo *fifo)
{
	pr_debug("%s(fifo: %p, fifo->capacity: %zu)\n", __func__,
				fifo, fifo->capacity);
	return fifo->capacity;
}

size_t fte_fifo_size(struct fte_fifo *fifo)
{
	pr_debug("%s(fifo: %p, fifo->size: %zu)\n", __func__, fifo, fifo->size);
	return fifo->size;
}

void fte_fifo_set_min_time_delta(struct fte_fifo *fifo,
				const u32 min_time_delta)
{
	pr_debug("%s(fifo: %p, min_time_delta: %u)\n", __func__,
				fifo, min_time_delta);
	fifo->min_time_delta = min_time_delta;
}

u64 fte_fifo_current_safety_margin(struct fte_fifo *fifo)
{
	return fte_fifo_size(fifo) * FTE_FIFO_EVENT_PROGRAMMING_TIME
			+ FTE_FIFO_REPROGRAMMING_OVERHEAD;
}

u64 fte_fifo_immediate_fte_max_delay(struct fte_fifo *fifo, const u64 curr_time)
{
	return curr_time
		+ 2 * fte_fifo_current_safety_margin(fifo)
		+ IMMEDIATE_FTE_MAX_DELAY_SETTING
		+ 2 * fifo->min_time_delta;
}

fte_fifo_itr_t fte_fifo_begin(struct fte_fifo *fifo)
{
	pr_debug("%s(fifo: %p, fifo->queued_ftes.next: %p)\n", __func__,
				fifo, fifo->queued_ftes.next);
	return fifo->queued_ftes.next;
}

fte_fifo_itr_t fte_fifo_end(struct fte_fifo *fifo)
{
	pr_debug("%s(fifo: %p, &fifo->queued_ftes: %p)\n", __func__,
				fifo, &fifo->queued_ftes);
	return &fifo->queued_ftes;
}

fte_fifo_itr_t fte_fifo_next(struct fte_fifo *fifo, fte_fifo_itr_t itr)
{
	pr_debug("%s(fifo: %p, itr->next: %p)\n", __func__,
				fifo, ((struct list_head *)itr)->next);
	return ((struct list_head *)itr)->next;
}

fte_fifo_itr_t fte_fifo_prev(struct fte_fifo *fifo, fte_fifo_itr_t itr)
{
	pr_debug("%s(fifo: %p, itr->prev: %p)\n", __func__,
				fifo, ((struct list_head *)itr)->prev);
	return ((struct list_head *)itr)->prev;
}

int fte_fifo_insert(struct fte_fifo *fifo, const u64 fte_time,
				const u64 term_bitmask, const u8 level,
				const u64 curr_time,
				size_t *earliest_modified_fte_index)
{
	struct list_head *itr;
	struct list_head *safe_itr;
	size_t _fte_index	= 0;
	u64 term_level		= level ? term_bitmask : 0;
	u64 prev_time		= 0;
	u64 prev_linestate	= fifo->current_linestate;
	u64 min_fte_time	= curr_time + fte_fifo_current_safety_margin(
								fifo);
	bool fte_inserted	= false;
	int insert_err		= 0;
	u64 time = fte_time;
	bool is_immediate_fte	= (time == NISYNC_TIME_IMMEDIATE_NANOS);
	u64 max_fte_time;

	pr_debug("%s(fifo: %p, time: %10llus %9lluns, level: %u, term_bitmask: %llx, curr_time: %10llus %9lluns)\n",
		 __func__, fifo,
		 time / NSEC_PER_SEC, time % NSEC_PER_SEC, level,
		term_bitmask, curr_time / NSEC_PER_SEC,
		curr_time % NSEC_PER_SEC);

	_fte_fifo_dbg_print(fifo);

	/*
	 * Schedule immediate FTE at specific time so we don't put it in front
	 * of fifo when scheduled events are close to being triggered
	 */
	if (is_immediate_fte) {
		time = min_fte_time;
		max_fte_time = min_fte_time + fte_fifo_current_safety_margin(fifo)
				+ IMMEDIATE_FTE_MAX_DELAY_SETTING;
	} else if (time < min_fte_time) {
		pr_debug("Can't insert FTE before min_fte_time. %llu < %llu\n",
				time, min_fte_time);
		return -ETIME;
	}

	list_for_each_safe(itr, safe_itr, &fifo->queued_ftes) {
		struct fte *curr_fte = list_entry(itr, struct fte, list_node);

		if (!fte_inserted) {
			/* if the given fte can be merged with current fte */
			if (curr_fte->time == time) {
				/*
				 * prev_time is too soon to update HW in time,
				 *   aka prev_time is <= safety margin from now
				 *   as provided by min_fte_time
				 */
				if (prev_time && (prev_time < min_fte_time)) {
					if (is_immediate_fte) {
						time++;

						prev_time = curr_fte->time;
						prev_linestate = curr_fte->linestate;
						_fte_index++;

						continue;
					}
					pr_debug("Failed merging FTE due to prev_time < min_fte_time\nprev_time:    %llu\ntime:         %llu\nmin_fte_time: %llu\n\n",
							prev_time, time,
							min_fte_time);
					return -ETIME;
				}
				fte_fifo_merge_event(itr, term_bitmask,
						     term_level);
				fte_inserted = true;
			} else if (time < curr_fte->time) {
				if (!is_immediate_fte) {
					/*
					 * if a minimum time delta violation will occur
					 *    OR
					 * prev_time is too soon to update HW in time,
					 *   aka prev_time is <= safety margin from now
					 *   as provided by min_fte_time
					 * (prev_time must be within safety margin to
					 * prevent hardware from advancing to an invalid
					 * event while we are reprogramming)
					 */
					if (fte_fifo_min_time_delta_err(prev_time, time,
						curr_fte->time)) {
						pr_debug("Failed inserting FTE in the middle of fifo\nprev_time:    %llu\ntime:         %llu\nmin_fte_time: %llu\n\n",
							prev_time, time, min_fte_time);
						return -EFTE_REARM_TIME_VIOLATION;
					}
					if ((prev_time && (prev_time < min_fte_time))) {
						pr_debug("Failed inserting FTE in the middle of fifo\nprev_time:    %llu\ntime:         %llu\nmin_fte_time: %llu\n\n",
							prev_time, time, min_fte_time);
						return -ETIME;
					}
					insert_err = fte_fifo_insert_before(itr, time,
						term_bitmask, term_level,
						prev_linestate);
					if (insert_err)
						return insert_err;
					fifo->size++;
				} else {
					if (prev_time && (prev_time < min_fte_time)) {
						if (curr_fte->time < max_fte_time) {
							/*
							 * Next event is not too far in
							 * future so schedule after it.
							 */
							time = curr_fte->time + 1;

							prev_time = curr_fte->time;
							prev_linestate = curr_fte->linestate;
							_fte_index++;

							continue;
						} else {
							/*
							 * Next event if too far
							 * in future. So wait
							 * until current event
							 * is triggered and
							 * retry scheduling the
							 * immediate FTE
							 */
							udelay(((prev_time - curr_time) / 1000) + 2);
							return -EAGAIN;
						}
					}

					/*
					 * Adjust time to prevent minimum time
					 * delta violation
					 */
					time = max(time, prev_time +
						fifo->min_time_delta);
					if (curr_fte->time && time +
					    fifo->min_time_delta >
					    curr_fte->time)
						time = curr_fte->time;

					if (curr_fte->time == time) {
						fte_fifo_merge_event(itr, term_bitmask,
								     term_level);
					} else {
						insert_err =
							fte_fifo_insert_before(itr,
								time,
								term_bitmask, term_level,
								prev_linestate);
						if (insert_err)
							return insert_err;
						fifo->size++;
					}
				}
				fte_inserted = true;
			} else {
				prev_time = curr_fte->time;
				prev_linestate = curr_fte->linestate;
				_fte_index++;
			}
		}
		/*
		 * Once we insert an fte into the fifo, we need to update the
		 * following linestates of the fte_fifo until we reach the next
		 * scheduled fte on the given terminal mask.
		 */
		if (fte_inserted) {
			/*
			 * If merged, must continue propagating level change
			 * through FTEs.
			 */
			if (curr_fte->time == time) {
				continue;
			/*
			 * If we find an FTE that explicitly sets the level for
			 * this terminal, stop updating FTEs.
			 */
			} else if (curr_fte->term & term_bitmask) {
				break;
			} else {
				curr_fte->linestate &= ~term_bitmask;
				curr_fte->linestate |= term_level;
			}
		}
	}

	/* insert at the end of the fifo */
	if (!fte_inserted) {
		if (fifo->size
			&& (fte_fifo_min_time_delta_err(prev_time, time, (u64)0))) {
			if (is_immediate_fte) {
				/*
				 * Adjust time to prevent minimum time
				 * delta violation
				 */
				time = max(time, prev_time +
					fifo->min_time_delta);
			} else {
				pr_debug("Failed adding fte to the end of fifo\nprev_time:    %llu\ntime:         %llu\nmin_fte_time: %llu\n\n",
					prev_time, time, min_fte_time);
				return -EFTE_REARM_TIME_VIOLATION;
			}
		}
		insert_err = fte_fifo_insert_before(itr, time, term_bitmask,
						    term_level, prev_linestate);
		if (insert_err)
			return insert_err;
		_fte_index = fifo->size++;
	}

	if (earliest_modified_fte_index
			&& _fte_index < *earliest_modified_fte_index)
		*earliest_modified_fte_index = _fte_index;

	_fte_fifo_dbg_print(fifo);
	pr_debug("() =>\n");

	return 0;
}

int fte_fifo_merge(struct fte_fifo *fifo, struct fte_fifo *events,
				const u64 curr_time,
				size_t *earliest_modified_fte_index)
{
	/* TODO: write efficient algorithm to avoid O(n^2) runtime */
	int err = 0;
	struct fte *curr_fte = NULL;

	list_for_each_entry(curr_fte, &events->queued_ftes, list_node) {
		err = fte_fifo_insert(fifo, curr_fte->time, curr_fte->term,
				hweight64(curr_fte->level & curr_fte->term),
				curr_time, earliest_modified_fte_index);
		if (err)
			break;
	}
	_fte_fifo_dbg_print(fifo);
	return err;
}

void fte_fifo_peek(struct fte_fifo *fifo, fte_fifo_itr_t itr, u64 *time,
				u64 *term_bitmask, u64 *linestate)
{
	struct fte *curr_fte = list_entry(itr, struct fte, list_node);

	pr_debug("%s(fifo: %p, itr: %p)\n", __func__, fifo, itr);

	_fte_fifo_dbg_print(fifo);

	if (time)
		*time = curr_fte->time;
	if (term_bitmask)
		*term_bitmask = curr_fte->term;
	if (linestate)
		*linestate = curr_fte->linestate;
}

int fte_fifo_del(struct fte_fifo *fifo, const u64 time, const u64 term_bitmask,
				const u64 curr_time,
				size_t *earliest_modified_fte_index)
{
	struct list_head *curr_itr;
	struct list_head *safe_itr;
	size_t _fte_index	= 0;
	u64 prev_linestate	= fifo->current_linestate;
	u64 prev_time		= 0;
	u64 min_fte_time	= curr_time + fte_fifo_current_safety_margin(
								fifo);
	bool fte_found		= false;

	pr_debug("%s(fifo: %p, term_bitmask: %llx, time: %10llus %9lluns, min_fte_time: %10llus %9lluns)\n",
		 __func__, fifo, term_bitmask,
		time		/ NSEC_PER_SEC, time		% NSEC_PER_SEC,
		min_fte_time	/ NSEC_PER_SEC, min_fte_time	% NSEC_PER_SEC);

	_fte_fifo_dbg_print(fifo);

	list_for_each_safe(curr_itr, safe_itr, &fifo->queued_ftes) {
		struct fte *curr_fte = list_entry(curr_itr, struct fte,
						list_node);
		if (!fte_found) {
			if (curr_fte->time == time) {
				/*
				 * prev_time is too soon to update HW in time,
				 *   aka prev_time is <= safety margin from now
				 *   as provided by min_fte_time
				 */
				if (prev_time && (prev_time < min_fte_time)) {
					pr_debug("Failed deleting FTE due to prev_time < min_fte_time\nprev_time:    %llu\ntime:         %llu\nmin_fte_time: %llu\n\n",
							prev_time, time,
							min_fte_time);
					return -ETIME;
				}
				curr_fte->term &= ~term_bitmask;
				if (curr_fte->term) {
					curr_fte->level &= ~term_bitmask;
					curr_fte->linestate &= ~term_bitmask;
					curr_fte->linestate |= prev_linestate &
								term_bitmask;
				} else {
					list_move(curr_itr, &fifo->pool);
					fifo->size--;
				}
				fte_found = true;
			} else {
				_fte_index++;
				prev_time = curr_fte->time;
			}
		} else if (curr_fte->term & term_bitmask) {
			/* line is explicitly set with an fte so we are done */
			break;
		} else {
			/*
			 * Once we find the event being del we need to clear
			 * following events of its impact until the next event
			 * on term_bitmask
			 */
			curr_fte->linestate &= ~term_bitmask;
			curr_fte->linestate |= prev_linestate &
							term_bitmask;
		}
		prev_linestate = curr_fte->linestate;
	}

	if (earliest_modified_fte_index &&
			_fte_index < *earliest_modified_fte_index)
		*earliest_modified_fte_index = _fte_index;

	_fte_fifo_dbg_print(fifo);

	if (!fte_found) {
		pr_debug("no matching FTE found\n");
		return -ENOENT;
	}

	pr_debug("() =>\n");
	return 0;
}

void fte_fifo_del_older_than(struct fte_fifo *fifo, const u64 time)
{
	struct fte *curr_fte;
	struct fte *safe_fte;

	pr_debug("%s(fifo: %p, time: %llu)\n", __func__, fifo, time);
	_fte_fifo_dbg_print(fifo);

	list_for_each_entry_safe(curr_fte, safe_fte, &fifo->queued_ftes,
					list_node) {
		if (curr_fte->time >= time)
			break;

		fifo->current_linestate = curr_fte->linestate;
		list_move(&curr_fte->list_node, &fifo->pool);
		fifo->size--;
	}

	_fte_fifo_dbg_print(fifo);

	pr_debug("() =>\n");
}

void fte_fifo_clear(struct fte_fifo *fifo)
{
	pr_debug("%s(fifo: %p)\n", __func__, fifo);

	_fte_fifo_dbg_print(fifo);

	_fte_fifo_reset(fifo);

	_fte_fifo_dbg_print(fifo);

	pr_debug("() =>\n");
}

int fte_fifo_clear_terminal(struct fte_fifo *fifo, const u64 term_bitmask,
				const u64 curr_time,
				size_t *earliest_modified_fte_index)
{
	struct list_head *curr_itr;
	struct list_head *safe_itr;
	size_t _fte_index	= 0;
	u64 prev_linestate	= fifo->current_linestate;
	u64 prev_time		= 0;
	u64 min_fte_time	= curr_time + fte_fifo_current_safety_margin(
								fifo);
	bool found_first_event	= false;

	pr_debug("%s(fifo: %p, term_bitmask: %llx)\n", __func__,
			   fifo, term_bitmask);
	_fte_fifo_dbg_print(fifo);

	list_for_each_safe(curr_itr, safe_itr, &fifo->queued_ftes) {
		struct fte *curr_fte = list_entry(curr_itr, struct fte,
						list_node);
		/* any FTE on terminals in term_bitmask will be removed */
		if (curr_fte->term & term_bitmask) {
			/*
			 * prev_time is too soon to update HW in time,
			 *   aka prev_time is <= safety margin from now
			 *   as provided by min_fte_time
			 */
			if (!found_first_event) {
				/*
				 * This event cannot be removed if either
				 * of the following is true
				 *  * If this is the first event of the fifo
				 *    and curr_fte->time is < min_fte_time
				 *
				 *  * If there was a previous event which was
				 *    scheduled before min_fte_time
				 */
				if ((!prev_time &&
					(curr_fte->time < min_fte_time))
					|| (prev_time &&
					(prev_time < min_fte_time))) {
				   _fte_index++;
				   prev_time = curr_fte->time;
				   prev_linestate = curr_fte->linestate;
				   continue;
				}
			}
			found_first_event = true;

			curr_fte->term &= ~term_bitmask;
			if (curr_fte->term) {
				curr_fte->level &= ~term_bitmask;
				curr_fte->linestate &= ~term_bitmask;
				curr_fte->linestate |= prev_linestate &
							term_bitmask;
			} else {
				/*
				 * no other terminals with an event
				 * at this time
				 */
				list_move(curr_itr, &fifo->pool);
				fifo->size--;
			}
		} else if (!found_first_event) {
			_fte_index++;
			prev_time = curr_fte->time;
		} else {
			/*
			 * Once we find an event all following events need to
			 * be cleared of changes on terminals in term_bitmask
			 */
			curr_fte->linestate &= ~term_bitmask;
			curr_fte->linestate |= prev_linestate & term_bitmask;
		}
		prev_linestate = curr_fte->linestate;
	}

	if (earliest_modified_fte_index &&
			_fte_index < *earliest_modified_fte_index)
		*earliest_modified_fte_index = _fte_index;

	_fte_fifo_dbg_print(fifo);

	pr_debug("() =>\n");
	return 0;
}

bool fte_fifo_is_terminal_scheduled(struct fte_fifo *fifo,
				const u64 term_bitmask)
{
	struct fte *curr_fte;

	pr_debug("%s(fifo: %p, term_bitmask: %llx)\n", __func__,
			   fifo, term_bitmask);
	_fte_fifo_dbg_print(fifo);

	list_for_each_entry(curr_fte, &fifo->queued_ftes, list_node) {
		if (term_bitmask & curr_fte->term)
			return true;
	}

	return false;
}

void fte_fifo_terminal_events(struct fte_fifo *fifo, const u64 term_bitmask,
				u32 *num_ftes)
{
	struct fte *curr_fte;

	pr_debug("%s(fifo: %p, term_bitmask: %llx)\n", __func__,
			   fifo, term_bitmask);
	_fte_fifo_dbg_print(fifo);

	*num_ftes = 0;
	list_for_each_entry(curr_fte, &fifo->queued_ftes, list_node) {
		*num_ftes += hweight64(curr_fte->term & term_bitmask);
	}
}

int fte_fifo_get_terminal_events(struct fte_fifo *fifo, const u64 term_bitmask,
					struct fte_fifo *matching_events)
{
	struct fte *curr_fte = NULL;

	list_for_each_entry(curr_fte, &fifo->queued_ftes, list_node) {
		/*
		 * TODO: write more efficient algorithm, since we'll always be
		 * inserting at the back
		 */
		if (term_bitmask & curr_fte->term) {
			pr_debug("found matching event, time: %10llus %9lluns, level: 0x%016llx\n",
				curr_fte->time / NSEC_PER_SEC,
				curr_fte->time % NSEC_PER_SEC, curr_fte->level);
			if (matching_events->size ==
				matching_events->capacity) {
				pr_debug("copying event would exceed capacity of matching_events (%zu)\n",
					matching_events->capacity);
				return -ENOMEM;
			}

			/*
			 * these events were screened for timing errors when
			 * they were inserted into the fifo, so we don't expect
			 * any errors from fte_fifo_insert
			 */
			fte_fifo_insert(matching_events, curr_fte->time,
				term_bitmask,
				hweight64(curr_fte->level & term_bitmask),
				0, NULL);
		}
	}
	return 0;
}

void fte_fifo_reset(struct fte_fifo *fifo)
{
	pr_debug("%s(fifo: %p)\n", __func__, fifo);

	fte_fifo_clear(fifo);
	fifo->current_linestate = 0;

	pr_debug("() =>\n");
}

void fte_fifo_check_errors(struct fte_fifo *fifo, const u64 term_bitmask,
				u32 *fte_errors)
{
	struct fte *curr_fte;
	u64 prev_linestate;
	u64 err_bitmask;

	pr_debug("%s(fifo: %p, term_bitmask: %llu, fte_errors: %p)\n", __func__,
			fifo, term_bitmask, fte_errors);

	*fte_errors = 0;

	prev_linestate = fifo->current_linestate;

	list_for_each_entry(curr_fte, &fifo->queued_ftes, list_node) {
		/* Get the bitmask of the lines that did not transition */
		err_bitmask = ~(curr_fte->linestate ^ prev_linestate);
		/* Mask off the bits in which we are not interested */
		err_bitmask &= curr_fte->term & term_bitmask;
		/* Count the number of bits set */
		*fte_errors += hweight64(err_bitmask);
		prev_linestate = curr_fte->linestate;
	}

	pr_debug("(fte_errors: %u) =>\n", fte_errors ? *fte_errors : 0);
}

