/************************************************************************
 * fte_fifo.h
 *
 * Author: Klemen Vodopivec (Cosylab)
 * Copyright (C) 2010-2015 ITER Organization
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License. See the file COPYING in the main directory of the Linux
 * distribution for more details.
 ************************************************************************/

#ifndef _FTE_FIFO_H
#define _FTE_FIFO_H

#include "nisync.h"

struct fte_fifo;
typedef void *fte_fifo_itr_t;

/*
 * Emperical testing performed by NI during Nov 2018 helped us choose these
 * values - specifically the stress FTE VI was run with different values to
 * arrive at these numbers.
 * Eventually, this decision should be left to the system integrator. On
 * systems with large PCI latency (e.g., MXI), this margin is not guaranteed
 * to be safe. The right balance of performance versus correctness is best
 * determined in a final system.
 * TODO: Make these values configurable by the user via sysfs.
 */
#define FTE_FIFO_EVENT_PROGRAMMING_TIME (4 * NSEC_PER_USEC)
#define FTE_FIFO_REPROGRAMMING_OVERHEAD (500 * NSEC_PER_USEC)
#define IMMEDIATE_FTE_MAX_DELAY_SETTING (100 * NSEC_PER_USEC)

/*
 * Allocate FTE FIFO.
 */
struct fte_fifo *fte_fifo_alloc(const size_t capacity);

/*
 * Deallocate FTE FIFO.
 */
void fte_fifo_destroy(struct fte_fifo *fifo);

/*
 * Returns capacity of FIFO.
 */
size_t fte_fifo_capacity(struct fte_fifo *fifo);

/*
 * Returns number of elements in FIFO.
 */
size_t fte_fifo_size(struct fte_fifo *fifo);

/*
 * Sets the minimum delta in time allowed between two consecutive FTEs.
 */
void fte_fifo_set_min_time_delta(struct fte_fifo *fifo,
				const u32 min_time_delta);

/*
 * Returns the number of nanoseconds it would take to safely reschedule all
 * existing events in the buffer.
 */
u64 fte_fifo_current_safety_margin(struct fte_fifo *fifo);

/*
 * Returns the maximum delay in scheduling immediate FTE
 */
u64 fte_fifo_immediate_fte_max_delay(struct fte_fifo *fifo, const u64 curr_time);

/*
 * Returns an iterator referring to the first element in the fifo.
 */
fte_fifo_itr_t fte_fifo_begin(struct fte_fifo *fifo);

/*
 * Returns an iterator referring to the 'past-the-end' element in the fifo.
 */
fte_fifo_itr_t fte_fifo_end(struct fte_fifo *fifo);

/*
 * Returns an iterator referring to the element after the element referred to
 * by the given iterator.
 */
fte_fifo_itr_t fte_fifo_next(struct fte_fifo *fifo, fte_fifo_itr_t itr);

/*
 * Returns an iterator referring to the element before the element referred to
 * by the given iterator.
 */
fte_fifo_itr_t fte_fifo_prev(struct fte_fifo *fifo, fte_fifo_itr_t itr);

/*
 * Insert the given future time event into the FIFO, sorted based on time in
 * chronological order. On success, the return value is zero and the parameter
 * 'earliest_modified_fte_index' is set to the index of the inserted event.
 * Otherwise, the return value is ENOMEM if the FIFO is full or ETIME if either
 * the minimum time between FTEs has been violated or if the given FTE or
 * previous FTE is too soon to safely schedule without corrupting the FTE HW
 * FIFO. term_bitmask can only specify _one_ terminal per call to
 * fte_fifo_insert.
 */
int fte_fifo_insert(struct fte_fifo *fifo, const u64 fte_time,
				const u64 term_bitmask, const u8 level,
				const u64 curr_time,
				size_t *earliest_modified_fte_index);

/*
 * Merge the entries from other_fifo into fifo. On success,
 * earliest_modified_fte_index is set to the index of the first event in fifo
 * that was changed. On error, some events may have been merged into fifo
 * before the error occurred.
 * TODO: If we gave callers more information about the memory allocation
 * strategy in fte_fifo, they could make better decisions about when errors
 * might arrive.
 */
int fte_fifo_merge(struct fte_fifo *fifo, struct fte_fifo *other_fifo,
				const u64 curr_time,
				size_t *earliest_modified_fte_index);

/*
 * Read the future time event referenced by the given iterator.
 */
void fte_fifo_peek(struct fte_fifo *fifo, fte_fifo_itr_t itr, u64 *time,
				u64 *term_bitmask, u64 *linestate);

/*
 * Deletes the given future time event from the FIFO.
 * On success, the return value is zero and the parameter
 * 'earliest_modified_fte_index' is set to the index of the first event after
 * the deleted event. Otherwise, the return value is ETIME if the FTE before
 * the one being deleted is too soon to safely modify the FTE HW FIFO without
 * corrupting it. term_bitmask can only specify _one_ terminal per call to
 * fte_fifo_del.
 */
int fte_fifo_del(struct fte_fifo *fifo, const u64 time, const u64 term_bitmask,
				const u64 curr_time,
				size_t *earliest_modified_fte_index);

/*
 * Delete all elements strictly older than time.
 */
void fte_fifo_del_older_than(struct fte_fifo *fifo, const u64 time);

/*
 * Clear all entries in FTE FIFO
 */
void fte_fifo_clear(struct fte_fifo *fifo);

/*
 * Clears all entries in the FTE FIFO on the terminals given by term_bitmask.
 * On success, the return value is zero and the parameter
 * 'earliest_modified_fte_index' is set to the index of the earliest event that
 * was modified. Otherwise, the return value is ETIME if the FTE before the
 * first one being deleted is too soon to safely modify the FTE HW FIFO without
 * corrupting it.
 */
int fte_fifo_clear_terminal(struct fte_fifo *fifo, const u64 term_bitmask,
				const u64 curr_time,
				size_t *earliest_modified_fte_index);

/*
 * Tests if there are any entries in the FTE FIFO on any of the terminals given
 * by term_bitmask
 */
bool fte_fifo_is_terminal_scheduled(struct fte_fifo *fifo,
				const u64 term_bitmask);

/*
 * Counts the number of events in the FTE FIFO on the terminal given by
 * term_bitmask.
 */
void fte_fifo_terminal_events(struct fte_fifo *fifo, const u64 term_bitmask,
				u32 *num_ftes);

/*
 * Searches fifo for events matching term_bitmask and copies them into
 * matching_events. If matching_events cannot hold all of the events that match,
 * will copy as many as possible and return ENOMEM. term_bitmask can contain any
 * combination of terminals. Information about simultaneous events on other
 * terminals (not matching term_bitmask) will not be copied.
 */
int fte_fifo_get_terminal_events(struct fte_fifo *fifo,
				const u64 term_bitmask,
				struct fte_fifo *matching_events);

/*
 * Reset FTE FIFO to initial state
 */
void fte_fifo_reset(struct fte_fifo *fifo);

/*
 * Checks that every scheduled FTE level is different from the previous level,
 * checking only terminals specified by the term_bitmask. It returns number of
 * FTE that did not result in a level change.
 */
void fte_fifo_check_errors(struct fte_fifo *fifo, const u64 term_bitmask,
				u32 *fte_errors);

#endif /* _FTE_FIFO_H */
