/**
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This header provides implementations for some required functions which are not
 * available in older kernels.
 * Implementations are taken from the first kernel version the function was
 * added to.
 */
#ifndef KERNEL_COMPAT_H_
#define KERNEL_COMPAT_H_

#include <linux/version.h>
#include <linux/phy.h>

#ifdef IOPOLL_H_HAS_READ_POLL
#include <linux/iopoll.h>
#else
#include <linux/hrtimer.h>
#define readx_poll_timeout(op, addr, val, cond, sleep_us, timeout_us)   \
({ \
        ktime_t timeout = ktime_add_us(ktime_get(), timeout_us); \
        might_sleep_if(sleep_us); \
        for (;;) { \
                (val) = op(addr); \
                if (cond) \
                        break; \
                if (timeout_us && ktime_compare(ktime_get(), timeout) > 0) { \
                        (val) = op(addr); \
                        break; \
                } \
                if (sleep_us) \
                        usleep_range((sleep_us >> 2) + 1, sleep_us); \
        } \
        (cond) ? 0 : -ETIMEDOUT; \
})

#define readl_poll_timeout(addr, val, cond, delay_us, timeout_us) \
	readx_poll_timeout(readl, addr, val, cond, delay_us, timeout_us)
#endif


#ifndef PHY_H_HAS_DEVM_MDIOBUS
static void _devm_mdiobus_free(struct device *dev, void *res)
{
	mdiobus_free(*(struct mii_bus **)res);
}

static struct mii_bus *devm_mdiobus_alloc_size(struct device *dev,
							int sizeof_priv)
{
	struct mii_bus **ptr, *bus;

	ptr = devres_alloc(_devm_mdiobus_free, sizeof(*ptr), GFP_KERNEL);
	if (!ptr)
		return NULL;

	/* use raw alloc_dr for kmalloc caller tracing */
	bus = mdiobus_alloc_size(sizeof_priv);
	if (bus) {
		*ptr = bus;
		devres_add(dev, ptr);
	} else {
		devres_free(ptr);
	}

	return bus;
}

static inline struct mii_bus *devm_mdiobus_alloc(struct device *dev)
{
	return devm_mdiobus_alloc_size(dev, 0);
}
#endif

#endif /* KERNEL_COMPAT_H_ */
