/*************************************************************************
 *
 * Description   : Types and ioctl interface of the nisync driver
 *
 * Authors       : Janez Golob (Cosylab)
 *                 James Blair (National Instruments)
 *                 Mike Hoogendyk (National Instruments)
 *
 * Copyright (c) : 2010-2015 ITER Organization,
 *                 CS 90 046
 *                 13067 St. Paul-lez-Durance Cedex
 *                 France
 *
 * This file is part of ITER CODAC software.
 * For the terms and conditions of redistribution or use of this software
 * refer to the file ITER-LICENSE.TXT located in the top level directory
 * of the distribution package.
 ************************************************************************/

/**
 * @file nisync.h
 * @brief Types and ioctl interface of the nisync driver
 *
 * This file contains ioctl implementation details of the nisync driver.
 * While types contained in this header define valid values of types provided
 * and used in the nisync-lib.h and nisync-lib-ptp.h API, the relevant types in
 * those headers should be used instead by other user programs.
 */

#ifndef NISYNC_H_
#define NISYNC_H_

#ifndef __KERNEL__
#  include <stdint.h>
#  include <time.h>
#else
#  include <linux/types.h>
#  include <linux/time.h>
#endif

/*
 * TODO: Inconsistently using enums in structs and in API instead of just as
 * valid values. Future API and usage should be to explicitly call out sized
 * types in API and in structures passed in the API to guarantee compatibility.
 * enums are then only for convenience in specifying the list of valid values.
 */

/**
 * @brief nisync Supported Terminals.
 *
 * Note that each hardware device supports a subset of the listed terminals;
 * refer to the user manual or terminals created in /dev/ for each device.
 */
enum nisync_terminal {
	/* Trigger Terminals */
	NISYNC_PFI0 = 0,
	NISYNC_PFI1 = 1,
	NISYNC_PFI2 = 2,

	NISYNC_RTSI0 = 3,
	NISYNC_RTSI1 = 4,
	NISYNC_RTSI2 = 5,
	NISYNC_RTSI3 = 6,
	NISYNC_RTSI4 = 7,
	NISYNC_RTSI5 = 8,
	NISYNC_RTSI6 = 9,
	NISYNC_RTSI7 = 10,

	NISYNC_PXI_TRIG0 = 11,
	NISYNC_PXI_TRIG1 = 12,
	NISYNC_PXI_TRIG2 = 13,
	NISYNC_PXI_TRIG3 = 14,
	NISYNC_PXI_TRIG4 = 15,
	NISYNC_PXI_TRIG5 = 16,
	NISYNC_PXI_TRIG6 = 17,
	NISYNC_PXI_TRIG7 = 18,

	NISYNC_PXI_STAR0 = 19,
	NISYNC_PXI_STAR1 = 20,
	NISYNC_PXI_STAR2 = 21,
	NISYNC_PXI_STAR3 = 22,
	NISYNC_PXI_STAR4 = 23,
	NISYNC_PXI_STAR5 = 24,
	NISYNC_PXI_STAR6 = 25,
	NISYNC_PXI_STAR7 = 26,
	NISYNC_PXI_STAR8 = 27,
	NISYNC_PXI_STAR9 = 28,
	NISYNC_PXI_STAR10 = 29,
	NISYNC_PXI_STAR11 = 30,
	NISYNC_PXI_STAR12 = 31,
	NISYNC_CLK10 = 36,
	NISYNC_GND = 37,

	/* Non-board terminals */
	NISYNC_SW = 40,
	NISYNC_ETHIN = 41,
	NISYNC_ETHOUT = 42,
	NISYNC_DEVICE = 43,
	NISYNC_TERMINAL_INVALID = 45,
};

/** Maximum number of terminals a device can have */
#define NISYNC_MAX_TERMINALS NISYNC_TERMINAL_INVALID
/**
 * Placeholder terminal identifier guaranteed to be larger than all valid
 * terminal identifiers.
 */
#define NISYNC_LAST_TERMINAL NISYNC_TERMINAL_INVALID

/**
 * Trigger terminal level definitions
 */
enum nisync_level {
	NISYNC_LEVEL_LOW	= 0,
	NISYNC_LEVEL_HIGH	= 1,
	NISYNC_LEVEL_INVALID	= 2,
};

/**
 * Edge Definitions used for activeEdge and detectedEdge
 */
enum nisync_edge {
	NISYNC_EDGE_RISING	= 0,
	NISYNC_EDGE_FALLING	= 1,
	NISYNC_EDGE_ANY		= 2,
	NISYNC_EDGE_INVALID	= 3,
};

/**
 * Terminal Timestamps in nanos for use with nisync_read_timestamps_ns
 */
struct nisync_timestamp_nanos {
	uint8_t edge;
	uint64_t nanos;
};

/**
 * Immediate time for use with _ns API to signify creation of an event as soon
 * as technically possible.
 */
#define NISYNC_TIME_IMMEDIATE_NANOS 0
/**
 * Indefinite time for use with _ns API to signify running a clock indefinitely
 * or forever.
 */
#define NISYNC_TIME_INDEFINITE_NANOS ULLONG_MAX

#if NISYNC_COMPAT_ERRORS
#define EFTE_REARM_TIME_VIOLATION	200
#else
#define EFTE_REARM_TIME_VIOLATION	ETIME
#endif

/**
 * Trigger Terminal Connection Mode Definitions
 */
enum nisync_invert {
	NISYNC_DONT_INVERT	= 0,
	NISYNC_INVERT		= 1,
};

/**
 * Trigger routing synchronization settings
 */
enum nisync_synchronous {
	NISYNC_ASYNCHRONOUS	= 0,	/**< Make route asynchronous */
	NISYNC_SYNCHRONOUS	= 1,	/**<
					 * Make route synchronous to
					 * backplane 10MHz clock
					 */
};

/**
 * Supported clocks
 */
enum nisync_clock {
	NISYNC_CLK_NONE		= 0,
	NISYNC_BOARD_CLK	= 1,	/**< On board TCXO 10MHz clock */
	NISYNC_BACKPLANE_CLK	= 2,	/**< Backplane 10MHz synchronization clock */
	NISYNC_TIMEKEEPER_CLK	= 3,	/**< Clock used for IEEE-1588v2 time */
	NISYNC_CLKIN		= 4,	/**< CLKIN connector on the front panel */
	NISYNC_CLKOUT		= 5,	/**< CLKOUT connector on the front panel */
	NISYNC_CLK_INVALID	= 6
};

/**
 * Latency attributes that can be queried by nisync_get_attribute
 */
enum nisync_attribute {
	NISYNC_FTE_LATENCY		= 0,
	NISYNC_INPUT_TIMESTAMP_LATENCY	= 1,
	NISYNC_SYS_TIME_LATENCY		= 2,
	NISYNC_ENET_TRANSMIT_LATENCY	= 3,
	NISYNC_ENET_RECEIVE_LATENCY	= 4,
};

/* TODO: missing PRE_MASTER, PASSIVE, and UNCALIBRATED states */
/**
 * PTP States as defined in the IEEE-1588v2 Specification
 */
enum nisync_ptpd_state {
	NISYNC_PTPD_DISABLED,
	NISYNC_PTPD_INITIALIZING,
	NISYNC_PTPD_SLAVE,
	NISYNC_PTPD_LISTENING,
	NISYNC_PTPD_MASTER,
	NISYNC_PTPD_FAULTY,
};

/**
 * nisync device information
 */
struct nisync_device_info {
	char driver_version[30];
	char model[10];
	char name[10];
	uint64_t time_since_sync;
	uint32_t serial_number;
	uint32_t signature;
	uint32_t revision;
	uint32_t oldest_compatible_revision;
	uint32_t hardware_revision;
};

/*
 * TODO: more consistently use status values to make it clearer what recovery
 * option to take.
 */
/**
 * Status values for nisync boards.
 */
enum nisync_status {
	/* Generic errors from SYNC-SRS-F-01 */
	NISYNC_STATUS_OK			= 0,
	/**< No error, fully operational state */
	NISYNC_STATUS_INITIALIZING		= 1,
	/**< During initialization phase board status shall go to this state */
	NISYNC_STATUS_RESETTING			= 2,
	/**< During reset phase board status shall go to this state */
	NISYNC_STATUS_HW_ERROR			= 3,
	/**< Firmware or hardware self-test failure */
	NISYNC_STATUS_UNAVAILABLE		= 4,
	/**< Cannot communicate with board */
	NISYNC_STATUS_STATIC_CONF_ERROR		= 5,
	/**<
	 * Static configuration error during initialization. Need to change
	 * board configuration.
	 */
	NISYNC_STATUS_DYNAMIC_CONF_ERROR	= 6,
	/**<
	 * Illegal configuration issued from a user task, which prevents normal
	 * operation.
	 */
	NISYNC_STATUS_RESERVED			= 7,
	/**< N/A for this board */

	/* Custom nisync status values */
	NISYNC_STATUS_HW_FIFO_OVERFLOW		= 8,
	/**<
	 * The HW timestamp buffer has overflowed because it is filling up
	 * faster than it can be processed by the driver software and timestamp
	 * data has been lost.
	 */
	NISYNC_STATUS_HW_FPGA_NOT_READY		= 9,
	/**<
	 * Firmware failed to load on reset. The board is not ready and should
	 * be reset again before attempting further use.
	 */
	NISYNC_STATUS_HW_CLK_NOT_LOCKED		= 10,
	/**<
	 * Board clock failed to lock to source clock. The board is not ready
	 * and should be reset before attempting further use.
	 */
	NISYNC_STATUS_DRV_BUFFER_OVERFLOW	= 11,
	/**<
	 * The driver software timestamp buffer for the terminal has overflowed
	 * before being read and timestamp data has been lost.
	 */
};

#define TENS_OF_YEARS(x)	(x >> 28)
#define UNITS_OF_YEARS(x)	((x >> 24) & 0xF)
#define HUNDREDS_OF_DAYS(x)	(x >> 30)
#define TENS_OF_DAYS(x)		((x >> 24) & 0xF)
#define UNITS_OF_DAYS(x)	((x >> 20) & 0xF)
#define TENS_OF_HOURS(x)	((x >> 28) & 0x3)
#define UNITS_OF_HOURS(x)	((x >> 16) & 0xF)
#define TENS_OF_MINS(x)		((x >> 12) & 0x7)
#define UNITS_OF_MINS(x)	((x >> 8) & 0xF)
#define TENS_OF_SECS(x)		((x >> 4) & 0x7)
#define UNITS_OF_SECS(x)	(x & 0xF)
#define STRAIGHT_BINARY_SECS(x)	(x & 0x1FFFF)

struct nisync_irig_packet {
	uint32_t time_and_days;
	uint32_t year_and_seconds;
	uint64_t arrival_hw_ts;
};

/**
 * arrival_hw_ts is the timestamp of the start of frame of an IRIG packet
 * The rest of the fields in this struct are the fields decoded from the
 * irig packet.
 * These values are reported by the hardware only after the whole packet
 * is decoded which is 10ms before the next seconds boundary.
 * So even though all the values are for the same second, we will probably
 * be into the next second before we read these values.
 */
struct nisync_irig_ts {
	uint16_t year;
	uint16_t day;
	uint16_t hour;
	uint16_t min;
	uint16_t sec;
	uint32_t straight_binary_seconds;
	uint64_t arrival_hw_ts;
};

/**
 * IRIG mode
 */
enum nisync_irig_mode {
	NISYNC_IRIGB_DC = 0,
	NISYNC_IRIGB_AM = 1,
	NISYNC_IRIG_INVALID = 2,
};

/*! \cond PRIVATE_IMPLEMENTATION_DETAIL */

/* Future Time Event structure */
struct nisync_write_buf {
	enum {
		NISYNC_WRITE_FTE,
		NISYNC_WRITE_CLOCK,
		NISYNC_WRITE_PULSE,
		NISYNC_ABORT_FTE,
		NISYNC_ABORT_CLOCK,
	} mode;

	union {
		struct {
			enum nisync_level level;/* level of FTE (0 or 1) */
			uint64_t time;		/* absoulute time of the FTE in ns */
		} fte;

		struct {
			uint64_t start;		/* absolute start time of the clock in ns */
			uint64_t end;		/* absolute end time of the clock in ns */
			uint32_t high_ticks;	/* high ticks in a period */
			uint32_t low_ticks;	/* low ticks in a period */
		} clock;

		struct {
			uint64_t time;		/* absolute start time of the pulse in ns */
			uint64_t width;		/* pulse width in ns */
			enum nisync_level level;/* line level during active pulse */
		} pulse;
	} data;
};

/* Terminal timestamp configuration */
struct nisync_timestamp_trigger {
	enum nisync_edge active_edge;
	uint32_t decimation_count;
};

/* Eth timestamp packets */
struct nisync_package_timestamp {
	uint32_t parts[5];
};

struct nisync_trigger_connection {
	enum nisync_terminal	src_terminal;
	enum nisync_terminal	dest_terminal;
	enum nisync_synchronous	synchronous;
	enum nisync_invert	invert;
};

struct nisync_clock_connection {
	/*
	 * NOTE: inconsistency between nisync_trigger_connection (src,dest) and
	 * nisync_clock_connection (dest,src)
	 */
	enum nisync_clock dest_clock;
	enum nisync_clock src_clock;
};

struct nisync_hw_attribute {
	enum nisync_attribute attr;
	uint32_t val;
};

/* Number of minor numbers requested by the nisync driver */
#define NISYNC_MINOR_NUMBERS 0xFFFF
/* Minor numbers allocated per device */
#define NISYNC_DEVICE_MINOR_STEP 50
#define NISYNC_ETH_PKT_BUFFER_SIZE 20

/*
 * IOCTL call definitions - use '0xA' as magic number.
 *
 * SET means "Set" through a ptr,
 * TTELL means "Tell" directly with the argument value
 * GET means "Get": reply by setting through a pointer
 * QUERY means "Query": response is on the return value
 * EXCHANGE means "eXchange": switch G and S atomically
 * SHIFT means "sHift": switch T and Q atomically
 *
 * _IO    an ioctl with no parameters
 * _IOW   an ioctl with write parameters (copy_from_user)
 * _IOR   an ioctl with read parameters  (copy_to_user)
 * _IOWR  an ioctl with both write and read parameters.
 *
 * 'Write' and 'read' are from the user's point of view!!!
 */
#define NISYNC_IOC_CLASS  0xA


/* HW dependent ioctl calls */

#define NISYNC_IOC_RESET			_IO(NISYNC_IOC_CLASS,	11)					/* Reset NI Sync device */

#define NISYNC_IOC_CONNECT_TRIG_TERMS		_IOW(NISYNC_IOC_CLASS,	13, struct nisync_trigger_connection *)	/* Connect trigger terminals */
#define NISYNC_IOC_DISCONNECT_TRIG_TERMS	_IOW(NISYNC_IOC_CLASS,	14, struct nisync_trigger_connection *)	/* Disconnect trigger terminals */
#define NISYNC_IOC_GET_TRIG_CONN		_IOWR(NISYNC_IOC_CLASS,	15, struct nisync_trigger_connection *)	/* Get the destination terminal */

#define NISYNC_IOC_CONNECT_CLOCK_TERM		_IOW(NISYNC_IOC_CLASS,	16, struct nisync_clock_connection *)	/* Connect trigger terminals */
#define NISYNC_IOC_GET_CLOCK_CONN		_IOWR(NISYNC_IOC_CLASS,	17, struct nisync_clock_connection *)	/* Disconnect trigger terminals */

#define NISYNC_IOC_SET_TIME			_IOW(NISYNC_IOC_CLASS,	34, uint64_t *)				/* Set time */
#define NISYNC_IOC_GET_TIME			_IOR(NISYNC_IOC_CLASS,	35, uint64_t *)				/* Get time */

#define NISYNC_IOC_ENABLE_FTE			_IO(NISYNC_IOC_CLASS,	37)					/* Enable future time events */
#define NISYNC_IOC_DISABLE_FTE			_IO(NISYNC_IOC_CLASS,	38)					/* Disable future time events */
#define NISYNC_IOC_READ_FTE_ENABLED		_IO(NISYNC_IOC_CLASS,	39)					/* Check if terminal is marked as FTE */

#define NISYNC_IOC_ENABLE_TIMESTAMP		_IOW(NISYNC_IOC_CLASS,	40, struct nisync_timestamp_trigger *)	/* Enable timestamping */
#define NISYNC_IOC_DISABLE_TIMESTAMP		_IO(NISYNC_IOC_CLASS,	41)					/* Disable timestamping */
#define NISYNC_IOC_GET_TIMESTAMP_CONF		_IOR(NISYNC_IOC_CLASS,	42, struct nisync_timestamp_trigger *)	/* Get current timestamp configuration */

#define NISYNC_IOC_ADJUST_TIME			_IOW(NISYNC_IOC_CLASS,	43, int64_t *)				/* Adjust time for specified offset */
#define NISYNC_IOC_RESET_FREQ			_IO(NISYNC_IOC_CLASS,	44)					/* Reset frequency */
#define NISYNC_IOC_GET_FREQ			_IOR(NISYNC_IOC_CLASS,	45, uint32_t *)				/* Get frequency */
#define NISYNC_IOC_SET_FREQ			_IOW(NISYNC_IOC_CLASS,	46, uint32_t *)				/* Set frequency */

#define NISYNC_IOC_SET_PTPD_STATE		_IOW(NISYNC_IOC_CLASS,	47, enum nisync_ptpd_state *)		/* Set PTP daemon state, drives LED lights and time validity */
#define NISYNC_IOC_GET_PTPD_STATE		_IOR(NISYNC_IOC_CLASS,	48, enum nisync_ptpd_state *)		/* Get PTP daemon state */

/*
 * TODO: this seems like more information about the HW than the user really
 * needs to know. Does this really need to be exposed?
 */
#define NISYNC_IOC_GET_ATTRIBUTE		_IOWR(NISYNC_IOC_CLASS,	53, struct nisync_hw_attribute *)	/* Get HW attribute */

#define NISYNC_IOC_GET_DEVICE_INFO		_IOWR(NISYNC_IOC_CLASS,	54, struct nisync_device_info *)	/* Get device info*/
#define NISYNC_IOC_ABORT_ALL_FTES		_IO(NISYNC_IOC_CLASS,	55)					/* Abort all pending FTEs */
#define NISYNC_IOC_GET_NUM_PENDING_FTES		_IOR(NISYNC_IOC_CLASS,	56, uint32_t *)				/* Read number of pending FTEs */
#define NISYNC_IOC_GET_MAX_QUEUED_FTES		_IOR(NISYNC_IOC_CLASS,	57, uint32_t *)				/* Read number of maximum schedules FTEs at any time */
#define NISYNC_IOC_CHECK_FTE_CONSISTENCY	_IOR(NISYNC_IOC_CLASS,	59, uint32_t *)				/* Check that scheduled FTE levels are alternating per terminal, return number of errors */

#define NISYNC_IOC_GET_ADJ_RATE			_IOR(NISYNC_IOC_CLASS,	60, int32_t *)				/* Get Adjustment Rate of HW timekeeper */
#define NISYNC_IOC_SET_ADJ_RATE			_IOW(NISYNC_IOC_CLASS,	61, int32_t *)				/* Set Adjustment Rate of HW timekeeper */

#define NISYNC_IOC_GET_STATUS			_IOR(NISYNC_IOC_CLASS,	62, enum nisync_status *)		/* Read device/terminal status */
#define NISYNC_IOC_GET_MAX_ADJ_RATE		_IOR(NISYNC_IOC_CLASS,	63, int32_t *)				/* Get the Maximum settable Adjustment Rate of HW timekeeper */
#define NISYNC_IOC_GET_MIN_ADJ_RATE		_IOR(NISYNC_IOC_CLASS,	64, int32_t *)				/* Get the Minimum settable Adjustment Rate of HW timekeeper */

/*
 * IOCTLs for IRIG
 */
#define NISYNC_IOC_ENABLE_IRIG			_IOW(NISYNC_IOC_CLASS,	65, uint8_t *)				/* Enable the specified IRIG mode */
#define NISYNC_IOC_GET_IRIG_TS			_IOR(NISYNC_IOC_CLASS,	66, struct nisync_irig_ts *)		/* Get the last IRIG packet data and timestamp */
#define NISYNC_IOC_DISABLE_IRIG			_IO(NISYNC_IOC_CLASS,	67)					/* Disable IRIG */

/*
 * IOCTLs for GPS
 */
#define NISYNC_IOC_GPS_ENABLE			_IO(NISYNC_IOC_CLASS,	68)					/* Reset GPS receiver and enable transmissions. */
#define NISYNC_IOC_GPS_DISABLE			_IO(NISYNC_IOC_CLASS,	69)					/* Disable GPS receiver transmissions. */
#define NISYNC_IOC_GPS_GET_TS			_IOR(NISYNC_IOC_CLASS,	70, int64_t *)				/* Get the latest GPS timestamp. */
#define NISYNC_IOC_GPS_IS_ANTENNA_CONNECTED	_IOR(NISYNC_IOC_CLASS,  71, bool *)				/* Get GPS antenna status. */
/*! \endcond */

#endif /* NISYNC_H_ */
