/*
 * nisync_buf.h
 *
 * Copyright (c) 2012-2015 National Instruments
 *
 * Authors: Josh Cartwright <josh.cartwright@ni.com>
 *          Mike Hoogendyk <mike.hoogendyk@ni.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */
#ifndef NISYNC_BUF_H_
#define NISYNC_BUF_H_

/*
 * TODO: add defines so that this header can be compiled into user mode unit
 * tests and validated.
 */
#include <asm/atomic.h>
#include <stdbool.h>
#include <linux/slab.h>
#include <linux/types.h>

/*
 * NISync buffer implementation for timestamps
 * - Terminal input timestamps leverage the bounded buffer (bb) implementation
 *   to report overflow when the buffer is full.
 * - Ethernet timestamps leverage the circular buffer (cb) implementation to
 *   always store the newest timestamp and throw away the oldest if the buffer
 *   is full.
 */
static atomic_t hack = ATOMIC_INIT(0);
struct nisync_buf {
	void *buf;

	u32 capacity;
	atomic_t size;
	u32 rd_idx;
	u32 wt_idx;
};

/* NISync buffer macros */
#if defined (NISYNC_BUF_DEBUG_PRINTS)
	#define nisync_buf_dbg_display(_msg_, _buf_) ({			\
		pr_debug(" %s: b: %p, c: %u, s: %u, ri: %u, wi: %u\n",	\
			_msg_, _buf_.buf, nisync_buf_capacity(_buf_),	\
			nisync_buf_size(_buf_),	_buf_.rd_idx, _buf_.wt_idx);\
	})
#else
	#define nisync_buf_dbg_display(_msg_, _buf_)
#endif

#define nisync_buf_init(_buf_, _buf_type_, _buf_capacity_) ({		\
	int err = 0;							\
	nisync_buf_dbg_display("nisync_buf_init <=", _buf_);		\
	(_buf_).size		= hack;					\
	(_buf_).rd_idx		= 0;					\
	(_buf_).wt_idx		= 0;					\
	(_buf_).buf = kmalloc(sizeof(_buf_type_) * (_buf_capacity_),	\
			      GFP_KERNEL);				\
	if ((_buf_).buf == NULL) {					\
		pr_debug("Failed to allocate space for nisync_buf\n");	\
		err = -ENOMEM;						\
	} else {							\
		(_buf_).capacity = (_buf_capacity_);			\
	}								\
	nisync_buf_dbg_display("nisync_buf_init =>", _buf_);		\
	pr_debug("nisync_buf_int => err: %d\n", err);			\
	err;								\
})

#define nisync_buf_destroy(_buf_) ({					\
	nisync_buf_dbg_display("nisync_buf_destroy <=", _buf_);		\
	kfree((_buf_).buf);						\
	(_buf_).capacity	= 0;					\
	(_buf_).size	= hack;						\
	(_buf_).rd_idx	= 0;						\
	(_buf_).wt_idx	= 0;						\
	nisync_buf_dbg_display("nisync_buf_destroy =>", _buf_);		\
})

#define nisync_buf_capacity(_buf_) ((_buf_).capacity)

#define nisync_buf_size(_buf_) (atomic_read(&(_buf_).size))

#define nisync_buf_empty(_buf_) (nisync_buf_size(_buf_) == 0)

#define nisync_buf_full(_buf_)						\
	(nisync_buf_size(_buf_) == nisync_buf_capacity(_buf_))

#define nisync_buf_clear(_buf_) ({					\
	nisync_buf_dbg_display("nisync_buf_clear <=", _buf_);		\
	(_buf_).size	= hack;						\
	(_buf_).rd_idx	= 0;						\
	(_buf_).wt_idx	= 0;						\
	nisync_buf_dbg_display("nisync_buf_clear =>", _buf_);		\
})

#define _nisync_buf_advance_idx(_buf_, _idx_, _step_) ({		\
	(_buf_)._idx_ = ((_buf_)._idx_ + _step_) % nisync_buf_capacity(_buf_);\
})

/* bounded buffer = return EOVERFLOW and do not overwrite data */
#define nisync_bb_write_element(_bb_, _element_) ({			\
	int err = 0;							\
	nisync_buf_dbg_display("nisync_bb_write_element <=", _bb_);	\
	if (nisync_buf_full(_bb_)) {					\
		pr_debug("Bounded buffer full, capacity: %d\n",		\
						nisync_buf_capacity(_bb_));\
		err = -EOVERFLOW;					\
	} else {							\
		((typeof(_element_) *)((_bb_).buf))[(_bb_).wt_idx] =	\
							(_element_);	\
		_nisync_buf_advance_idx(_bb_, wt_idx, 1);		\
		atomic_inc(&(_bb_).size);				\
	}								\
	nisync_buf_dbg_display("nisync_bb_write_element =>", _bb_);	\
	pr_debug("nisync_bb_write_element => err: %d\n", err);		\
	err;								\
})

/* circular buffer = silently overwrite oldest data with newest */
#define nisync_cb_write_element(_cb_, _element_) ({			\
	nisync_buf_dbg_display("nisync_cb_write_element <=", _cb_);	\
	if (nisync_buf_full(_cb_)) {					\
		pr_debug("Circular buffer full, overwriting oldest element.\n");\
		_nisync_buf_advance_idx(_cb_, rd_idx, 1);		\
		atomic_sub(1, &(_cb_).size);				\
	}								\
	((typeof(_element_) *)((_cb_).buf))[(_cb_).wt_idx] = (_element_);\
	_nisync_buf_advance_idx(_cb_, wt_idx, 1);			\
	atomic_inc(&(_cb_).size);					\
									\
	nisync_buf_dbg_display("nisync_cb_write_element =>", _cb_);	\
})

/*
 * Technically a bounded buffer checks if the buffer is empty on read where as
 * a circular buffer does not. This distinction doesn't matter to nisync so we
 * always just assume the caller has checked if the buffer is empty.
 */
#define nisync_buf_read_block_to_user(_buf_, _type_, _read_buffer_,	\
				_read_buffer_size_) ({			\
	int err = 0;							\
	typeof(_type_) *src_buf = (_type_ *)(_buf_).buf;		\
	typeof(_type_) *dst_buf = (_type_ *)(_read_buffer_);		\
	unsigned long elements_to_copy = (_read_buffer_size_) / sizeof(_type_);\
	unsigned long elements_to_end = nisync_buf_capacity(_buf_)	\
					- (_buf_).rd_idx;		\
	unsigned long elements_in_cb = nisync_buf_size(_buf_);		\
	nisync_buf_dbg_display("nisync_buf_read_block_to_user <=", _buf_);\
									\
	if (elements_to_copy > elements_in_cb)				\
		elements_to_copy = elements_in_cb;			\
	_read_buffer_size_ = elements_to_copy * sizeof(_type_);		\
									\
	if (elements_to_copy > elements_to_end) {			\
		err = copy_to_user(dst_buf, src_buf + (_buf_).rd_idx,	\
					elements_to_end * sizeof(_type_));\
		if (!err) {						\
			_nisync_buf_advance_idx(_buf_, rd_idx,		\
						elements_to_end);	\
			elements_to_copy -= elements_to_end;		\
			dst_buf += elements_to_end;			\
			atomic_sub(elements_to_end, &(_buf_).size);	\
		} else {						\
			pr_debug("nisync_buf_read_block_to_user: error %d in copy_to_user\n",\
				 err);					\
			err = -EFAULT;					\
		}							\
	}								\
	if (!err) {							\
		err = copy_to_user(dst_buf, src_buf + (_buf_).rd_idx,	\
					elements_to_copy * sizeof(_type_));\
		if (!err) {						\
			_nisync_buf_advance_idx(_buf_, rd_idx,		\
						elements_to_copy);	\
			atomic_sub(elements_to_copy, &(_buf_).size);	\
		} else {						\
			pr_debug("nisync_buf_read_block_to_user: error %d in copy_to_user\n",\
				 err);					\
			err = -EFAULT;					\
		}							\
	}								\
	nisync_buf_dbg_display("nisync_buf_read_block_to_user =>", _buf_);\
	pr_debug("nisync_buf_read_block_to_user => err: %d\n", err);	\
	err;								\
})

#endif /* NISYNC_BUF_H_ */

