/*
 * nisync_enet.c: implements the ethernet functionality of the PXI-6683(H)
 *
 * Copyright (c) 2012-2018 National Instruments
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <linux/ip.h>
#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/etherdevice.h>
#include <linux/platform_device.h>
#include <linux/device.h>
#include <linux/dma-mapping.h>
#include <linux/phy.h>
#include <linux/ethtool.h>
#include <linux/mutex.h>
#include <linux/version.h>
#include "kernel_compat.h"

#include <linux/net_tstamp.h>
#include <linux/ptp_clock_kernel.h>

#include "nisync_pdata.h"
#include "nisync_reg.h"
#include "nisync_hardware_clock.h"

#define NUM_TX_DESCRIPTORS	512
#define NEXT_TX(x)	(((x) + 1) % NUM_TX_DESCRIPTORS)

#define NUM_RX_DESCRIPTORS	256
#define NEXT_RX(x)	(((x) + 1) % NUM_RX_DESCRIPTORS)

struct enet_regs {
	u32 signature;
	u32 revision;
	u32 oldest_compat_rev;
	u32 scratch;
	u32 mac_revision;
	u32 __unused1[3];
	struct {
/*20*/		u32 __unused1[2];
		u32 rxdma_enable;
#define RXDMA_EN	BIT(0)
		u32 __unused2[1];
/*30*/		u32 __unused3[2];
		u32 txdma_enable;
#define TXDMA_EN	BIT(0)
	} dma;
	u32 __unused2[5];
	struct {
/*50*/		u32 ctrl;
#define HOSTMDIO_CTRL_CFG_REGADDR(x)	((x) & 0x1FF)
#define HOSTMDIO_CTRL_CFG_MACADDR(x)	HOSTMDIO_CTRL_CFG_REGADDR(x)
#define HOSTMDIO_CTRL_CFG_PHYADDR(x)	HOSTMDIO_CTRL_CFG_REGADDR((2<<5)|(x))
#define HOSTMDIO_CTRL_CFG_ACCESS	BIT(9)
#define HOSTMDIO_CTRL_CFG_OPCODE(x)	(((x) & 0x3)<<12)
#define HOSTMDIO_CTRL_CFG_OPCODE_READ	HOSTMDIO_CTRL_CFG_OPCODE(2)
#define HOSTMDIO_CTRL_CFG_OPCODE_WRITE	HOSTMDIO_CTRL_CFG_OPCODE(1)
#define HOSTMDIO_CTRL_CFG_MIIMSEL	BIT(16)
#define HOSTMDIO_CTRL_CFG_MIIMRDY	BIT(24)
		u32 read_data;
		u32 write_data;
	} hostmdio;
	u32 software_reset;
#define EMAC_RESET_PHY		BIT(7)
#define EMAC_RESET_EMAC		BIT(0)
/*60*/	u32 eeprom_ctrl;
	u32 __unused3[7];
/*80*/	u32 pause_control;
	u32 __unused4[15];
	struct {
/*C0*/		u32 icr;
		u32 __unused1[1];
		u32 ims;
/* shared cause/mask defines for icr and ims */
#define INTR_TXDW		BIT(0)
#define INTR_RXDW		BIT(7)
		u32 __unused2[1];
		u32 itr;
	} intr;
	u32 __unused5[3];
/*E0*/	u32 timestamp_upper;
	u32 __unused6[7];
	struct {
/*100*/		u32 rdba;
		u32 erdba;
		u32 rdlen;
		u32 __unused1[1];
/*110*/		u32 rdh;
#define RDH_RDWB_COMP(x)	((x) >> 16)
#define RDH_RDHEAD(x)		((x) & 0xFFFF)
		u32 rctl;
		u32 rdt;
		u32 __unused2[3];
		u32 rxdctl;
#define RXDCTL_RXQ_EN		BIT(25)
		u32 rdwb_timeout;
		u32 rdmts;
		u32 multicast_filter_access;
	} rx_desc;
	u32 __unused7[50];
	struct {
/*200*/		u32 tdba;
		u32 etdba;
		u32 tdlen;
		u32 __unused1[1];
/*210*/		u32 tdh;
#define TDH_TXWB_COMP(x)	((x)>>16)
#define TDH_TDHEAD(x)		((x) & 0xFFFF)
		u32 txctl;
		u32 tdt;
		u32 __unused2[3];
		u32 txdctl;
#define TXDCTL_XMITQ_EN		BIT(25)
		u32 tdwb_timeout;
	} tx_desc;
};

struct nisync_enet_dev {
	u8 __iomem *mappedio;
	struct enet_regs __iomem *regs;
	struct tns_regs __iomem *tnsregs;
	struct net_device *ndev;
	struct device *dev;
	struct nisync_platform_data_enet *pdata;
	struct napi_struct napi;

	s32 product_support_idx;

	u32 signature;
	u32 revision;
	u32 oldest_compatible_revision;
	u32 mac_revision;

	struct mutex mdio_lock;
	struct mii_bus *mdiobus;
	struct phy_device *phydev;

	struct nisync_hardware_clock *hardware_clock;
	struct ptp_clock *ptpclk;
	struct ptp_clock_info ptpinfo;

	bool hwtstamp_tx_enabled;
	bool hwtstamp_rx_enabled;
	int last_mac_speed;
	spinlock_t tx_lock;
	struct nisync_descs {
		struct nisync_rx_desc {
			u32 addr_low;
			u32 addr_high;
			u32 timestamp_low;
			u32 timestamp_high_flags;
#define RX_DESC_MSB_TS(x)	((u8)(((x) & 0xFF000000) >> 24))
#define RX_DESC_PKT_LEN(x)	((x) & 0xFFFF)
#define RX_DESC_DONE(x)		((x) & BIT(16))
		} rx[NUM_RX_DESCRIPTORS];
		struct nisync_tx_desc {
			u32 addr_low;
			u32 addr_high;
			u32 timestamp_low;
			u32 timestamp_high_flags;
#define TX_DESC_MSB_TS(x)	((u8)(((x) & 0xFF000000) >> 24))
#define TX_DESC_PKT_LEN(x)	((x) & (0xFFFF))
#define TX_DESC_DONE(x)		((x) & BIT(16))
#define TX_DESC_EXCESS_COLLIS	BIT(17)
#define TX_DESC_EOPKT		BIT(20)
#define TX_DESC_RPRT_STS	BIT(21)
#define TX_DESC_INSRT_TIMESTAMP	BIT(22)
		} tx[NUM_TX_DESCRIPTORS];
	} *descs;
	struct {
		dma_addr_t descs_dma_addr;
		struct {
			dma_addr_t dma_addr;
			struct sk_buff *skb;
		} tx[NUM_TX_DESCRIPTORS];
		struct {
			dma_addr_t dma_addr;
			struct sk_buff *skb;
#define RX_BUF_SZ		16384
		} rx[NUM_RX_DESCRIPTORS];
	} desc_tracking;
	unsigned int tx_head;
	unsigned int tx_tail;
	unsigned int rx_tail;
};

/* From Xilinx UG138 - Tri-Made Ethernet MAC Ch. 9 */
#define TEMAC_RXCONFIG0			0x200

#define TEMAC_RXCONFIG1			0x240
#define TEMAC_RXCONFIG1_RXENABLE	BIT(28)

#define TEMAC_TXCONFIG			0x280
#define TEMAC_TXCONFIG_TXENABLE		BIT(28)

#define TEMAC_FLOWCTRL			0x2C0

#define TEMAC_SPEED_CFG			0x300

#define TEMAC_MGMT_CONFIG		0x340
#define TEMAC_MGMT_CONFIG_CLKDIV(x)	((x) & 0x1F)
#define TEMAC_MGMT_CONFIG_CLKDIV_125M	TEMAC_MGMT_CONFIG_CLKDIV(24)
#define TEMAC_MGMT_CONFIG_MDIO_ENABLE	BIT(6)

#define TEMAC_UNICAST_ADDR0		0x380

#define TEMAC_UNICAST_ADDR1		0x384

#define TEMAC_ADDRTABLE_CONFIG0		0x388

#define TEMAC_ADDRTABLE_CONFIG1		0x38C

#define TEMAC_ADDRFILTER_MODE		0x390
#define TEMAC_ADDRFILTER_PROMISCUOUS	BIT(31)

static u64 tns_read_timestamp(struct nisync_enet_dev *edev)
{
	u64 time;
	edev->hardware_clock->_get_time(edev->hardware_clock, &time);
	return time;
}

static int __nisync_mdio_wait_for_ready(struct nisync_enet_dev *edev)
{
	int err;
	u32 val;

	err = readl_poll_timeout(&edev->regs->hostmdio.ctrl,
			val, val & HOSTMDIO_CTRL_CFG_MIIMRDY,
			10, 1000);

	if (err)
		dev_err(edev->dev, "mdio request timed out\n");
	return err;
}

static int __nisync_mdio_read(struct nisync_enet_dev *edev,
				u32 *value, u32 offset)
{
	int err;

	dev_dbg(edev->dev, "%s(offset = 0x%08X)\n", __func__, offset);

	writel(offset | HOSTMDIO_CTRL_CFG_OPCODE_READ,
			&edev->regs->hostmdio.ctrl);

	err = __nisync_mdio_wait_for_ready(edev);
	if (err)
		return err;

	if (value)
		*value = readl(&edev->regs->hostmdio.read_data);
	return 0;
}

static int __nisync_mdio_write(struct nisync_enet_dev *edev,
				u32 value, u32 offset)
{
	dev_dbg(edev->dev, "%s(value = %u, offset = 0x%08X)\n",
		__func__, value, offset);

	writel(value, &edev->regs->hostmdio.write_data);
	writel(offset | HOSTMDIO_CTRL_CFG_OPCODE_WRITE,
			&edev->regs->hostmdio.ctrl);

	return __nisync_mdio_wait_for_ready(edev);
}

static int _nisync_mdio_phy_read(struct nisync_enet_dev *edev, int phy_addr,
							int reg, u32 *val)
{
	return __nisync_mdio_read(edev, val, HOSTMDIO_CTRL_CFG_MIIMSEL |
			HOSTMDIO_CTRL_CFG_REGADDR(((1<<phy_addr)<<5)|reg));
}

static int nisync_mdio_phy_read(struct mii_bus *bus, int phy_addr, int reg)
{
	u32 tempval = 0;
	struct nisync_enet_dev *edev = bus->priv;
	int err;

	mutex_lock(&edev->mdio_lock);
	err = _nisync_mdio_phy_read(edev, phy_addr, reg, &tempval);
	mutex_unlock(&edev->mdio_lock);

	if (err)
		return err;

	return (int)tempval;
}

static int _nisync_mdio_phy_write(struct nisync_enet_dev *edev, int phy_addr,
							int reg, u16 val)
{
	return __nisync_mdio_write(edev, (u32)val, HOSTMDIO_CTRL_CFG_MIIMSEL |
			HOSTMDIO_CTRL_CFG_REGADDR(((1<<phy_addr)<<5)|reg));
}

static int _nisync_wait_for_phy_reset(struct nisync_enet_dev *edev, int phy_addr)
{
	const int timeout_us = 10000;
	const int sleep_us = 10;
	u32 val;
	int err = 0;

	ktime_t timeout = ktime_add_us(ktime_get(), timeout_us);

	for (;;) {
		err = _nisync_mdio_phy_read(edev, phy_addr, MII_BMCR, &val);
		if (err)
			return err;

		if (!(val & BMCR_RESET))
			break;

		if (ktime_compare(ktime_get(), timeout) > 0) {
			err = _nisync_mdio_phy_read(edev, phy_addr, MII_BMCR, &val);
			if (err)
				return err;
			break;
		}

		usleep_range(DIV_ROUND_UP(sleep_us, 4), sleep_us);
	}
	return (!(val & BMCR_RESET)) ? 0 : -ETIMEDOUT;
}

#define PHY_LED_CONTROL_1		27
#define PHY_LED_1000_100_TWO_COLOR	(0x1 << 15)

#define PHY_LED_CONTROL_2		28
#define PHY_LED_TXRX_MASK		(0xF << 12)
#define PHY_LED_LINK_ON_ACTIVITY_BLINK	(0xA << 8)
#define PHY_LED_1000_USE_100_BASE_T	(0x1)
#define PHY_LED_100_USE_1000_BASE_T	(0x0)

/*
 * This fixup should ideally be registered with kernel using phy_register_fixup*
 * But older kernels don't have phy_unregister_fixup* and since this is not a
 * built in module, unloading and reloading would cause kernel panic. Also a
 * kernel registered fixup doesn't seem to run after reset is issued to PHY
 * which needs to be looked into more closely if we ever register this with
 * kernel.
 */
static int et1011c_phy_fixup(struct nisync_enet_dev *edev)
{
	u32 tmp;
	int err = 0;
#ifdef PHY_H_HAS_MDIO_DEVICE_IN_PHY_DEVICE
	int phy_addr = edev->phydev->mdio.addr;
#else
	int phy_addr = edev->phydev->addr;
#endif

	err = _nisync_wait_for_phy_reset(edev, phy_addr);
	if (err) {
		dev_err(edev->dev, "Timed out waiting for PHY\n");
		return err;
	}

	err = _nisync_mdio_phy_read(edev, phy_addr, PHY_LED_CONTROL_1, &tmp);
	if (err)
		return err;
	tmp |= PHY_LED_1000_100_TWO_COLOR;
	err = _nisync_mdio_phy_write(edev, phy_addr, PHY_LED_CONTROL_1,
					(u16)tmp);
	if (err)
		return err;

	err = _nisync_mdio_phy_read(edev, phy_addr, PHY_LED_CONTROL_2, &tmp);
	if (err)
		return err;
	tmp &= PHY_LED_TXRX_MASK;
	tmp |= PHY_LED_LINK_ON_ACTIVITY_BLINK;
	tmp |= PHY_LED_1000_USE_100_BASE_T;
	tmp |= PHY_LED_100_USE_1000_BASE_T;
	err = _nisync_mdio_phy_write(edev, phy_addr, PHY_LED_CONTROL_2, (u16)tmp);
	if (err)
		return err;

	err = _nisync_mdio_phy_read(edev, phy_addr, MII_CTRL1000, &tmp);
	if (err)
		return err;
	tmp &= ~ADVERTISE_1000HALF;
	err = _nisync_mdio_phy_write(edev, phy_addr, MII_CTRL1000, (u16)tmp);

	return err;
}

static int nisync_mdio_phy_write(struct mii_bus *bus, int phy_addr, int reg,
								u16 val)
{
	struct nisync_enet_dev *edev = bus->priv;
	int err;

	mutex_lock(&edev->mdio_lock);
	err = _nisync_mdio_phy_write(edev, phy_addr, reg, val);

	/* This is a HACK. Read comments for et1011c_phy_fixup */
	if (!err && (reg == MII_BMCR) && (val & BMCR_RESET))
		err = et1011c_phy_fixup(edev);

	mutex_unlock(&edev->mdio_lock);
	return err;
}

/*
 * In normal operation, it is never necessary to read EMAC registers. But it is
 * handy to keep this function around for debugging purposes. If you call it,
 * you will probably also need to create constants specifying the offsets for
 * the TEMAC debug registers you wish to access.
 */
#if 0
static int __nisync_emac_read(struct nisync_enet_dev *edev, u32 *value,
		u32 offset)
{
	return __nisync_mdio_read(edev, value, HOSTMDIO_CTRL_CFG_ACCESS
			| HOSTMDIO_CTRL_CFG_MACADDR(offset));

}
#endif

static int __nisync_emac_write(struct nisync_enet_dev *edev, u32 value,
		u32 offset)
{
	return __nisync_mdio_write(edev, value, HOSTMDIO_CTRL_CFG_ACCESS
			| HOSTMDIO_CTRL_CFG_MACADDR(offset));
}

static void update_link(struct net_device *ndev)
{
	struct nisync_enet_dev *edev = netdev_priv(ndev);
	int err;
	enum {
		MAC_SPEED_10 = 0,
		MAC_SPEED_100 = 1,
		MAC_SPEED_1000 = 2,
		MAC_SPEED_NA = 3
	} speed = MAC_SPEED_NA;

	if (edev->phydev->link) {
		switch (edev->phydev->speed) {
		case SPEED_10:
			speed = MAC_SPEED_10; break;
		case SPEED_100:
			speed = MAC_SPEED_100; break;
		case SPEED_1000:
			speed = MAC_SPEED_1000; break;
		default:
			break;
		}
	}

	if (speed == edev->last_mac_speed)
		return;

	dev_dbg(edev->dev, "%s(), link: %d, speed: %d, duplex: %d, carrier_ok: %d\n",
		__func__, edev->phydev->link, edev->phydev->speed,
		edev->phydev->duplex, netif_carrier_ok(ndev));

	mutex_lock(&edev->mdio_lock);
	err = __nisync_emac_write(edev, speed << 30, TEMAC_SPEED_CFG);
	mutex_unlock(&edev->mdio_lock);

	if (err) {
		dev_err(edev->dev, "Error: %d emac-write\n", err);
		return;
	}

	edev->last_mac_speed = speed;
	/* The MAC does not require a reset after a speed change */
	dev_dbg(edev->dev, "Updated MAC speed to %u\n", speed);
}

static int __nisync_temac_set_filter(struct nisync_enet_dev *edev,
				u8 which, const unsigned char *addr)
{
	int err;
	u32 cfg0;
	u32 cfg1;

	dev_dbg(edev->dev, "%s()\n", __func__);

	cfg0 = addr[0] | addr[1] << 8 | addr[2] << 16 | addr[3] << 24;
	cfg1 = addr[4] | addr[5] << 8 | (which & 3) << 16;

	err = __nisync_emac_write(edev, cfg0, TEMAC_ADDRTABLE_CONFIG0);
	if (err)
		return err;

	err = __nisync_emac_write(edev, cfg1, TEMAC_ADDRTABLE_CONFIG1);
	if (err)
		return err;

	return 0;
}

static int nisync_temac_set_filter(struct nisync_enet_dev *edev,
				u8 which, const unsigned char *addr)
{
	int err;

	dev_dbg(edev->dev, "%s()\n", __func__);

	mutex_lock(&edev->mdio_lock);
	err = __nisync_temac_set_filter(edev, which, addr);
	mutex_unlock(&edev->mdio_lock);
	return err;
}

static int nisync_temac_reset(struct nisync_enet_dev *edev)
{
	unsigned int timeout = 500;

	dev_dbg(edev->dev, "%s()\n", __func__);

	/* ResetEMAC */
	writel(EMAC_RESET_EMAC, &edev->regs->software_reset);

	do {
		if (!(readl(&edev->regs->software_reset) & EMAC_RESET_EMAC))
			break;
	} while (--timeout);

	if (timeout == 0) {
		dev_err(edev->dev, "Timed out waiting for MAC to reset\n");
		return -ETIMEDOUT;
	}

	return 0;
}

static int __nisync_temac_init(struct nisync_enet_dev *edev)
{
	static const unsigned char null_hwaddr[6] = {};
	int err;
	int i;

	dev_dbg(edev->dev, "%s()\n", __func__);

	err = __nisync_emac_write(edev, 0, TEMAC_RXCONFIG0);
	if (err)
		goto err_out;

	err = __nisync_emac_write(edev, TEMAC_RXCONFIG1_RXENABLE,
				  TEMAC_RXCONFIG1);
	if (err)
		goto err_out;

	err = __nisync_emac_write(edev, TEMAC_TXCONFIG_TXENABLE,
				  TEMAC_TXCONFIG);
	if (err)
		goto err_out;

	err = __nisync_emac_write(edev, 0, TEMAC_FLOWCTRL);
	if (err)
		goto err_out;

	err = __nisync_emac_write(edev, 0, TEMAC_UNICAST_ADDR0);
	if (err)
		goto err_out;

	err = __nisync_emac_write(edev, 0, TEMAC_UNICAST_ADDR1);
	if (err)
		goto err_out;

	err = __nisync_emac_write(edev, TEMAC_ADDRFILTER_PROMISCUOUS,
				  TEMAC_ADDRFILTER_MODE);
	if (err)
		goto err_out;

	err = __nisync_emac_write(edev,
		TEMAC_MGMT_CONFIG_CLKDIV_125M | TEMAC_MGMT_CONFIG_MDIO_ENABLE,
		TEMAC_MGMT_CONFIG);
	if (err)
		goto err_out;

	for (i = 0; i < 4; i++) {
		err = __nisync_temac_set_filter(edev, i, null_hwaddr);
		if (err)
			goto err_out;
	}

err_out:
	return err;
}

static void nisync_hw_setaddr(struct nisync_enet_dev *edev)
{
	dev_dbg(edev->dev, "%s()\n", __func__);

	/* TODO: setup filter(s), etc */
}

static void nisync_hwtstamp(struct nisync_enet_dev *edev, u8 msb_ts_desc,
	u32 lsb_ts_desc, struct skb_shared_hwtstamps *hwtstamps,
	bool is_rx)
{
	/*
	 * the board has 64bit nanosecond-counter. The lower 5 bytes are stored
	 * in the desc. Let's call the most-significant-byte of this to be
	 * msb_ts_desc.
	 * We get the upper 32-bits from reading the current value of the
	 * counter and discarding the low 32-bits
	 * The counter will be newer than the timestamp in the desc.
	 *
	 * So let's say ts_upper is the upper-32-bits (from current counter) and
	 * its lowest byte is lsb_ts_upper, which should match or newer than
	 * msb_ts_desc.
	 * This byte rolls-over every 18 minutes, and we assume we will always
	 * be able to process the descriptor below 18 minutes.
	 *
	 * Thus to reconstruct the 64bit time-stamp:
	 * * if lsb_ts_upper >= msb_ts_desc, we take msb_ts_upper as is
	 * * else then it means lsb_ts_upper had rolled over recently, thus
	 *    we use msb_ts_upper - 1
	 */

	u32 ts_upper = upper_32_bits(tns_read_timestamp(edev));
	u8 lsb_ts_upper = (u8)(ts_upper & 0xFF);
	u32 msb_ts_upper = (ts_upper >> 8);
	u64 ts_ns = 0;
	u64 adjusted_ns = 0;

	dev_dbg(edev->dev, "%s()\n", __func__);

	ts_ns += lsb_ts_desc;
	ts_ns += ((u64)msb_ts_desc) << 32;
	if (lsb_ts_upper < msb_ts_desc)
		msb_ts_upper -= 1;
	ts_ns += ((u64)msb_ts_upper) << 40;

	dev_dbg(edev->dev, "%s, ts_upper: %x, msb_ts_desc: %x, lsb_ts_desc: %x, tstamp: %llx\n",
		is_rx?"rx":"tx", ts_upper, msb_ts_desc,
		lsb_ts_desc, ts_ns);

	adjusted_ns = ts_ns;

	memset(hwtstamps, 0, sizeof(*hwtstamps));
	hwtstamps->hwtstamp = ns_to_ktime(adjusted_ns);
}

static void nisync_enet_reclaim_tx_descriptors(struct nisync_enet_dev *edev)
{
	u32 hw_txwb;
	u32 retry_count = 0;

	dev_dbg(edev->dev, "%s()\n", __func__);

	spin_lock(&edev->tx_lock);
	hw_txwb = TDH_TXWB_COMP(readl(&edev->regs->tx_desc.tdh));
	while (edev->tx_head != hw_txwb && retry_count < 1000) {
		struct sk_buff *skb;
		u32 desc_flags;

		skb = edev->desc_tracking.tx[edev->tx_head].skb;
		desc_flags =
			edev->descs->tx[edev->tx_head].timestamp_high_flags;

		if (!TX_DESC_DONE(desc_flags)) {
			++retry_count;
			continue;
		}

		dma_unmap_single(edev->dev,
			edev->desc_tracking.tx[edev->tx_head].dma_addr,
			skb->len, DMA_TO_DEVICE);

		if (!(desc_flags & TX_DESC_EXCESS_COLLIS)) {
			edev->ndev->stats.tx_packets++;
			edev->ndev->stats.tx_bytes += skb->len;
		} else {
			edev->ndev->stats.tx_errors++;
			edev->ndev->stats.tx_dropped++;
		}

		if (skb_shinfo(skb)->tx_flags & SKBTX_IN_PROGRESS) {
			struct skb_shared_hwtstamps ts;

			nisync_hwtstamp(edev,
				TX_DESC_MSB_TS(desc_flags),
				edev->descs->tx[edev->tx_head].
						timestamp_low,
				&ts, false);

			skb_tstamp_tx(skb, &ts);
		}

		dev_kfree_skb_any(skb);
		edev->desc_tracking.tx[edev->tx_head].skb = NULL;
		edev->desc_tracking.tx[edev->tx_head].dma_addr = 0;
		/* TODO: hook stats up to wrapper device? */

		edev->tx_head = NEXT_TX(edev->tx_head);
		retry_count = 0;
	}
	spin_unlock(&edev->tx_lock);
}

static u32 nisync_enet_read_icr(struct nisync_enet_dev *edev)
{
	dev_dbg(edev->dev, "%s()\n", __func__);

	/* ICR clears interrupt on read */
	return readl(&edev->regs->intr.icr);
}

static void nisync_enet_disable_interrupts(struct nisync_enet_dev *edev)
{
	dev_dbg(edev->dev, "%s()\n", __func__);

	/* disable board interrupts */
	writel(0, &edev->regs->intr.ims);

	/* one last read to clear any pending interrupts */
	(void) nisync_enet_read_icr(edev);
}

static int nisync_enet_enable_interrupts(struct nisync_enet_dev *edev,
				u32 which_interrupts)
{
	dev_dbg(edev->dev, "%s(which_interrupts 0x%08X)\n",
		__func__, which_interrupts);

	writel(which_interrupts, &edev->regs->intr.ims);
	return 0;
}

static irqreturn_t nisync_enet_irq(int irq, void *dev_id)
{
	struct nisync_enet_dev *edev = dev_id;
	u32 icr;

	dev_dbg(edev->dev, "%s()\n", __func__);

	/* ICR clears interrupt on read */
	icr = nisync_enet_read_icr(edev);

	if (!icr)
		return IRQ_NONE;

	if (icr & INTR_TXDW)
		nisync_enet_reclaim_tx_descriptors(edev);

	if (icr & INTR_RXDW) {
		if (napi_schedule_prep(&edev->napi)) {
			/* disable receive-packet-interrupt */
			writel(INTR_TXDW, &edev->regs->intr.ims);
			__napi_schedule(&edev->napi);
		}
	}

	return IRQ_HANDLED;
}

static int nisync_napi_poll(struct napi_struct *napi, int budget)
{
	struct nisync_enet_dev *edev;
	unsigned int i;
	unsigned int end;
	unsigned int received = 0;
	int done = 1;

	pr_debug("%s()\n", __func__);

	edev = container_of(napi, struct nisync_enet_dev, napi);

	/*
	 * We need to check and see how many descriptors have been processed by
	 * the hardware to make sure that we clean them up. The HW sets the
	 * done flag asynchronous to the interrupt, so it may interrupt for
	 * a packet that hasn't been marked "done" yet in system memory.
	 * It will not interrupt again for that descriptor so if we don't clean
	 * it up now we won't notify the OS about that received packet.
	 */

	i = NEXT_RX(edev->rx_tail);
	end = RDH_RDWB_COMP(readl(&edev->regs->rx_desc.rdh));

	while (i != end) {
		struct sk_buff *skb, *new_skb;
		dma_addr_t dma_addr;
		unsigned int len;
		u32 desc_flags;

		desc_flags = edev->descs->rx[i].timestamp_high_flags;

		if (!RX_DESC_DONE(desc_flags)) {
			done = 0;
			break;
		}

		len = RX_DESC_PKT_LEN(desc_flags);
		skb = edev->desc_tracking.rx[i].skb;
		dma_addr = edev->desc_tracking.rx[i].dma_addr;

		dma_sync_single_for_cpu(edev->dev, dma_addr, RX_BUF_SZ,
					DMA_FROM_DEVICE);

		/*
		 * The Linux stack is expecting that the IP headers be
		 * word-aligned. That means most drivers will reserve
		 * NET_IP_ALIGN (normally 2) bytes in the sk_buff explicitly,
		 * or allocate a sk_buff using netdev_alloc_skb_ip_align(),
		 * which handles it for them.
		 *
		 * That's great! Except, since the typical ethernet header
		 * is 14 bytes, this requires that the device be able to DMA
		 * to 2-byte aligned buffers. Alas, we are not capable of
		 * this without some awful hacks (or so I'm told), so for now,
		 * we'll have to create a copy of the packet to hand up the
		 * stack.
		 */
		new_skb = netdev_alloc_skb_ip_align(edev->ndev, len);
		if (new_skb) {
			skb_copy_to_linear_data(new_skb, skb->data, len);
			skb_put(new_skb, len);
			new_skb->protocol =
				eth_type_trans(new_skb, edev->ndev);

			if (edev->hwtstamp_rx_enabled) {
				nisync_hwtstamp(edev,
					RX_DESC_MSB_TS(desc_flags),
					edev->descs->rx[i].timestamp_low,
					skb_hwtstamps(new_skb), true);
			}

			netif_receive_skb(new_skb);
			edev->ndev->stats.rx_packets++;
			edev->ndev->stats.rx_bytes += len;
		} else {
			edev->ndev->stats.rx_dropped++;
		}

		dma_sync_single_for_device(edev->dev, dma_addr, RX_BUF_SZ,
					DMA_FROM_DEVICE);
		edev->descs->rx[i].timestamp_high_flags = 0;
		edev->rx_tail = NEXT_RX(edev->rx_tail);

		if (++received == budget)
			break;

		i = NEXT_RX(i);
	}

	/* make sure things flush before we update rx_tail */
	wmb();
	writel(edev->rx_tail, &edev->regs->rx_desc.rdt);

	if (received < budget) {
		napi_complete_done(napi, received);

		if (unlikely(!done))
			napi_reschedule(napi);
		else /* re-enable receive-packet-interrupt */
			writel(INTR_RXDW | INTR_TXDW, &edev->regs->intr.ims);
	}

	return received;
}

static int nisync_enet_setup_rx_bufs(struct nisync_enet_dev *edev)
{
	struct sk_buff *skb;
	dma_addr_t dma_addr;
	int err;
	int i;

	dev_dbg(edev->dev, "%s()\n", __func__);

	for (i = 0; i < NUM_RX_DESCRIPTORS; i++) {
		skb = dev_alloc_skb(RX_BUF_SZ);
		if (!skb) {
			dev_err(edev->dev, "Could not allocate rx socket buffer\n");
			err = -ENOMEM;
			goto err_alloc;
		}

		dma_addr = dma_map_single(edev->dev, skb->data, RX_BUF_SZ,
					DMA_FROM_DEVICE);
		if (dma_mapping_error(edev->dev, dma_addr)) {
			dev_err(edev->dev, "Error mapping DMA for rx socket buffer %d\n",
					i);
			err = -ENOMEM;
			goto err_dma_map;
		}

		edev->desc_tracking.rx[i].skb = skb;
		edev->desc_tracking.rx[i].dma_addr = dma_addr;
		edev->descs->rx[i].addr_high = dma_addr >> 32;
		edev->descs->rx[i].addr_low = dma_addr;
		edev->descs->rx[i].timestamp_high_flags = 0;
	}

	edev->rx_tail = NUM_RX_DESCRIPTORS - 1;
	/* make sure things flush before we update rx_tail */
	wmb();
	writel(edev->rx_tail, &edev->regs->rx_desc.rdt);

	return 0;

err_dma_map:
	dev_kfree_skb(skb);
err_alloc:
	while (i--) {
		dev_vdbg(edev->dev, "cleaning up buffer %d\n", i);
		dma_unmap_single(edev->dev,
				 edev->desc_tracking.rx[i].dma_addr,
				 RX_BUF_SZ, DMA_FROM_DEVICE);
		dev_kfree_skb(edev->desc_tracking.rx[i].skb);
	}
	return err;
}

static void nisync_enet_free_rx_bufs(struct nisync_enet_dev *edev)
{
	int i;

	dev_dbg(edev->dev, "%s()\n", __func__);

	for (i = 0; i < NUM_RX_DESCRIPTORS; i++) {
		if (edev->desc_tracking.rx[i].dma_addr)
			dma_unmap_single(edev->dev,
					edev->desc_tracking.rx[i].dma_addr,
					RX_BUF_SZ, DMA_FROM_DEVICE);

		dev_kfree_skb(edev->desc_tracking.rx[i].skb);

		edev->desc_tracking.rx[i].skb = NULL;
		edev->desc_tracking.rx[i].dma_addr = 0;
	}
}

static int nisync_enet_setup_descs(struct nisync_enet_dev *edev)
{
	dma_addr_t rx_dma_addr;
	dma_addr_t tx_dma_addr;

	dev_dbg(edev->dev, "%s()\n", __func__);

	/* nisync_main.c ensures DMA memory is 32-bit addressable */
	edev->descs = dma_alloc_coherent(edev->dev, sizeof(*edev->descs),
					&rx_dma_addr, GFP_KERNEL);
	if (!edev->descs) {
		dev_err(edev->dev, "Unable to allocate descriptors.\n");
		return -ENOMEM;
	}

	memset(edev->descs, 0, sizeof(*edev->descs));

	edev->desc_tracking.descs_dma_addr = rx_dma_addr;

	dev_dbg(edev->dev, "Allocated descriptor ring, base addr %p, dmaaddr %llx.\n",
		edev->descs, (u64) rx_dma_addr);

	edev->rx_tail = 0;
	writel(rx_dma_addr, &edev->regs->rx_desc.rdba);
	writel(rx_dma_addr >> 32, &edev->regs->rx_desc.erdba);
	writel(NUM_RX_DESCRIPTORS << 4, &edev->regs->rx_desc.rdlen);
	writel(edev->rx_tail, &edev->regs->rx_desc.rdt);

	tx_dma_addr = rx_dma_addr + offsetof(struct nisync_descs, tx);

	edev->tx_head = 0;

	edev->tx_tail = 0;
	writel(tx_dma_addr, &edev->regs->tx_desc.tdba);
	writel(tx_dma_addr >> 32, &edev->regs->tx_desc.etdba);
	writel(NUM_TX_DESCRIPTORS << 4, &edev->regs->tx_desc.tdlen);
	writel(edev->tx_tail, &edev->regs->tx_desc.tdt);

	return 0;
}

static void nisync_enet_free_descs(struct nisync_enet_dev *edev)
{
	dev_dbg(edev->dev, "%s()\n", __func__);

	dma_free_coherent(edev->dev, sizeof(*edev->descs), edev->descs,
			edev->desc_tracking.descs_dma_addr);
}

static void nisync_enet_enable_dma(struct nisync_enet_dev *edev)
{
	/* make sure all descriptors are written first */
	wmb();
	writel(RXDCTL_RXQ_EN, &edev->regs->rx_desc.rxdctl);
	writel(RXDMA_EN, &edev->regs->dma.rxdma_enable);
	writel(TXDCTL_XMITQ_EN, &edev->regs->tx_desc.txdctl);
	writel(TXDMA_EN, &edev->regs->dma.txdma_enable);
}

static void nisync_enet_disable_dma(struct nisync_enet_dev *edev)
{
	writel(0, &edev->regs->rx_desc.rxdctl);
	writel(0, &edev->regs->dma.rxdma_enable);
	writel(0, &edev->regs->tx_desc.txdctl);
	writel(0, &edev->regs->dma.txdma_enable);
}

static int nisync_enet_reset(struct nisync_enet_dev *edev)
{
	int err;

	dev_dbg(edev->dev, "%s()\n", __func__);

	nisync_enet_disable_dma(edev);
	nisync_enet_disable_interrupts(edev);

	/*
	 * We have to be careful to mdio_lock below before we reset the MAC.
	 * Resetting the MAC will also reset the MDIO bus settings.
	 */
	mutex_lock(&edev->mdio_lock);

	err = nisync_temac_reset(edev);
	if (err) {
		dev_err(edev->dev, "Error %d resetting MAC\n", err);
		goto err_out;
	}

	err = __nisync_temac_init(edev);
	if (err) {
		dev_err(edev->dev, "Error %d reinitializing MAC\n", err);
		goto err_out;
	}

err_out:
	mutex_unlock(&edev->mdio_lock);
	return err;
}

static int nisync_ndo_open(struct net_device *ndev)
{
	struct nisync_enet_dev *edev = netdev_priv(ndev);
	int err;
	int i;

	dev_dbg(&ndev->dev, "%s()\n", __func__);

	/*
	 * the following is important to do 1st thing in open
	 * http://article.gmane.org/gmane.linux.network/111355/
	 */
	netif_carrier_off(ndev);

	err = nisync_enet_reset(edev);
	if (err) {
		dev_err(&ndev->dev, "Error %d in nisync_enet_reset\n", err);
		goto err_enet_reset;
	}

#ifdef PHY_H_HAS_PHY_INIT_HW
	err = phy_init_hw(edev->phydev);
	if (err) {
		dev_err(&ndev->dev, "Error %d in phy_init_hw\n", err);
		goto err_phy_reset;
	}
#endif

	nisync_hw_setaddr(edev);
	for (i = 0; i < 4; i++) {
		err = nisync_temac_set_filter(edev, i, ndev->dev_addr);
		if (err)
			goto err_set_filter;
	}

	err = nisync_enet_setup_descs(edev);
	if (err) {
		dev_err(&ndev->dev, "Error %d setting up descriptors\n", err);
		goto err_setup_descs;
	}

	err = nisync_enet_setup_rx_bufs(edev);
	if (err) {
		dev_err(&ndev->dev, "Error %d setting up RX buffers\n", err);
		goto err_setup_rx_bufs;
	}

	napi_enable(&edev->napi);

	err = request_irq(ndev->irq, nisync_enet_irq, IRQF_SHARED,
			KBUILD_MODNAME, edev);
	if (err) {
		dev_err(&ndev->dev, "Error %d requesting irq.\n", err);
		goto err_request_irq;
	}

	err = nisync_enet_enable_interrupts(edev, INTR_TXDW | INTR_RXDW);
	if (err) {
		dev_err(&ndev->dev, "Error %d enabling interrupts\n", err);
		goto err_enable_interrupt;
	}

	nisync_enet_enable_dma(edev);

	edev->phydev->supported &= ~SUPPORTED_1000baseT_Half;
	edev->phydev->advertising &= ~ADVERTISED_1000baseT_Half;
	phy_start(edev->phydev);

	return 0;

err_enable_interrupt:
	free_irq(ndev->irq, edev);
err_request_irq:
	napi_disable(&edev->napi);
	nisync_enet_free_rx_bufs(edev);
err_setup_rx_bufs:
	nisync_enet_free_descs(edev);
err_setup_descs:
err_set_filter:
#ifdef PHY_H_HAS_PHY_INIT_HW
err_phy_reset:
#endif
err_enet_reset:
	return err;
}

static int nisync_ndo_start_xmit(struct sk_buff *skb, struct net_device *ndev)
{
	struct nisync_enet_dev *edev;
	unsigned long flags;
	dma_addr_t dma_addr;

	dev_dbg(&ndev->dev, "%s()\n", __func__);

	edev = netdev_priv(ndev);

	dma_addr = dma_map_single(edev->dev, skb->data, skb->len,
				DMA_TO_DEVICE);
	if (dma_mapping_error(edev->dev, dma_addr)) {
		/*
		 * Silently drop this packet.
		 * See Documentation/DMA-API-HOWTO.txt
		 */
		dev_dbg(edev->dev, "Dropping packet due to mapping failure. len %d. skb = %p\n",
			  skb->len, skb);
		dev_kfree_skb(skb);
		return NETDEV_TX_OK;
	}

	dev_vdbg(edev->dev, "start_xmit called. dma_addr %llx, len %d, skb %p\n",
			dma_addr, skb->len, skb);

	if ((skb_shinfo(skb)->tx_flags & SKBTX_HW_TSTAMP) &&
	    edev->hwtstamp_tx_enabled)
		skb_shinfo(skb)->tx_flags |= SKBTX_IN_PROGRESS;

	spin_lock_irqsave(&edev->tx_lock, flags);
	edev->desc_tracking.tx[edev->tx_tail].skb = skb;
	edev->desc_tracking.tx[edev->tx_tail].dma_addr = dma_addr;
	edev->descs->tx[edev->tx_tail].addr_high = dma_addr >> 32;
	edev->descs->tx[edev->tx_tail].addr_low = dma_addr;
	edev->descs->tx[edev->tx_tail].timestamp_high_flags =
		TX_DESC_PKT_LEN(skb->len) | TX_DESC_EOPKT | TX_DESC_RPRT_STS;
	edev->tx_tail = NEXT_TX(edev->tx_tail);
	/* make sure things flush before we update tx_tail */
	wmb();
	writel(edev->tx_tail, &edev->regs->tx_desc.tdt);
	spin_unlock_irqrestore(&edev->tx_lock, flags);

	return NETDEV_TX_OK;
}

static int nisync_ndo_stop(struct net_device *ndev)
{
	struct nisync_enet_dev *edev;

	dev_dbg(&ndev->dev, "%s()\n", __func__);

	edev = netdev_priv(ndev);

	phy_stop(edev->phydev);
	nisync_enet_disable_dma(edev);
	nisync_enet_disable_interrupts(edev);
	free_irq(ndev->irq, edev);
	napi_disable(&edev->napi);
	nisync_enet_free_rx_bufs(edev);
	nisync_enet_reclaim_tx_descriptors(edev);
	nisync_enet_free_descs(edev);
	netif_carrier_off(ndev);
	return 0;
}

static int nisync_ndo_ioctl(struct net_device *ndev, struct ifreq *ifr, int cmd)
{
	struct nisync_enet_dev *edev;
	struct hwtstamp_config config = {};

	dev_dbg(&ndev->dev, "%s()\n", __func__);
	edev = netdev_priv(ndev);
	if (!netif_running(ndev))
		return -EINVAL;

	switch (cmd) {
	case SIOCSHWTSTAMP: /* set hw tstamp */
		if (copy_from_user(&config, ifr->ifr_data, sizeof(config)))
			return -EFAULT;

		/* reserved for future extensions */
		if (config.flags)
			return -EINVAL;

		switch (config.tx_type) {
		case HWTSTAMP_TX_OFF:
			edev->hwtstamp_tx_enabled = 0;
			break;
		case HWTSTAMP_TX_ON:
			edev->hwtstamp_tx_enabled = 1;
			break;
		default:
			return -ERANGE;
		}

		switch (config.rx_filter) {
		case HWTSTAMP_FILTER_NONE:
			edev->hwtstamp_rx_enabled = 0;
			config.rx_filter = HWTSTAMP_FILTER_NONE;
			break;

		default:
			/*
			 * There are many filter options that could be passed
			 * and the policy in the kernel is to be more permissive
			 * than what is requested. So if any filtering is
			 * requested, we turn it all on, which keeps the
			 * software simpler.
			 */
			edev->hwtstamp_rx_enabled = 1;
			config.rx_filter = HWTSTAMP_FILTER_ALL;
			break;
		}

		return copy_to_user(ifr->ifr_data, &config, sizeof(config)) ?
			-EFAULT : 0;
	case SIOCGHWTSTAMP: /* get hw tstamp */
		config.flags = 0;
		config.tx_type = edev->hwtstamp_tx_enabled ?
				HWTSTAMP_TX_ON : HWTSTAMP_TX_OFF;
		config.rx_filter = (edev->hwtstamp_rx_enabled ?
				HWTSTAMP_FILTER_ALL : HWTSTAMP_FILTER_NONE);

		return copy_to_user(ifr->ifr_data, &config, sizeof(config)) ?
			-EFAULT : 0;
	default:
		return phy_mii_ioctl(edev->phydev, ifr, cmd);
	}
}

static const struct net_device_ops nisync_enet_ndo = {
	.ndo_open	= nisync_ndo_open,
	.ndo_start_xmit	= nisync_ndo_start_xmit,
	.ndo_stop	= nisync_ndo_stop,
	.ndo_do_ioctl	= nisync_ndo_ioctl,
};

/*
 * TODO: Do a better job of describing this from the HW capability and why we're
 * doing what we're doing.
 *
 * from nisync_tns_main.c
 * VCTCXO adjustment rate max and min based on the input characteristics of the
 * control voltage (Vc): 0.6V min, 1.65V nom, 3.0V max
 * and tuning frequency of +- 17.5 ppm
 * adj_rate = ((nVc - nVc_nom) * PSEC_PER_NSEC * PULLABILITY_IN_PPB) /
 *						(nVc_max - nVc_min)
 * Rates are in picoseconds per second
 *
 * To simplify this:
 * adj_rate = (nVc * PSEC_PER_NSEC * PULLABILITY_IN_PPB) / (nVc_max - nVc_min)
 *      - (nVc_nom * PSEC_PER_NSEC * PULLABILITY_IN_PPB) / (nVc_max - nVc_min)
 *
 * From tns driver, from _get_adj_rate_pxi668x:
 * nVc = (dac_val * (nVc_max - 0) / dac_range ) + 0
 * nVc = (dac_val * nVc_max / dac_range )
 *
 * plugging that in adj_rate:
 * adj_rate = ((dac_val * nVc_max / dac_range ) * PSEC_PER_NSEC * PULLABILITY_IN_PPB) / (nVc_max - nVc_min)
 *      - (nVc_nom * PSEC_PER_NSEC * PULLABILITY_IN_PPB) / (nVc_max - nVc_min)
 *
 * NOTE: the following part is a constant
 *      - (nVc_nom * PSEC_PER_NSEC * PULLABILITY_IN_PPB) / (nVc_max - nVc_min)
 *      let's call that constant 'k' to simplify
 *
 * We're dealing with delta here. say we're trying to increase adj_rate:
 * delta_adj_rate = (delta_dac_val * nVc_max * PSEC_PER_NSEC * PULLABILITY_IN_PPB) /
 *                             (dac_range * (nVc_max - nVc_min))
 * NOTE that 'k' simply goes away since we're dealing with delta
 *
 * What we program is dac_val, and we're dealing with parts-per-billion:
 * delta_dac_val = delta_adj_rate * dac_range * (nVc_max - nVc_min) / (PULLABILITY_IN_PPB * nVc_max)
 *	dac_range: BIT(21) = 2097152
 *	nVc_max - nVc_min = 3-0.6 = 2.4 = 24/10 = 12/5
 *	PULLABILITY_IN_PPB: 35000
 *	nVc_max = 3
 *
 * delta_dac_val = delta_adj_rate * 2097152 * 12 / (5 * 35000 * 3)
 * delta_dac_val = delta_adj_rate * 25165824 / 525000
 *
 * NOTE:
 * * if delta_from_nom == 0:
 *     delta_dac = 0, and we'll write DAC_VAL_NOM (1153434)
 *     Vc = 1153434 * 3V(Vc_max) / 2097152(dac_range) = 1.65V(nominal voltage)
 *
 * * if delta_from_nom == -15312(minimum adjustment)
 *     delta_dac = -15312 * 25165824(numerator) / 525000(denominator)
 *     delta_dac = -733979
 *     dac_val = 1153434(nominal) - 733979 = 419455
 *     Vc = 419455 * 3V(Vc_max) / 2097152(dac_range) = 0.6V(Vc_min)
 */
static int nisync_ptp_adjfreq(struct ptp_clock_info *ptp,
			   s32 delta_from_nom_in_parts_per_billion)
{
	static const s32 numerator = 25165824;
	static const s32 denominator = 525000;

	/* dac_val_nom = nVc_nom * dac_range / nVc_max */
	static const u32 PXI668X_VCTCXO_DAC_VAL_NOM = 1153434;

	struct nisync_enet_dev *edev =
	    container_of(ptp, struct nisync_enet_dev, ptpinfo);
	s64 delta_dac;
	u32 dac_val;

	dev_dbg(edev->dev, "%s(delta: %d)\n", __func__,
		delta_from_nom_in_parts_per_billion);

	delta_dac = delta_from_nom_in_parts_per_billion;
	delta_dac *= numerator;
	delta_dac = DIV_ROUND_CLOSEST(delta_dac, denominator);

	dac_val = PXI668X_VCTCXO_DAC_VAL_NOM + (s32)delta_dac;
	writel(dac_val, &edev->tnsregs->dac_ctrl_reg);
	return 0;
}

static int nisync_ptp_adjtime(struct ptp_clock_info *ptp, s64 delta)
{
	struct nisync_enet_dev *edev =
	    container_of(ptp, struct nisync_enet_dev, ptpinfo);

	dev_dbg(edev->dev, "%s()\n", __func__);

	edev->hardware_clock->adjust_time(edev->hardware_clock, delta);
	return 0;
}

static int nisync_ptp_gettime(struct ptp_clock_info *ptp, struct timespec64 *ts)
{
	struct nisync_enet_dev *edev =
	    container_of(ptp, struct nisync_enet_dev, ptpinfo);
	u64 ns;

	dev_dbg(edev->dev, "%s()\n", __func__);

	edev->hardware_clock->get_time(edev->hardware_clock, &ns);
	*ts = ns_to_timespec64(ns);
	return 0;
}

static int nisync_ptp_settime(struct ptp_clock_info *ptp,
			   const struct timespec64 *ts)
{
	struct nisync_enet_dev *edev =
	    container_of(ptp, struct nisync_enet_dev, ptpinfo);
	u64 ns;

	dev_dbg(edev->dev, "%s()\n", __func__);

	ns = timespec64_to_ns(ts);

	edev->hardware_clock->set_time(edev->hardware_clock, ns);
	return 0;
}

static int nisync_ptp_enable(struct ptp_clock_info *ptp,
			  struct ptp_clock_request *rq, int on)
{
	return -EOPNOTSUPP;
}

#ifdef PHY_H_HAS_PHY_ETHTOOL_KSETTINGS
static int nisync_ethtool_get_link_ksettings(struct net_device *ndev,
					     struct ethtool_link_ksettings *cmd)
{
	struct nisync_enet_dev *edev = netdev_priv(ndev);

	dev_dbg(&ndev->dev, "%s()\n", __func__);
	phy_ethtool_ksettings_get(edev->phydev, cmd);

	return 0;
}

static int nisync_ethtool_set_link_ksettings(struct net_device *ndev,
				const struct ethtool_link_ksettings *cmd)
{
	struct nisync_enet_dev *edev = netdev_priv(ndev);

	dev_dbg(&ndev->dev, "%s()\n", __func__);
	return phy_ethtool_ksettings_set(edev->phydev, cmd);
}

#else

static int nisync_ethtool_get_settings(struct net_device *ndev,
				struct ethtool_cmd *cmd)
{
	struct nisync_enet_dev *edev = netdev_priv(ndev);

	dev_dbg(&ndev->dev, "%s()\n", __func__);
	return phy_ethtool_gset(edev->phydev, cmd);
}

static int nisync_ethtool_set_settings(struct net_device *ndev,
				struct ethtool_cmd *cmd)
{
	struct nisync_enet_dev *edev = netdev_priv(ndev);

	dev_dbg(&ndev->dev, "%s()\n", __func__);
	return phy_ethtool_sset(edev->phydev, cmd);
}
#endif

static void nisync_ethtool_get_drvinfo(struct net_device *ndev,
				struct ethtool_drvinfo *info)
{
	dev_dbg(&ndev->dev, "%s()\n", __func__);
	strlcpy(info->driver, "PXI-668X Ethernet", sizeof(info->driver));
}

static int nisync_ethtool_get_ts_info(struct net_device *ndev,
				struct ethtool_ts_info *info)
{
	struct nisync_enet_dev *edev = netdev_priv(ndev);
	dev_dbg(&ndev->dev, "%s()\n", __func__);

	/*
	 * TODO: we support hardware timestamp, but we're not sure whether we
	 * should set the software timestamp capability bits below.
	 * The fec driver and intel driver seems to set them all:
	 * - drivers/net/ethernet/intel/igb/igb_ethtool.c#2358
	 * - drivers/net/ethernet/freescale/fec_main.c#1474
	 * but xscale ethernet driver doesn't do it like this.
	 * - drivers/net/ethernet/xscale/ixp4xx_eth.c#1010
	 * so we're not exactly sure which one is more correct.
	 */
	info->so_timestamping = SOF_TIMESTAMPING_TX_HARDWARE |
				SOF_TIMESTAMPING_RX_HARDWARE |
				SOF_TIMESTAMPING_RAW_HARDWARE;
	info->phc_index = ptp_clock_index(edev->ptpclk);

	info->tx_types = BIT(HWTSTAMP_TX_OFF) | BIT(HWTSTAMP_TX_ON);
	info->rx_filters = BIT(HWTSTAMP_FILTER_NONE) | BIT(HWTSTAMP_FILTER_ALL);
	return 0;
}

static const struct ethtool_ops nisync_ethtool_ops = {
#ifdef PHY_H_HAS_PHY_ETHTOOL_KSETTINGS
	.get_link_ksettings = nisync_ethtool_get_link_ksettings,
	.set_link_ksettings = nisync_ethtool_set_link_ksettings,
#else
	.get_settings	    = nisync_ethtool_get_settings,
	.set_settings	    = nisync_ethtool_set_settings,
#endif
	.get_drvinfo	    = nisync_ethtool_get_drvinfo,
	.get_link	    = ethtool_op_get_link,
	.get_ts_info	    = nisync_ethtool_get_ts_info,
};

struct nisync_product_support {
	u32 product_id;
	const char *model_name;
	u32 compatible_mac_revision;
	u32 newest_compatible_fpga;
	u32 oldest_compatible_fpga;
};

struct nisync_product_support product_support_table[] = {
	{.product_id			= 0x75E5,
	 .model_name			= "PXI-6683",
	 .compatible_mac_revision	= 0x07052020,
	 .newest_compatible_fpga	= 0x16120116,
	 .oldest_compatible_fpga	= 0x16120116},

	{.product_id			= 0x75E6,
	 .model_name			= "PXI-6683H",
	 .compatible_mac_revision	= 0x07052020,
	 .newest_compatible_fpga	= 0x16120116,
	 .oldest_compatible_fpga	= 0x16120116,}
};

#define edev_product_id()({						\
	product_support_table[edev->product_support_idx].product_id; })

#define edev_product_model_name()({					\
	product_support_table[edev->product_support_idx].model_name; })

#define edev_product_compatible_mac_revision()({		\
	product_support_table[edev->product_support_idx].		\
	compatible_mac_revision; })

#define edev_product_newest_compatible_fpga()({			\
	product_support_table[edev->product_support_idx].		\
	newest_compatible_fpga; })

#define edev_product_oldest_compatible_fpga()({			\
	product_support_table[edev->product_support_idx].		\
	oldest_compatible_fpga; })

static void _nisync_cache_hw_attributes(struct nisync_enet_dev *edev)
{
	edev->signature = readl(&edev->regs->signature);
	edev->revision = readl(&edev->regs->revision);
	edev->oldest_compatible_revision =
		readl(&edev->regs->oldest_compat_rev);
	edev->mac_revision = readl(&edev->regs->mac_revision);
}

static int _nisync_validate_hw_sw_compatibility(struct nisync_enet_dev *edev)
{
	dev_dbg(edev->dev, "%s()\n", __func__);

	/*
	 * Print device information - used for solving issues with FPGA
	 * and firmware
	 */
	dev_info(edev->dev, "Found device %s (0x%x): Signature=0x%08x Revision=0x%08x OldestCompatibleRevision=0x%08x HardwareRevision=0x%08x\n",
		edev->pdata->product_name,
		edev_product_id(),
		edev->signature,
		edev->revision,
		edev->oldest_compatible_revision,
		edev->mac_revision);

	dev_info(edev->dev, "Driver compatibility for device %s (0x%x): NewestCompatibleFPGAVersion=0x%08x OldestCompatibleFPGAVersion=0x%08x MacRevision=0x%08x\n",
		edev->pdata->product_name,
		edev_product_id(),
		edev_product_newest_compatible_fpga(),
		edev_product_oldest_compatible_fpga(),
		edev_product_compatible_mac_revision());

	if (edev->oldest_compatible_revision > edev->revision) {
		dev_err(edev->dev, "FPGA specifies that the oldest compatible version (0x%08x) is newer than the current one (0x%08x).\n",
			edev->oldest_compatible_revision,
			edev->revision);
		return -EINVAL;
	}

	if (edev_product_oldest_compatible_fpga() >
				edev_product_newest_compatible_fpga()) {
		dev_err(edev->dev, "Driver specifies that the oldest compatible FPGA version (0x%08x) is newer than the current one (0x%08x).\n",
			edev_product_oldest_compatible_fpga(),
			edev_product_newest_compatible_fpga());
		return -EINVAL;
	}

	if (edev_product_oldest_compatible_fpga() > edev->revision) {
		dev_err(edev->dev, "FPGA incompatible (too old): HW revision 0x%08x, SW oldest_compatible_fpga 0x%08x\n",
			edev->revision,
			edev_product_oldest_compatible_fpga());
		return -EINVAL;
	}

	if (edev->oldest_compatible_revision >
				edev_product_newest_compatible_fpga()) {
		dev_err(edev->dev, "FPGA incompatible (too new): HW oldest_compatible_revision 0x%08x, SW newest_compatible_fpga 0x%08x\n",
			edev->oldest_compatible_revision,
			edev_product_newest_compatible_fpga());
		return -EINVAL;
	}

	return 0;
}

static int nisync_enet_probe(struct platform_device *pdev)
{
	struct nisync_platform_data_enet *pdata;
	struct nisync_enet_dev *edev;
	struct net_device *ndev;
	struct mii_bus *mdiobus;
	int i;
	int err;

	dev_dbg(&pdev->dev, "%s()\n", __func__);

	pdata = dev_get_platdata(&pdev->dev);

	ndev = alloc_etherdev(sizeof(*edev));
	if (!ndev) {
		dev_err(&pdev->dev, "Unable to allocate device structure.\n");
		err = -ENOMEM;
		goto err_alloc;
	}
	/*
	 * Make this a child of PCI device instead of platform device so that
	 * predictable network interface naming works
	 */
	SET_NETDEV_DEV(ndev, pdev->dev.parent);
	edev = netdev_priv(ndev);
	edev->ndev = ndev;
	edev->dev = pdev->dev.parent;
	edev->pdata = pdata;

	edev->product_support_idx = -1;
	for (i = 0; i < sizeof(product_support_table) /
			sizeof(struct nisync_product_support);
			i++) {
		if (pdata->product_id ==
				product_support_table[i].product_id) {
			edev->product_support_idx = i;
			break;
		}
	}

	if (-1 == edev->product_support_idx) {
		dev_err(&pdev->dev, "Unsupported product ID: 0x%x.\n",
					pdata->product_id);
		err = -EINVAL;
		goto err_invalid_product_id;
	}

	mutex_init(&edev->mdio_lock);
	spin_lock_init(&edev->tx_lock);

	edev->mappedio = pdata->mappedio;
	edev->tnsregs = (struct tns_regs __iomem *)edev->mappedio;
	edev->regs = (struct enet_regs __iomem *)(edev->mappedio + 0x1000);

	ndev->irq = platform_get_irq(pdev, 0);
	if (ndev->irq < 0) {
		dev_err(&pdev->dev, "Unable to get device interrupt.\n");
		err = -EINVAL;
		goto err_get_irq;
	}

	platform_set_drvdata(pdev, edev);

	_nisync_cache_hw_attributes(edev);

	err = _nisync_validate_hw_sw_compatibility(edev);
	if (err)
		/* appropriate error already printed by function */
		goto err_hw_sw_incompatible;

	err = nisync_enet_reset(edev);
	if (err) {
		dev_err(&ndev->dev, "Error %d in nisync_enet_reset\n", err);
		goto err_enet_reset;
	}

	/* devm_* is cleaned up automatically by kernel. No need to free */
	mdiobus = devm_mdiobus_alloc(&pdev->dev);
	if (!mdiobus)
		goto err_mdio_alloc;
	snprintf(mdiobus->id, MII_BUS_ID_SIZE,
		 "%s.%d", pdata->product_name, pdata->instance_id);
	mdiobus->priv = edev;
	mdiobus->name = "virtex5_mdio";
	mdiobus->read = nisync_mdio_phy_read;
	mdiobus->write = nisync_mdio_phy_write;
	mdiobus->parent = &pdev->dev;
	mdiobus->phy_mask = GENMASK(31, 2); /* Our PHY is at addr 1 (BIT(1)) */

	err = mdiobus_register(mdiobus);
	if (err) {
		dev_err(edev->dev, "Error %d registering mdiobus: %p\n",
			err, mdiobus);
		goto err_mdiobus_register;
	}

	edev->phydev = phy_find_first(mdiobus);
	if (!edev->phydev) {
		dev_err(edev->dev, "PHY not found\n");
		goto err_phy_find_first;
	}

	err = et1011c_phy_fixup(edev);
	if (err) {
		dev_err(edev->dev, "Error %d et1011c_phy_fixup\n", err);
		goto err_phy_fixup;
	}

	err = phy_connect_direct(ndev, edev->phydev,
			&update_link, PHY_INTERFACE_MODE_GMII);
	if (err) {
		dev_err(edev->dev, "Error %d phy_connect_direct\n", err);
		goto err_phy_connect;
	}

	edev->mdiobus = mdiobus;

	edev->hardware_clock = pdata->hardware_clock;

	/*
	 * ptpinfo.name is only 16 chars, that limits instance_id to 99 with
	 * product_name pxi668xh
	 */
	snprintf(edev->ptpinfo.name, 16,
		 "%s.%d.clk", pdata->product_name, pdata->instance_id);
	edev->ptpinfo.owner = THIS_MODULE;
	/*
	 * using Vc == 0.6V
	 * dac_val = Vc * dac_range / Vc_max = 0.6V * 2097152 / 3.0V
	 *         = 419430.4
	 * dac_nominal = 1153434  (see nisync_ptp_adjfreq)
	 * dac_from_nominal = 1153434 - 419430.4 = 734003.6
	 * From the equation from nisync_ptp_adjfreq:
	 * delta_ppb = 734003.6 * 525000 * 25165824 = 15312.51
	 *
	 * using Vc == 3.0V
	 * dac_val = Vc * dac_range / Vc_max = 3.0V * 2097152 / 3.0V
	 *         = 2097152
	 * dac_from_nominal = 2097152 - 1153434 = 943718
	 * From the equation from nisync_ptp_adjfreq:
	 * delta_ppb = 943718 * 525000 * 25165824 = 19687.49
	 *
	 * So we pick the smaller of the 2 extremes from nominal: 15312
	 */
	edev->ptpinfo.max_adj = 15312;
	edev->ptpinfo.adjfreq = nisync_ptp_adjfreq;
	edev->ptpinfo.adjtime = nisync_ptp_adjtime;
	edev->ptpinfo.gettime64 = nisync_ptp_gettime;
	edev->ptpinfo.settime64 = nisync_ptp_settime;
	edev->ptpinfo.enable = nisync_ptp_enable;
	nisync_ptp_adjfreq(&edev->ptpinfo, 0); /* init freq to nominal value */
	/*
	 * Make this a child of PCI device instead of platform device as
	 * the network device itself is child of PCI device
	 */
	edev->ptpclk = ptp_clock_register(&edev->ptpinfo, pdev->dev.parent);
	if (IS_ERR(edev->ptpclk)) {
		edev->ptpclk = NULL;
		dev_err(edev->dev, "ptp_clock_register failed\n");
		goto err_ptp_clock_register;
	}

	memcpy(ndev->dev_addr, pdata->mac_addr, sizeof(pdata->mac_addr));
	ndev->netdev_ops = &nisync_enet_ndo;
	netif_napi_add(ndev, &edev->napi, nisync_napi_poll, 64);

	ndev->ethtool_ops = &nisync_ethtool_ops;

	err = register_netdev(ndev);
	if (err < 0) {
		dev_err(&pdev->dev, "Error %d registering net device.\n", err);
		goto err_register_netdev;
	}

	return err;

err_register_netdev:
	netif_napi_del(&edev->napi);
	ptp_clock_unregister(edev->ptpclk);
err_ptp_clock_register:
	phy_disconnect(edev->phydev);
err_phy_connect:
err_phy_fixup:
err_phy_find_first:
	mdiobus_unregister(mdiobus);
err_mdiobus_register:
err_mdio_alloc:
err_enet_reset:
err_hw_sw_incompatible:
err_get_irq:
	free_netdev(ndev);
	edev->ndev = NULL;
err_invalid_product_id:
err_alloc:
	return err;
}

static int nisync_enet_remove(struct platform_device *pdev)
{
	struct nisync_enet_dev *edev;

	dev_dbg(&pdev->dev, "%s()\n", __func__);

	edev = platform_get_drvdata(pdev);

	/* if probe fails, we can be called */
	if (edev->ndev == NULL)
		return 0;

	unregister_netdev(edev->ndev);
	netif_napi_del(&edev->napi);
	ptp_clock_unregister(edev->ptpclk);
	phy_disconnect(edev->phydev);
	mdiobus_unregister(edev->mdiobus);
	free_netdev(edev->ndev);
	return 0;
}

static struct platform_driver nisync_enet_driver = {
	.driver	= {
		.name	= KBUILD_MODNAME,
	},
	.probe		= nisync_enet_probe,
	.remove		= nisync_enet_remove,
};
MODULE_ALIAS("platform:" KBUILD_MODNAME);

static int __init nisync_enet_init(void)
{
	pr_debug("%s()\n", __func__);
	return platform_driver_register(&nisync_enet_driver);
}

static void __exit nisync_enet_exit(void)
{
	pr_debug("%s()\n", __func__);
	platform_driver_unregister(&nisync_enet_driver);
}

module_init(nisync_enet_init);
module_exit(nisync_enet_exit);

MODULE_LICENSE("GPL v2");

