#ifndef NISYNC_HARDWARE_CLOCK_H_
#define NISYNC_HARDWARE_CLOCK_H_

#include <linux/mutex.h>
#include <linux/rwsem.h>
#include <linux/spinlock.h>

#include "nisync_reg.h"

struct nisync_hardware_clock {
	struct tns_regs __iomem *tnsregs;

	/* Prevent concurrent calls to the device */
	struct mutex dev_lock;
	/* Prevent device reset with down_read */
	struct rw_semaphore dev_reset_lock;
	/* Prevent concurrent access to timecounter */
	raw_spinlock_t timestamp_counter_lock;

	u64 last_phase_adjust_offset;
	s32 adjustment_rate_in_ps;

	/* Device constants */
	u32 nominal_frequency;
	u32 nominal_clock_period_in_ns;
	u32 time_snap_boundary_in_ns;

	u32 vco_max_adj_rate;
	u32 vco_min_adj_rate;

	s32 vco_pullability_in_ppb;
	s64 vco_max_nanovolts;
	s64 vco_min_nanovolts;
	s64 vco_nom_nanovolts;

	s64 dac_min_nanovolts;
	s64 dac_max_nanovolts;
	u32 dac_max_value;

	u32 system_time_latency;

	/*
	 * _get_time is equivalent to get_time but uses no locks other
	 * than timestamp_counter_lock. Other functions with names starting
	 * with _ are equivalent to non _ functions but don't use any locks.
	 * The _ functions are for use in IRQ context or when locks are already
	 * held
	 */
	void (*_get_time)(struct nisync_hardware_clock *clk,
					u64 *time);
	int (*get_time)(struct nisync_hardware_clock *clk, u64 *time);

	int (*_set_time)(struct nisync_hardware_clock *clk, u64 time);
	int (*set_time)(struct nisync_hardware_clock *clk, u64 time);

	int (*_reset_time)(struct nisync_hardware_clock *clk);

	int (*_adjust_time)(struct nisync_hardware_clock *clk, s64 offset);
	int (*adjust_time)(struct nisync_hardware_clock *clk, s64 offset);

	int (*get_adj_rate)(struct nisync_hardware_clock *clk,
					s32 *rate_in_ps_per_sec);

	int (*_set_adj_rate)(struct nisync_hardware_clock *clk,
					const s32 rate_in_ps_per_sec);
	int (*set_adj_rate)(struct nisync_hardware_clock *clk,
					const s32 rate_in_ps_per_sec);

	int (*get_frequency)(struct nisync_hardware_clock *clk, u32 *frequency);
	int (*set_frequency)(struct nisync_hardware_clock *clk, u32 frequency);
};

#endif /* NISYNC_HARDWARE_CLOCK_H_ */
