#include <linux/timer.h>
#include "nisync_reg.h"
#include "nisync_hardware_clock.h"

#define hardware_clock_product_regmap(_reg_)({				\
	void __iomem *register_address;					\
	register_address = (void __iomem *)				\
		&clk->tnsregs->_reg_;					\
	register_address; })

#define create_u64(_upper_u32_, _lower_u32_)				\
	((((u64)(_upper_u32_)) << 32) | ((u64)(_lower_u32_)))

#define start_timer(_timer_, _nanoseconds_) ({				\
	getrawmonotonic(&(_timer_));					\
	timespec_add_ns(&(_timer_), ((s64)_nanoseconds_)); })

#define is_timer_expired(_timeout_) ({					\
	bool timer_expired = false;					\
	struct timespec curr_time;					\
									\
	getrawmonotonic(&curr_time);					\
	if (timespec_compare(&curr_time, &(_timeout_)) > 0)		\
		timer_expired = true;					\
	timer_expired; })

#define nisync_spin_event_timeout(_condition_, _timeout_in_ns_) ({	\
	bool timeout = true;						\
	struct timespec timer;						\
	start_timer(timer, (_timeout_in_ns_));				\
	do {								\
		if (_condition_) {					\
			timeout = false;				\
			break;						\
		}							\
	} while (!is_timer_expired(timer));				\
	timeout; })

#define round_to_nearest(_value_, _interval_)				\
       ((_interval_) == 0 ? (_value_) :					\
       ((_value_) < 0 ?							\
       (((_value_) - (_interval_) / (s64)2) / (_interval_) * (_interval_)) : \
       (((_value_) + (_interval_) / (s64)2) / (_interval_) * (_interval_))))

/* DIV_ROUND_CLOSEST that works for negative values of x */
#define DIV_ROUND_CLOSEST_GENERIC(x, divisor)({				\
	((x) < 0)?							\
		(-DIV_ROUND_CLOSEST((-(x)), (divisor))) :		\
		(DIV_ROUND_CLOSEST((x), (divisor)));			\
})

#define PSEC_PER_NSEC 1000
#define PSEC_PER_SEC ((u64)NSEC_PER_SEC * PSEC_PER_NSEC)

void _nisync_hardware_clock_get_time(struct nisync_hardware_clock *clk,
					u64 *time)
{
	u32 high_word;
	u32 low_word;
	unsigned long flags;

	raw_spin_lock_irqsave(&clk->timestamp_counter_lock, flags);

	low_word = readl(hardware_clock_product_regmap(system_time_low_word));
	high_word = readl(hardware_clock_product_regmap(system_time_high_word));

	raw_spin_unlock_irqrestore(&clk->timestamp_counter_lock, flags);

	*time = create_u64(high_word, low_word);
	*time += clk->system_time_latency;
}

int nisync_hardware_clock_get_time(struct nisync_hardware_clock *clk,
					u64 *time)
{
	down_read(&clk->dev_reset_lock);
	mutex_lock(&clk->dev_lock);

	_nisync_hardware_clock_get_time(clk, time);

	mutex_unlock(&clk->dev_lock);
	up_read(&clk->dev_reset_lock);

	return 0;
}

#define SYSTEM_TIME_RESET_ENABLE_BIT	BIT(24)
int _nisync_hardware_clock_reset_time(struct nisync_hardware_clock *clk)
{
	bool timeout;

	/* Reset hardware's system time and phase adjustment */
	writel(SYSTEM_TIME_RESET_ENABLE_BIT,
		hardware_clock_product_regmap(servo_control_enable));

	/* Wait for reset to complete */
	timeout = nisync_spin_event_timeout(
		!(readl(hardware_clock_product_regmap(servo_control_enable)) &
			SYSTEM_TIME_RESET_ENABLE_BIT),
		NSEC_PER_SEC);
	if (timeout) {
		pr_err("timeout waiting for servo_control_enable SystemTimeReset register\n");
		return -ETIMEDOUT;
	}

	clk->last_phase_adjust_offset = 0;

	return 0;
}

int _nisync_hardware_clock_adjust_time(struct nisync_hardware_clock *clk,
					s64 offset)
{
	/*
	 * There is no need to reset time counter to go back in time.
	 *
	 * We need to go back in time when (last_phase_adjust_offset + offset)
	 * is negative, but in unsigned arithmetic, this will calculate to a
	 * large positive value. Using this value for phase adjustment will
	 * overflow the time counter and wrap time back to the correct lower
	 * value.
	 *
	 * This also preserves phase relationship between ClkOut and TK
	 */
	clk->last_phase_adjust_offset =
		(u64)round_to_nearest(clk->last_phase_adjust_offset + offset,
					clk->nominal_clock_period_in_ns);

	/*
	 * HW will adjust curr_time by the amount servo_phase_adjust changes
	 * from last write
	 */
	writel(upper_32_bits(clk->last_phase_adjust_offset),
		hardware_clock_product_regmap(servo_phase_adjust_high_word));
	writel(lower_32_bits(clk->last_phase_adjust_offset),
		hardware_clock_product_regmap(servo_phase_adjust_low_word));

	return 0;
}

int _nisync_hardware_clock_set_time(struct nisync_hardware_clock *clk,
					u64 time)
{
	u64 curr_time;
	s64 offset;

	_nisync_hardware_clock_get_time(clk, &curr_time);

	offset = time - curr_time;
	offset = round_to_nearest(offset, clk->time_snap_boundary_in_ns);

	return _nisync_hardware_clock_adjust_time(clk, offset);
}

int nisync_hardware_clock_set_time(struct nisync_hardware_clock *clk,
					u64 time)
{
	int err;

	down_read(&clk->dev_reset_lock);
	mutex_lock(&clk->dev_lock);

	err = _nisync_hardware_clock_set_time(clk, time);

	mutex_unlock(&clk->dev_lock);
	up_read(&clk->dev_reset_lock);

	return err;
}

int nisync_hardware_clock_adjust_time(struct nisync_hardware_clock *clk,
					s64 offset)
{
	int err;

	if (offset == 0)
		return 0;

	offset = round_to_nearest(offset, clk->time_snap_boundary_in_ns);

	down_read(&clk->dev_reset_lock);
	mutex_lock(&clk->dev_lock);

	err = _nisync_hardware_clock_adjust_time(clk, offset);

	mutex_unlock(&clk->dev_lock);
	up_read(&clk->dev_reset_lock);

	return err;
}

static int __set_dac_val_pxi668x(struct nisync_hardware_clock *clk, u32 dac_val)
{
	if (dac_val > clk->dac_max_value)
		return -EINVAL;

	writel(dac_val, hardware_clock_product_regmap(dac_ctrl_reg));

	return 0;
}

static void __get_dac_val_pxi668x(struct nisync_hardware_clock *clk,
				u32 *dac_val)
{
	*dac_val = readl(hardware_clock_product_regmap(dac_ctrl_reg));
}

static int _get_adj_rate_pxi668x(struct nisync_hardware_clock *clk, s32 *rate)
{
	u32 dac_val;
	s64 voltage;
	s64 tmp_rate;

	__get_dac_val_pxi668x(clk, &dac_val);

	/* convert dac_val back to a voltage */
	voltage = dac_val *
		(clk->dac_max_nanovolts - clk->dac_min_nanovolts);
	voltage = DIV_ROUND_CLOSEST(voltage, clk->dac_max_value);
	voltage += clk->dac_min_nanovolts;

	if (voltage < clk->dac_min_nanovolts ||
					voltage > clk->dac_max_nanovolts) {
		pr_err("voltage %lld out of range; min: %lld, max: %lld\n",
				voltage, clk->dac_min_nanovolts,
				clk->dac_max_nanovolts);
		return -EIO;
	}

	/* convert voltage back to a rate in picosecondsPerSecond */
	tmp_rate = voltage - clk->vco_nom_nanovolts;
	tmp_rate *= PSEC_PER_NSEC;
	tmp_rate *= clk->vco_pullability_in_ppb;

	/*
	 * DIV_ROUND_CLOSEST does not always round a negative value to the
	 * correct number so use DIV_ROUND_CLOSEST_GENERIC.
	 */
	*rate = DIV_ROUND_CLOSEST_GENERIC(tmp_rate,
					     clk->vco_max_nanovolts -
					     clk->vco_min_nanovolts);

	return 0;
}

int nisync_hardware_clock_get_adj_rate(struct nisync_hardware_clock *clk,
					s32 *rate_in_ps_per_sec)
{
	*rate_in_ps_per_sec = clk->adjustment_rate_in_ps;
	return 0;
}

int _nisync_hardware_clock_set_adj_rate(struct nisync_hardware_clock *clk,
					s32 rate)
{
	int err = 0;
	s64 voltage;
	u64 dac_val;
	s32 new_rate;

	/*
	 * Calculate the nanovoltage that will induce the requested rate
	 * adjustment
	 */
	voltage = rate; /* rate is in picosecondsPerSecond */
	voltage *= clk->vco_max_nanovolts -
				clk->vco_min_nanovolts;
	voltage = DIV_ROUND_CLOSEST(voltage,
				clk->vco_pullability_in_ppb);
	voltage = DIV_ROUND_CLOSEST(voltage, PSEC_PER_NSEC);
	voltage += clk->vco_nom_nanovolts;

	if (voltage < clk->vco_min_nanovolts ||
			voltage > clk->vco_max_nanovolts) {
		return -EINVAL;
	}

	/* Calculate the DAC value that will output the calculated voltage */
	dac_val = voltage - clk->dac_min_nanovolts;
	dac_val *= clk->dac_max_value;
	dac_val = DIV_ROUND_CLOSEST(dac_val, clk->dac_max_nanovolts -
						clk->dac_min_nanovolts);

	/* Commit DAC value to hardware */
	err = __set_dac_val_pxi668x(clk, dac_val);
	if (err)
		return err;

	/*
	 * Get the new rate. Coercion can result in a different rate than what
	 * was requested due to rounding and scaling when calculating and
	 * committing the DAC value.
	 */
	err = _get_adj_rate_pxi668x(clk, &new_rate);
	if (err)
		return err;

	clk->adjustment_rate_in_ps = new_rate;

	return 0;

}

int nisync_hardware_clock_set_adj_rate(struct nisync_hardware_clock *clk,
					const s32 rate_in_ps_per_sec)
{
	int err;

	down_read(&clk->dev_reset_lock);
	mutex_lock(&clk->dev_lock);

	err = _nisync_hardware_clock_set_adj_rate(clk, rate_in_ps_per_sec);

	mutex_unlock(&clk->dev_lock);
	up_read(&clk->dev_reset_lock);

	return err;
}

static int _nisync_hardware_clock_get_frequency(struct nisync_hardware_clock *clk,
					u32 *frequency)
{
	/*
	 * DIV_ROUND_CLOSEST does not always round a negative value to the
	 * correct number so use DIV_ROUND_CLOSEST_GENERIC.
	 */
	*frequency = DIV_ROUND_CLOSEST_GENERIC((s64) clk->nominal_frequency *
					       (s64) clk->adjustment_rate_in_ps,
					       (s64) PSEC_PER_SEC);
	*frequency += clk->nominal_frequency;

	return 0;
}

int nisync_hardware_clock_get_frequency(struct nisync_hardware_clock *clk,
					u32 *frequency)
{
	int err;

	down_read(&clk->dev_reset_lock);
	mutex_lock(&clk->dev_lock);

	err = _nisync_hardware_clock_get_frequency(clk, frequency);

	mutex_unlock(&clk->dev_lock);
	up_read(&clk->dev_reset_lock);

	return err;
}

int nisync_hardware_clock_set_frequency(struct nisync_hardware_clock *clk,
					u32 frequency)
{
	int err;
	s64 rate;

	/* Convert frequency to a rate in picosecondsPerSecond */
	rate = (s64) frequency;
	rate -= (s64) clk->nominal_frequency;
	rate *= (s64) PSEC_PER_SEC;
	rate /= (s64) clk->nominal_frequency;

	if (INT_MAX < rate || INT_MIN > rate) {
		err = -EINVAL;
		goto err_rate_too_large;
	}

	down_read(&clk->dev_reset_lock);
	mutex_lock(&clk->dev_lock);

	err = _nisync_hardware_clock_set_adj_rate(clk, (s32)rate);

	mutex_unlock(&clk->dev_lock);
	up_read(&clk->dev_reset_lock);

err_rate_too_large:
	return err;
}
