/*
 * nisync.c: implements board startup and registers MFD cells
 *
 * Copyright (c) 2012-2018 National Instruments
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <linux/mfd/core.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/version.h>
#include <linux/device.h>
#include <linux/sysfs.h>
#include <linux/firmware.h>
#include <linux/delay.h>
#include <linux/cdev.h>
#include <linux/dma-mapping.h>
#include "kernel_compat.h"

#include <linux/mutex.h>

#include "nisync_pdata.h"
#include "nisync_kern.h"

#include "nisync_hardware_clock.h"
#include "nisync_hardware_clock_impl.h"

#define NISYNC_CONCAT_(x, y) x ## y
#define NISYNC_CONCAT(x, y) NISYNC_CONCAT_(x, y)

#define NISYNC_PADDING(b) u8 NISYNC_CONCAT(__reserved, __LINE__)[b]

struct nisync_regs {
/*000*/	NISYNC_PADDING(0x08);
/*008*/	u32 __reg1;
#define REGISTER1_OFFSET1	BIT(25)
/*00C*/	NISYNC_PADDING(0x04);
/*010*/	u32 __reg2;
#define REGISTER2_OFFSET1	BIT(31)
/*014*/	NISYNC_PADDING(0x0c);
/*020*/	u32 __reg3;
#define REGISTER3_OFFSET1	BIT(14)
/*024*/	NISYNC_PADDING(0x0c);
/*030*/ u32 __reg4;
/*034*/	NISYNC_PADDING(0x8c);
/*0C0*/	u32 __reg5;
#define REGISTER5_OFFSET1(x)	((x)&0xFFFFF000)
#define REGISTER5_OFFSET2	BIT(7)
/*0C4*/	u32 __reg6;
#define REGISTER6_OFFSET1(x)	((x)&0xFFFFFF00)
#define REGISTER6_OFFSET2	BIT(7)
#define REGISTER6_OFFSET3	(0xC)
/*0C8*/	NISYNC_PADDING(0x0c);
/*0D4*/	u32 __reg7;
/*0D8*/	NISYNC_PADDING(0x1c);
/*0F4*/	u32 __reg8;
#define REGISTER8_OFFSET1	BIT(0)
/*0F8*/	NISYNC_PADDING(0x3fc);
/*4F4*/	u32 __reg9;
#define REGISTER9_OFFSET1		BIT(19)
#define REGISTER9_OFFSET2		BIT(17)
#define REGISTER9_OFFSET3		BIT(3)
/*4F8*/	NISYNC_PADDING(0x8);
	struct __struct1 {
	/*000*/	u32 __reg10;
#define REGISTER10_OFFSET1	BIT(31)
#define REGISTER10_OFFSET2	BIT(4)
#define REGISTER10_OFFSET3		BIT(0)
	/*004*/	u32 __reg11;
#define REGISTER11_OFFSET1		BIT(14)
#define REGISTER11_OFFSET2	GENMASK(12, 11)
#define REGISTER11_OFFSET3	GENMASK(9, 8)
#define REGISTER11_OFFSET4	GENMASK(5, 4)
#define REGISTER11_OFFSET5	BIT(3)
#define REGISTER11_OFFSET6	GENMASK(1, 0)
	/*008*/	NISYNC_PADDING(0x4);
	/*00C*/	u32 __reg12;
#define REGISTER12_OFFSET1		GENMASK(23, 21)
#define REGISTER12_OFFSET3		BIT(10)
#define REGISTER12_OFFSET4	GENMASK(9, 8)
	/*010*/	NISYNC_PADDING(0x4);
	/*014*/	u32 __reg13;
#define REGISTER13_OFFSET2		GENMASK(23, 21)
#define REGISTER13_OFFSET3		BIT(10)
#define REGISTER13_OFFSET4	GENMASK(9, 8)
#define REGISTER13_OFFSET5		BIT(6)
	/*018*/	NISYNC_PADDING(0x4);
	/*01C*/	u32 __reg14;
#define REGISTER14_OFFSET1		GENMASK(23, 21)
#define REGISTER14_OFFSET2	GENMASK(18, 17)
#define REGISTER14_OFFSET3		BIT(10)
#define REGISTER14_OFFSET4	GENMASK(9, 8)
#define REGISTER14_OFFSET5		BIT(6)
#define REGISTER14_OFFSET6	BIT(0)
	/*020*/	NISYNC_PADDING(0xe0);
/*500*/	} __struct1_array[3];
};

struct nisync_device_data;

struct nisync_flash_sector_group {
	u32 sector_size;
	u32 sector_count;
};

struct nisync_flash_sector_iterator {
	const struct nisync_flash_sector_group *group;
	u32 offset;
	u32 sector;
	u32 sector_in_group;
};

struct nisync_device {
	atomic_t dev_counter;
	const char *name;
	const char *fw_path;

	u32 nominal_clock_period_in_ns;
	u32 time_snap_boundary_in_ns;
};

#define NUM_CELLS 2 /* tns and enet */

struct nisync_device_data {
	struct device *dev;
	struct pci_dev *pdev;
	struct mfd_cell cells[NUM_CELLS];
	bool mfd_devices_exist;
	int mfd_id;
	struct nisync_device *sync_dev;
	struct nisync_hardware_clock hardware_clock;
	struct nisync_platform_data_tns pdata_tns;
	struct nisync_platform_data_enet pdata_enet;
	struct {
		u32 signature;
		u32 revision;
		u32 oldest_compatible_revision;
		u32 hardware_revision;
	} firmware;
	struct mutex sysfs_firmware_lock;
	struct nisync_regs __iomem *asic;
	u8 __iomem *window;
};

static struct class *nisync_class;
static atomic_t total_devices = ATOMIC_INIT(0);

#define PCI_VENDOR_ID_NI	0x1093
#define PCI_DEV_ID_NI_PXI6683	0x75E5
#define PCI_DEV_ID_NI_PXI6683H	0x75E6

static u8 nisync_s29gl032n_read8(struct nisync_device_data *ddata, u32 offset);
static int pxi668x_flash_enable(struct nisync_device_data *ddata,
				const struct firmware *fw);
/*
 * Both the PXI-6683 and PXI-6683H use the S29GL032N flash (model 03) for
 * storing the devices FPGA image. Functions and data specific to the model of
 * the flash have _s29gl032n_03_ in their names. Otherwise, those generic to the
 * family have only _s29gl032n_ in their names.
 */
static void nisync_s29gl032n_open(struct nisync_device_data *ddata);
static void nisync_s29gl032n_close(struct nisync_device_data *ddata);
static int nisync_s29gl032n_write8(struct nisync_device_data *ddata, u8 data,
				u32 offset);
static int nisync_s29gl032n_erase_sector(struct nisync_device_data *ddata,
			const struct nisync_flash_sector_iterator *iter);

static const struct nisync_flash_sector_group nisync_s29gl032n_03_group[] = {
	{
		.sector_size	= 0x010000,
		.sector_count	= 63,
	},
	{
		.sector_size	= 0x002000,
		.sector_count	= 8,
	},
	{ }
};


/* Generic flash access methods */
#define FLASH_READ(offset) nisync_s29gl032n_read8(ddata, offset)

static void inc_flash_sector_iterator(
	struct nisync_flash_sector_iterator *iter)
{
	++iter->sector;
	++iter->sector_in_group;
	iter->offset += iter->group->sector_size;
	if (iter->sector_in_group == iter->group->sector_count) {
		++iter->group;
		iter->sector_in_group = 0;
	}
}

static void init_flash_sector_iterator_to_first_sector(
	const struct nisync_flash_sector_group *group,
	struct nisync_flash_sector_iterator *iter)
{
	iter->group = group;
	iter->offset = iter->sector = iter->sector_in_group = 0;
	/* start from sector 1. Sector 0 is off limit */
	inc_flash_sector_iterator(iter);
}

static bool flash_sector_iterator_at_end(
	struct nisync_flash_sector_iterator *iter)
{
	return (iter->group->sector_count == 0);
}

/* PXI-668x have an 8k memory window for accessing the flash */
#define FLASH_WINDOW_SIZE		(0x2000)
#define FLASH_WINDOW_PAGE(offset)	(offset & ~(FLASH_WINDOW_SIZE - 1))
#define FLASH_WINDOW_INDEX(offset)	(offset & (FLASH_WINDOW_SIZE - 1))

static u8 nisync_s29gl032n_read8(struct nisync_device_data *ddata, u32 offset)
{
	writel(FLASH_WINDOW_PAGE(offset), &ddata->asic->__reg7);
	return readb(&ddata->window[FLASH_WINDOW_INDEX(offset)]);
}

static void nisync_flash_write8(struct nisync_device_data *ddata, u8 data,
				u32 offset)
{
	writel(FLASH_WINDOW_PAGE(offset), &ddata->asic->__reg7);
	writeb(data, &ddata->window[FLASH_WINDOW_INDEX(offset)]);
}

static void nisync_s29gl032n_open(struct nisync_device_data *ddata)
{
	dev_dbg(ddata->dev, "%s()\n", __func__);

	/* Unlock bypass command sequence */
	nisync_flash_write8(ddata, 0xAA, 0xAAA);
	nisync_flash_write8(ddata, 0x55, 0x555);
	nisync_flash_write8(ddata, 0x20, 0xAAA);
}

static void nisync_s29gl032n_close(struct nisync_device_data *ddata)
{
	dev_dbg(ddata->dev, "%s()\n", __func__);

	/* Unlock bypass reset command sequence */
	nisync_flash_write8(ddata, 0x90, 0x000);
	nisync_flash_write8(ddata, 0x00, 0x000);
}

static int nisync_s29gl032n_write8(struct nisync_device_data *ddata, u8 value,
				u32 offset)
{
	int tmp;
	int err;

	if (value == 0xFF) {
		if (nisync_s29gl032n_read8(ddata, offset) != 0xFF)
			return -EIO;
		return 0;
	}

	/* Unlock bypass program command sequence */
	nisync_flash_write8(ddata, 0xA0, 0x000);
	nisync_flash_write8(ddata, value, offset);

	/*
	 * with tight-loop, the wait below takes about 7 or more times.
	 * with 20us wait, most of the time it's done after 1 sleep
	 */
	err = readx_poll_timeout(FLASH_READ, offset, tmp, tmp == value,
		20, 2 * USEC_PER_MSEC);
	return err;
}

static int nisync_s29gl032n_erase_sector(struct nisync_device_data *ddata,
			const struct nisync_flash_sector_iterator *iter)
{
	int tmp;
	int err;

	/* Sector error command sequence */
	nisync_flash_write8(ddata, 0xAA, 0xAAA);
	nisync_flash_write8(ddata, 0x55, 0x555);
	nisync_flash_write8(ddata, 0x80, 0xAAA);
	nisync_flash_write8(ddata, 0xAA, 0xAAA);
	nisync_flash_write8(ddata, 0x55, 0x555);
	nisync_flash_write8(ddata, 0x30, iter->offset);

	/* Wait for the flash erase to finish
	 * Note: The maximum expected time to wait is 3.5sec (time required to
	 * erase a sector). The typical time to wait is 0.5sec. The timeout
	 * is padded by 0.5sec.
	 */
	err = readx_poll_timeout(FLASH_READ, iter->offset, tmp, tmp == 0xFF,
		20 * USEC_PER_MSEC, 4 * USEC_PER_SEC);
	return err;
}


static int pxi668x_flash_enable(struct nisync_device_data *ddata,
				const struct firmware *fw)
{
	dev_dbg(ddata->dev, "%s(fw: %p)\n", __func__, fw);

	/* PXI-668x's flash has 64kB sectors; enable flag at sector 1  */
	return nisync_s29gl032n_write8(ddata, 0x7F, 0x10000);
}

static void nisync_setup_flash_window(struct nisync_regs __iomem *asic,
				u32 window)
{
	pr_debug("%s()\n", __func__);

	writel(0, &asic->__reg5);
	writel(REGISTER6_OFFSET1(window) | REGISTER6_OFFSET2 |
			REGISTER6_OFFSET3, &asic->__reg6);
	writel(0, &asic->__reg7);
	writel(REGISTER8_OFFSET1, &asic->__reg8);
}

static void nisync_setup_device_window(struct nisync_regs __iomem *asic,
				u32 window)
{
	pr_debug("%s()\n", __func__);

	writel(REGISTER5_OFFSET1(window) | REGISTER5_OFFSET2,
			&asic->__reg5);
	writel(0, &asic->__reg4);
	writel(0, &asic->__reg6);
}

static void nisync_setup_ram_window(struct nisync_regs __iomem *asic,
				u32 window)
{
	pr_debug("%s()\n", __func__);

	writel(REGISTER6_OFFSET1(window) | REGISTER6_OFFSET2 |
			REGISTER6_OFFSET3, &asic->__reg6);
	writel(0, &asic->__reg5);
	writel(0, &asic->__reg7);
	writel(readl(&asic->__reg8) & ~REGISTER8_OFFSET1,
			&asic->__reg8);
}

/* NOTE: 6683x = 6683 and 6683h */
static struct resource pxi668x_tns_resources[] = {
	{
		.start	= 0x0,
		.end	= 0x0,
		.flags	= IORESOURCE_IRQ,
	}
};
static struct resource pxi668x_enet_resources[] = {
	{
		.start	= 0x0,
		.end	= 0x0,
		.flags	= IORESOURCE_IRQ,
	}
};

static void pxi668x_setup_io_window(struct nisync_regs __iomem *asic,
				u32 window)
{
	pr_debug("%s()\n", __func__);

	nisync_setup_ram_window(asic, window);
}

static void pxi668x_setup_dma(struct nisync_regs __iomem *asic)
{
	pr_debug("%s()\n", __func__);

	writel(REGISTER9_OFFSET1 |
		REGISTER9_OFFSET2 |
		REGISTER9_OFFSET3,
		&asic->__reg9);

	writel(REGISTER10_OFFSET1, &asic->__struct1_array[1].__reg10);
	writel(REGISTER10_OFFSET2, &asic->__struct1_array[1].__reg10);

	writel(REGISTER11_OFFSET1 | REGISTER11_OFFSET2 |
		REGISTER11_OFFSET3 | REGISTER11_OFFSET4 |
		REGISTER11_OFFSET5 | REGISTER11_OFFSET6,
		&asic->__struct1_array[1].__reg11);
	writel(REGISTER12_OFFSET1 |
		REGISTER12_OFFSET3 | REGISTER12_OFFSET4,
		&asic->__struct1_array[1].__reg12);
	writel(REGISTER13_OFFSET2 |
		REGISTER13_OFFSET3 | REGISTER13_OFFSET4 |
		REGISTER13_OFFSET5,
		&asic->__struct1_array[1].__reg13);
	writel(REGISTER14_OFFSET1 | REGISTER14_OFFSET2 |
		REGISTER14_OFFSET3 | REGISTER14_OFFSET4 |
		REGISTER14_OFFSET5 | REGISTER14_OFFSET6,
		&asic->__struct1_array[1].__reg14);

	writel(REGISTER10_OFFSET3, &asic->__struct1_array[1].__reg10);

	writel(readl(&asic->__reg1) | REGISTER1_OFFSET1,
		&asic->__reg1);
	writel(readl(&asic->__reg2) | REGISTER2_OFFSET1,
		&asic->__reg2);
}

static int pxi668x_setup_pdata(struct nisync_device_data *ddata)
{
	struct nisync_platform_data_tns *pdata_tns;
	struct nisync_platform_data_enet *pdata_enet;
	struct pci_dev *pdev;
	u32 tmp;
	u32 word1;
	u32 word2;
	u32 instance_id;
	struct nisync_regs __iomem *asic;
	u8 __iomem *window;
	int err = 0;

	dev_dbg(ddata->dev, "%s()\n", __func__);

	pdata_tns = &ddata->pdata_tns;
	pdata_enet = &ddata->pdata_enet;
	pdev = ddata->pdev;
	asic = ddata->asic;
	window = ddata->window;

	instance_id = atomic_inc_return(&ddata->sync_dev->dev_counter)-1;

	tmp = readl(&asic->__reg8);
	writel(tmp | REGISTER8_OFFSET1, &asic->__reg8);

	pdata_enet->product_id = pdev->device;
	word1 = readl(window + 0x100);
	word2 = readl(window + 0x104);
	pdata_enet->mac_addr[0] = word1 >> 24;
	pdata_enet->mac_addr[1] = word1 >> 16;
	pdata_enet->mac_addr[2] = word1 >> 8;
	pdata_enet->mac_addr[3] = word1;
	pdata_enet->mac_addr[4] = word2 >> 24;
	pdata_enet->mac_addr[5] = word2 >> 16;
	pdata_enet->hardware_clock = &ddata->hardware_clock;
	pdata_enet->mappedio = ddata->window;
	pdata_enet->instance_id = instance_id;
	pdata_enet->product_name = ddata->sync_dev->name;

	pdata_tns->serial_num = readl(window + 0x4);

	writel(tmp & ~REGISTER8_OFFSET1, &asic->__reg8);

	pdata_tns->product_id = pdev->device;
	pdata_tns->instance_id = instance_id;
	pdata_tns->product_name = ddata->sync_dev->name;
	pdata_tns->hardware_clock = &ddata->hardware_clock;
	pdata_tns->mappedio = ddata->window;

	ddata->cells[0].name		= "nisync_tns";
	ddata->cells[0].resources	= pxi668x_tns_resources;
	ddata->cells[0].num_resources	= ARRAY_SIZE(pxi668x_tns_resources);
	ddata->cells[0].platform_data	= &ddata->pdata_tns;
	ddata->cells[0].pdata_size	= sizeof(ddata->pdata_tns);

	ddata->cells[1].name		= "nisync_enet";
	ddata->cells[1].resources	= pxi668x_enet_resources;
	ddata->cells[1].num_resources	= ARRAY_SIZE(pxi668x_enet_resources);
	ddata->cells[1].platform_data	= &ddata->pdata_enet;
	ddata->cells[1].pdata_size	= sizeof(ddata->pdata_enet);

	return err;
}

#define CLK10_PERIOD_IN_NS 100
static struct nisync_device pxi6683_device = {
	.dev_counter			= ATOMIC_INIT(0),
	.name				= "pxi6683",
	.fw_path			= "ni/pxi6683x-latest.bin",

	.nominal_clock_period_in_ns	= 10,
	.time_snap_boundary_in_ns	= CLK10_PERIOD_IN_NS,
};

static struct nisync_device pxi6683h_device = {
	.dev_counter			= ATOMIC_INIT(0),
	.name				= "pxi6683h",
	.fw_path			= "ni/pxi6683x-latest.bin",

	.nominal_clock_period_in_ns	= 10,
	.time_snap_boundary_in_ns	= CLK10_PERIOD_IN_NS,
};

static int nisync_init_firmware_attr(struct nisync_device_data *ddata)
{
	struct fw {
		u32 signature;
		u32 revision;
		u32 oldest_compatible;

		NISYNC_PADDING(12);
		u32 hardware_revision;
	} __iomem *device;

	dev_dbg(ddata->dev, "%s()\n", __func__);

	device = (struct fw __iomem *)ddata->window;
	ddata->firmware.signature = readl(&device->signature);
	ddata->firmware.revision = readl(&device->revision);
	ddata->firmware.oldest_compatible_revision =
		readl(&device->oldest_compatible);
	ddata->firmware.hardware_revision =
		readl(&device->hardware_revision);

	return 0;
}

static int pxi668x_download_firmware(struct nisync_device_data *ddata,
				const struct firmware *fw)
{
	struct nisync_flash_sector_iterator iter;
	u32 flash_offset = 0x20000;
	int i;
	int err = 0;

	dev_dbg(ddata->dev, "%s(fw: %p)\n", __func__, fw);

	if (!fw) {
		dev_err(ddata->dev, "Invalid fw ptr\n");
		return -EINVAL;
	}

	nisync_setup_flash_window(ddata->asic,
		pci_resource_start(ddata->pdev, 1));

	nisync_s29gl032n_open(ddata);

	/* Disable automatic firmare loading */
	init_flash_sector_iterator_to_first_sector(
					nisync_s29gl032n_03_group, &iter);
	err = nisync_s29gl032n_erase_sector(ddata, &iter);
	if (err) {
		dev_err(ddata->dev, "Error %d disabling automatic firmware loading.\n",
			err);
		goto err_disable_firmware_loading;
	}

	/* Assume the firmware offset is at the beginning of the sector */
	while (iter.offset != flash_offset)
		inc_flash_sector_iterator(&iter);
	for (i = 0; i < fw->size; ) {
		if (flash_sector_iterator_at_end(&iter))
			return -EFAULT;
		nisync_s29gl032n_erase_sector(ddata, &iter);
		i += iter.group->sector_size;
		inc_flash_sector_iterator(&iter);
	}

	/* Program firmare */
	for (i = 0; i < fw->size; ++i) {
		err = nisync_s29gl032n_write8(ddata, fw->data[i],
				flash_offset + i);
		if (err) {
			dev_err(ddata->dev, "Error %d writing 0x%02X to flash offset 0x%08X.\n",
				err, fw->data[i], flash_offset + i);
			goto err_program_firmware;
		}
	}

	/* Verify firmware */
	for (i = 0; i < fw->size; ++i) {
		if (fw->data[i] !=
			nisync_s29gl032n_read8(ddata, flash_offset + i)) {
			err = -EIO;
			dev_err(ddata->dev, "Error %d verifying firmware.\n",
				err);
			goto err_verify_firmware;
		}
	}

	/* Enable automatic firmare loading */
	err = pxi668x_flash_enable(ddata, fw);
	if (err) {
		dev_err(ddata->dev, "Error %d enabling automatic firmware loading.\n",
			err);
		goto err_enable_firmware;
	}

err_enable_firmware:
err_verify_firmware:
err_program_firmware:
err_disable_firmware_loading:
	nisync_s29gl032n_close(ddata);
	return err;
}

#define FLASH_OFFSET1		(0x304)

#define REGISTER15		(0x1000)
#define REGISTER15_OFFSET1		BIT(31)
#define REGISTER15_OFFSET2	BIT(30)
#define REGISTER15_OFFSET3	BIT(24)
#define REGISTER15_OFFSET4	BIT(2)

#define REGISTER16		(0x1004)

#define REGISTER17		(0x1008)
#define REGISTER17_OFFSET1	BIT(18)
#define REGISTER17_OFFSET2	BIT(17)
#define REGISTER17_OFFSET3	(0x13)

#define REGISTER18		(0x100C)
#define REGISTER18_OFFSET1		((0x1) << 8)
#define REGISTER18_OFFSET2	BIT(3)
#define REGISTER18_OFFSET3		BIT(1)

#define REGISTER19		(0x1010)
#define REGISTER19_OFFSET1		(0x1)

#define REGISTER20		(0x1014)
#define REGISTER20_OFFSET1		((0x64) << 24)
#define REGISTER20_OFFSET2		BIT(18)
#define REGISTER20_OFFSET3	BIT(17)
#define REGISTER20_OFFSET4	BIT(16)

#define REGISTER21		(0x1018)

#define REGISTER22		(0x101C)
#define REGISTER22_OFFSET1	BIT(31)
#define REGISTER22_OFFSET2	BIT(24)
#define REGISTER22_OFFSET3	BIT(12)
#define REGISTER22_OFFSET4	BIT(6)
#define REGISTER22_OFFSET5	BIT(4)
#define REGISTER22_OFFSET_GROUP1 \
	(REGISTER22_OFFSET2 | \
	REGISTER22_OFFSET4)

#define REGISTER23		(0x1020)
#define REGISTER24	(0x1024)

static int pxi668x_busload_firmware(struct nisync_device_data *ddata,
				const struct firmware *fw)
{
	struct nisync_regs __iomem *asic;
	u8 __iomem *window;
	u32 tmp;
	int i;
	int err = 0;
	struct pci_dev *pdev;
	const int retry_count = 10000;

	pdev = ddata->pdev;
	window = ddata->window;
	asic = ddata->asic;

	nisync_setup_device_window(asic, pci_resource_start(pdev, 1));

	writel(REGISTER15_OFFSET1, &window[REGISTER15]);
	writel(REGISTER15_OFFSET2, &window[REGISTER15]);

	err = readl_poll_timeout(&window[REGISTER22], tmp,
		!(tmp & REGISTER22_OFFSET1) &&
		(tmp & REGISTER22_OFFSET2),
			1 * USEC_PER_MSEC, 1 * USEC_PER_SEC);
	if (err) {
		dev_err(&pdev->dev, "Error %d waiting for configuration engine .\n", err);
		goto err_timeout_in_reset;
	}

	writel(REGISTER15_OFFSET3, &window[REGISTER15]);

	writel(0, &window[REGISTER16]);
	writel(REGISTER17_OFFSET1 |
	       REGISTER17_OFFSET2 |
	       REGISTER17_OFFSET3,
			&window[REGISTER17]);
	writel(REGISTER18_OFFSET1 |
	       REGISTER18_OFFSET2 |
	       REGISTER18_OFFSET3,
			&window[REGISTER18]);
	writel(REGISTER19_OFFSET1, &window[REGISTER19]);
	writel(REGISTER20_OFFSET1 |
	       REGISTER20_OFFSET2 |
	       REGISTER20_OFFSET3 |
	       REGISTER20_OFFSET4,
			&window[REGISTER20]);
	writel(0, &window[REGISTER21]);

	writel(REGISTER15_OFFSET4, &window[REGISTER15]);

	err = readl_poll_timeout(&window[REGISTER22], tmp,
		(tmp & REGISTER22_OFFSET3),
			1 * USEC_PER_MSEC, 1 * USEC_PER_SEC);
	if (err) {
		dev_err(&pdev->dev, "Error %d waiting for ready\n", err);
		goto err_timeout_in_generate_data;
	}

	/* Program firmare */
	if (fw->size % 4 != 0) {
		err = -EINVAL;
		dev_err(&pdev->dev, "Invalid firmware size\n");
		goto err_invalid_firmware_size;
	}

	for (i = 0; i < fw->size; i += 4) {
		u32 data = cpu_to_be32(*((u32 *)&fw->data[i]));

		tmp = readl(&window[REGISTER22]);
		if (tmp & REGISTER22_OFFSET4)
			break;
		if (tmp & REGISTER22_OFFSET5) {
			dev_err(&pdev->dev, "Error configuring FPGA\n");
			err = -EFAULT;
			goto err_fpga_config;
		}

		writel(data, &window[REGISTER23]);
	}

	for (i = 0; i < retry_count; i++) {
		tmp = readl(&window[REGISTER22]);
		if (tmp & REGISTER22_OFFSET4)
			break;
		if (tmp & REGISTER22_OFFSET5) {
			dev_err(&pdev->dev, "Error configuring FPGA\n");
			err = -EFAULT;
			goto err_fpga_config;
		}

		writel(0xFFFFFFFF, &window[REGISTER23]);
	}

	if (i == retry_count) {
		dev_err(&pdev->dev, "Error waiting for FPGA\n");
		err = -EFAULT;
		goto err_fpga_wait;
	}

	pxi668x_setup_io_window(ddata->asic,
				 pci_resource_start(ddata->pdev, 1));

	/*
	 * HACK: This delay is somehow needed or we fail to find the
	 * PHY chip... like the MDIO bus is in a bad state and we can't
	 * communicate through it anymore.
	 */
	msleep(10);

err_fpga_wait:
err_fpga_config:
err_invalid_firmware_size:
err_timeout_in_generate_data:
err_timeout_in_reset:
	return err;
}

static ssize_t nisync_store_sysfs_busload_firmware(struct device *dev,
					struct device_attribute *attr,
					const char *buf,
					size_t count)
{
	struct nisync_device_data *ddata;
	const struct firmware *fw;
	int err;
	int i;
	char *path;

	ddata = dev_get_drvdata(dev);

	path = kzalloc(count + 1, GFP_KERNEL);
	if (!path) {
		err = -ENOMEM;
		dev_err(ddata->dev, "Error no memory to allocate firmware path\n");
		goto err_kzalloc;
	}

	for (i = 0; i < count && buf[i] != '\n'; ++i)
		path[i] = buf[i];

	err = request_firmware(&fw, path, &ddata->pdev->dev);
	if (err) {
		dev_err(ddata->dev, "Error %d requesting firmware %s.\n",
			err, path);
		goto err_request_firmware;
	}

	mutex_lock(&ddata->sysfs_firmware_lock);

	if (ddata->mfd_devices_exist) {
		mfd_remove_devices(&ddata->pdev->dev);
		ddata->mfd_devices_exist = false;
	}

	err = pxi668x_busload_firmware(ddata, fw);
	if (err) {
		dev_err(ddata->dev, "Error %d busloading firmware.\n", err);
		goto err_busload_firmware;
	}

	err = nisync_init_firmware_attr(ddata);
	if (err) {
		dev_err(ddata->dev, "Error %d initializing firmware attributes.\n",
			err);
		goto err_init_firmware_attr;
	}

	err = mfd_add_devices(&ddata->pdev->dev, ddata->mfd_id,
			ddata->cells, NUM_CELLS,
			&ddata->pdev->resource[1], ddata->pdev->irq, NULL);
	if (err) {
		dev_err(ddata->dev, "Error %d adding mfd devices.\n", err);
		goto err_mfd_add_devices;
	}

	ddata->mfd_devices_exist = true;

	mutex_unlock(&ddata->sysfs_firmware_lock);
	release_firmware(fw);

	kfree(path);
	return count;

err_mfd_add_devices:
err_init_firmware_attr:
err_busload_firmware:
	mutex_unlock(&ddata->sysfs_firmware_lock);
	release_firmware(fw);
err_request_firmware:
	kfree(path);
err_kzalloc:
	return err;
}

static int pxi668x_reload_firmware(struct nisync_device_data *ddata)
{
	struct nisync_regs __iomem *asic;
	u8 __iomem *window;
	int i;
	u32 header_offset;
	u32 header[8];
	u32 tmp;
	int err = 0;
	struct pci_dev *pdev;

	dev_dbg(ddata->dev, "%s()\n", __func__);

	pdev = ddata->pdev;
	window = ddata->window;
	asic = ddata->asic;

	nisync_setup_flash_window(asic, pci_resource_start(pdev, 1));

	header_offset = readl(&window[FLASH_OFFSET1]);
	dev_dbg(&pdev->dev, "header_offset: 0x%04X\n", header_offset);

	for (i = 0; i < 8; i++) {
		header[i] = readl(&window[i * 4 + header_offset]);
		dev_dbg(&pdev->dev, "offset: 0x%04X, value: 0x%08X\n",
			i * 4 + header_offset, header[i]);
	}

	nisync_setup_device_window(asic, pci_resource_start(pdev, 1));

	writel(REGISTER15_OFFSET1, &window[REGISTER15]);
	writel(REGISTER15_OFFSET2, &window[REGISTER15]);
	writel(REGISTER15_OFFSET3, &window[REGISTER15]);

	/* Configure firmware loader */
	writel(header[0], &window[REGISTER16]);
	writel(header[1], &window[REGISTER17]);
	writel(header[2], &window[REGISTER18]);
	writel(header[3], &window[REGISTER20]);
	writel(header[4], &window[REGISTER19]);
	/* Data at header[5] is not used */
	writel(header[6], &window[REGISTER21]);
	writel(header[7], &window[REGISTER24]);

	/* Start firmware loader */
	writel(REGISTER15_OFFSET4, &window[REGISTER15]);

	/* Wait for firmware to load */
	err = readl_poll_timeout(&window[REGISTER22], tmp,
		(tmp & REGISTER22_OFFSET_GROUP1) ==
			REGISTER22_OFFSET_GROUP1,
			1 * USEC_PER_MSEC, 1 * USEC_PER_SEC);
	if (err)
		dev_err(&pdev->dev, "Error %d reloading firmware.\n", err);

	pxi668x_setup_io_window(ddata->asic,
				 pci_resource_start(ddata->pdev, 1));

	/*
	 * HACK: This delay is somehow needed or we fail to find the
	 * PHY chip... like the MDIO bus is in a bad state and we can't
	 * communicate through it anymore.
	 */
	msleep(10);

	return err;
}

static ssize_t nisync_store_sysfs_load_firmware_from_device(struct device *dev,
					struct device_attribute *attr,
					const char *buf,
					size_t count)
{
	struct nisync_device_data *ddata;
	int err;

	ddata = dev_get_drvdata(dev);

	mutex_lock(&ddata->sysfs_firmware_lock);

	if (ddata->mfd_devices_exist) {
		mfd_remove_devices(&ddata->pdev->dev);
		ddata->mfd_devices_exist = false;
	}

	pxi668x_reload_firmware(ddata);

	err = nisync_init_firmware_attr(ddata);
	if (err) {
		dev_err(ddata->dev, "Error %d initializing firmware attributes.\n",
			err);
		goto err_init_firmware_attr;
	}

	err = mfd_add_devices(&ddata->pdev->dev, ddata->mfd_id,
			ddata->cells, NUM_CELLS,
			&ddata->pdev->resource[1], ddata->pdev->irq, NULL);
	if (err) {
		dev_err(ddata->dev, "Error %d adding mfd devices.\n", err);
		goto err_mfd_add_devices;
	}

	ddata->mfd_devices_exist = true;

	mutex_unlock(&ddata->sysfs_firmware_lock);

err_mfd_add_devices:
err_init_firmware_attr:
	return count;
}

static ssize_t nisync_store_sysfs_firmware(struct device *dev,
					struct device_attribute *attr,
					const char *buf,
					size_t count)
{
	struct nisync_device_data *ddata;
	const struct firmware *fw;
	int err;
	int i;
	char *path;

	pr_debug("%s(dev: %p, attr: %p, buf: %p, count: %zu)\n", __func__,
				dev, attr, buf, count);

	ddata = dev_get_drvdata(dev);

	path = kzalloc(count + 1, GFP_KERNEL);
	if (!path) {
		err = -ENOMEM;
		dev_err(ddata->dev, "Error no memory to allocate firmware path\n");
		goto err_kzalloc;
	}

	for (i = 0; i < count && buf[i] != '\n'; ++i)
		path[i] = buf[i];

	err = request_firmware(&fw, path, &ddata->pdev->dev);
	if (err) {
		dev_err(ddata->dev, "Error %d requesting firmware %s.\n",
			err, path);
		goto err_request_firmware;
	}

	mutex_lock(&ddata->sysfs_firmware_lock);

	if (ddata->mfd_devices_exist) {
		mfd_remove_devices(&ddata->pdev->dev);
		ddata->mfd_devices_exist = false;
	}

	err = pxi668x_download_firmware(ddata, fw);
	if (err) {
		dev_err(ddata->dev, "Error %d downloading firmware.\n", err);
		goto err_download_firmware;
	}

	err = pxi668x_reload_firmware(ddata);
	if (err) {
		dev_err(ddata->dev, "Error %d reloading firmware.\n", err);
		goto err_reload_firmware;
	}

	err = nisync_init_firmware_attr(ddata);
	if (err) {
		dev_err(ddata->dev, "Error %d initializing firmware attributes.\n",
			err);
		goto err_init_firmware_attr;
	}

	err = mfd_add_devices(&ddata->pdev->dev, ddata->mfd_id,
			ddata->cells, NUM_CELLS,
			&ddata->pdev->resource[1], ddata->pdev->irq, NULL);
	if (err) {
		dev_err(ddata->dev, "Error %d adding mfd devices.\n", err);
		goto err_mfd_add_devices;
	}

	ddata->mfd_devices_exist = true;

	mutex_unlock(&ddata->sysfs_firmware_lock);

	release_firmware(fw);
	kfree(path);
	return count;

err_mfd_add_devices:
err_init_firmware_attr:
err_reload_firmware:
err_download_firmware:
	mutex_unlock(&ddata->sysfs_firmware_lock);
	release_firmware(fw);
err_request_firmware:
	kfree(path);
err_kzalloc:
	return err;
}

static ssize_t nisync_show_sysfs(struct device *dev,
				struct device_attribute *attr, char *buf);

#define S_IRALL (S_IRUSR | S_IRGRP | S_IROTH)

static DEVICE_ATTR(firmware,		S_IWUSR, NULL,
		nisync_store_sysfs_firmware);
static DEVICE_ATTR(busload_firmware,	S_IWUSR, NULL,
		nisync_store_sysfs_busload_firmware);
static DEVICE_ATTR(load_firmware_from_device, S_IWUSR, NULL,
		nisync_store_sysfs_load_firmware_from_device);
static DEVICE_ATTR(signature,		S_IRALL, nisync_show_sysfs, NULL);
static DEVICE_ATTR(revision,		S_IRALL, nisync_show_sysfs, NULL);
static DEVICE_ATTR(oldest_compatible_revision,	S_IRALL, nisync_show_sysfs,
		NULL);
static DEVICE_ATTR(hardware_revision,	S_IRALL, nisync_show_sysfs, NULL);
static DEVICE_ATTR(serial_number,	S_IRALL, nisync_show_sysfs, NULL);

static ssize_t nisync_show_sysfs(struct device *dev,
				struct device_attribute *attr, char *buf)
{
	struct nisync_device_data *ddata;
	u32 var;

	pr_debug("%s(dev: %p, attr: %p, buf: %p)\n", __func__, dev, attr, buf);

	ddata = dev_get_drvdata(dev);

	if (attr == &dev_attr_signature)
		var = ddata->firmware.signature;
	else if (attr == &dev_attr_revision)
		var = ddata->firmware.revision;
	else if (attr == &dev_attr_oldest_compatible_revision)
		var = ddata->firmware.oldest_compatible_revision;
	else if (attr == &dev_attr_hardware_revision)
		var = ddata->firmware.hardware_revision;
	else if (attr == &dev_attr_serial_number)
		var = ddata->pdata_tns.serial_num;
	else
		return -ENOENT;

	return scnprintf(buf, PAGE_SIZE, "0x%08X\n", var);
}

static struct attribute *firmware_attribute[] = {
	&dev_attr_firmware.attr,
	&dev_attr_busload_firmware.attr,
	&dev_attr_load_firmware_from_device.attr,
	&dev_attr_signature.attr,
	&dev_attr_revision.attr,
	&dev_attr_oldest_compatible_revision.attr,
	&dev_attr_hardware_revision.attr,
	&dev_attr_serial_number.attr,
	NULL
};

static const struct attribute_group firmware_attr_group = {
	.attrs = firmware_attribute,
};

static const struct attribute_group *sysfs_groups[] = {
	&firmware_attr_group,
	NULL
};

#define NISYNC_DEVICE(pid, dev) \
	{ PCI_DEVICE(PCI_VENDOR_ID_NI, pid), \
	.driver_data = (kernel_ulong_t) dev }

static struct pci_device_id nisync_pci_tbl[] = {
	NISYNC_DEVICE(PCI_DEV_ID_NI_PXI6683, &pxi6683_device),
	NISYNC_DEVICE(PCI_DEV_ID_NI_PXI6683H, &pxi6683h_device),
	{ },
};

MODULE_DEVICE_TABLE(pci, nisync_pci_tbl);

/*
 * The pxi6683(h) has a software controlled DAC connected to the control pin of
 * a VCO, so we program the DAC to adjust the clock output of the VCO.
 * - The DAC is referenced to 3.0V, giving us a usable range of 0V to 3.0V
 * - The VCO has a 10 Mhz nominal clock and input characteristics of:
 *   Control Voltage (Vc): 0.6Vc minimum, 1.65Vc nominal, 3.0Vc maximum
 *   Tuning Frequency: +- 17.5 ppm minimum
 * Note: the VCO operates down to Vc = 0.0V, but the pull curve is flat
 * below Vc = 0.6V.
 */

/*
 * Following values are for 6683
 */
#define VCO_MAX_ADJUSTMENT_RATE (19687500)
#define VCO_MIN_ADJUSTMENT_RATE (-15312500)
#define PXI668X_VCO_PULLABILITY_IN_PPB 35000;
#define PXI668X_VCO_MAX_NANOVOLTS (3.00 * 1e9)
#define PXI668X_VCO_MIN_NANOVOLTS (0.60 * 1e9)
#define PXI668X_VCO_NOM_NANOVOLTS (1.65 * 1e9)

#define PXI668X_DAC_MIN_NANOVOLTS (0.00 * 1e9)
#define PXI668X_DAC_MAX_NANOVOLTS (3.00 * 1e9)

#define PXI668X_DAC_BITS_OF_RESOLUTION 21
#define PXI668X_DAC_MAX_VALUE ((BIT(PXI668X_DAC_BITS_OF_RESOLUTION)) - 1)
static void nisync_setup_hardware_clock(struct nisync_device_data *ddata)
{
	struct nisync_hardware_clock *hc = &ddata->hardware_clock;

	hc->tnsregs = (struct tns_regs __iomem *)ddata->window;
	mutex_init(&hc->dev_lock);
	init_rwsem(&hc->dev_reset_lock);
	raw_spin_lock_init(&hc->timestamp_counter_lock);

	hc->last_phase_adjust_offset = 0;
	hc->adjustment_rate_in_ps = 0;

	hc->nominal_clock_period_in_ns = ddata->sync_dev->nominal_clock_period_in_ns;
	hc->time_snap_boundary_in_ns = ddata->sync_dev->time_snap_boundary_in_ns;
	hc->nominal_frequency = NSEC_PER_SEC / hc->nominal_clock_period_in_ns;

	hc->vco_max_adj_rate = VCO_MAX_ADJUSTMENT_RATE;
	hc->vco_min_adj_rate = VCO_MIN_ADJUSTMENT_RATE;

	hc->vco_pullability_in_ppb = PXI668X_VCO_PULLABILITY_IN_PPB;
	hc->vco_max_nanovolts = PXI668X_VCO_MAX_NANOVOLTS;
	hc->vco_min_nanovolts = PXI668X_VCO_MIN_NANOVOLTS;
	hc->vco_nom_nanovolts = PXI668X_VCO_NOM_NANOVOLTS;

	hc->dac_min_nanovolts = PXI668X_DAC_MIN_NANOVOLTS;
	hc->dac_max_nanovolts = PXI668X_DAC_MAX_NANOVOLTS;
	hc->dac_max_value = PXI668X_DAC_MAX_VALUE;

	hc->system_time_latency = readl(&hc->tnsregs->system_time_latency);

	hc->_get_time = _nisync_hardware_clock_get_time;
	hc->get_time = nisync_hardware_clock_get_time;

	hc->_set_time = _nisync_hardware_clock_set_time;
	hc->set_time = nisync_hardware_clock_set_time;

	hc->_reset_time = _nisync_hardware_clock_reset_time;
	hc->_adjust_time = _nisync_hardware_clock_adjust_time;
	hc->adjust_time = nisync_hardware_clock_adjust_time;

	hc->get_adj_rate = nisync_hardware_clock_get_adj_rate;

	hc->_set_adj_rate = _nisync_hardware_clock_set_adj_rate;
	hc->set_adj_rate = nisync_hardware_clock_set_adj_rate;

	hc->get_frequency = nisync_hardware_clock_get_frequency;
	hc->set_frequency = nisync_hardware_clock_set_frequency;
}

static int nisync_probe(struct pci_dev *pdev,
				const struct pci_device_id *id)
{
	struct nisync_device_data *ddata;
	struct nisync_regs __iomem *asic;
	int err;
	u32 tmp;
	const struct firmware *fw;

	dev_dbg(&pdev->dev, "%s()\n", __func__);

	if (dma_set_mask(&pdev->dev, DMA_BIT_MASK(32)) ||
			dma_set_coherent_mask(&pdev->dev, DMA_BIT_MASK(32))) {
		dev_err(&pdev->dev, "No usable DMA addressing method.\n");
		return -EIO;
	}

	ddata = devm_kzalloc(&pdev->dev, sizeof(*ddata), GFP_KERNEL);
	if (!ddata)
		return -ENOMEM;

	ddata->pdev = pdev;
	pci_set_drvdata(pdev, ddata);

	ddata->sync_dev = (struct nisync_device *) id->driver_data;
	mutex_init(&ddata->sysfs_firmware_lock);

	err = pci_enable_device(pdev);
	if (err) {
		dev_err(&pdev->dev, "Error %d enabling pci device!\n", err);
		goto err_pci_enable_device;
	}

	pci_set_master(pdev);

	err = pci_request_regions(pdev, KBUILD_MODNAME);
	if (err < 0) {
		dev_err(&pdev->dev, "Error %d: memory address range already allocated!\n",
			err);
		goto err_request_regions;
	}

	ddata->asic = pci_ioremap_bar(ddata->pdev, 0);
	ddata->window = pci_ioremap_bar(ddata->pdev, 1);
	if ((ddata->asic == NULL) || (ddata->window == NULL))
		goto err_map_windows;
	asic = ddata->asic;

	err = readl_poll_timeout(&asic->__reg3, tmp,
			!(tmp & REGISTER3_OFFSET1),
			1 * USEC_PER_MSEC, 50 * USEC_PER_MSEC);
	if (err) {
		dev_err(&pdev->dev, "Timed out waiting for FPGA autoload.\n");
		goto err_fpga_load_timeout;
	}

	pxi668x_setup_io_window(asic, pci_resource_start(pdev, 1));

	err = request_firmware(&fw, ddata->sync_dev->fw_path, &ddata->pdev->dev);
	if (err) {
		dev_warn(&pdev->dev, "Error %d requesting firmware %s. Loading firmware from device.\n",
			err, ddata->sync_dev->fw_path);

		err = pxi668x_reload_firmware(ddata);
		if (err) {
			dev_err(&pdev->dev, "Error %d loading firmware from device.\n", err);
			goto err_reload_firmware;
		}
	} else {
		err = pxi668x_busload_firmware(ddata, fw);
		if (err) {
			dev_err(&pdev->dev, "Error %d busloading firmware.\n", err);
			goto err_fpga_load_err;
		}

		release_firmware(fw);
	}

	pxi668x_setup_dma(asic);

	err = nisync_init_firmware_attr(ddata);
	if (err) {
		dev_err(&pdev->dev, "Error %d initializing firmware attributes.\n",
			err);
		goto err_init_firmware_attr;
	}

	nisync_setup_hardware_clock(ddata);

	err = pxi668x_setup_pdata(ddata);
	if (err) {
		dev_err(&pdev->dev, "Error %d setting up platform data.\n",
			err);
		goto err_setup_pdata;
	}

	ddata->mfd_id = atomic_inc_return(&total_devices) - 1;

	err = mfd_add_devices(&pdev->dev, ddata->mfd_id,
		ddata->cells, NUM_CELLS,
		&pdev->resource[1], pdev->irq, NULL);
	if (err) {
		dev_err(&pdev->dev, "Error %d adding multi-function devices.\n",
			err);
		goto err_mfd_add_devices;
	}
	ddata->mfd_devices_exist = true;

	ddata->dev = device_create_with_groups(nisync_class, &ddata->pdev->dev,
		MKDEV(0, 0), ddata, sysfs_groups,
		"%s.%d", ddata->pdata_tns.product_name,
		ddata->pdata_tns.instance_id);

	if (IS_ERR(ddata->dev)) {
		err = PTR_ERR(ddata->dev);
		dev_err(&pdev->dev, "Error %d creating sysfs device and files.\n",
			err);
		goto err_create_sysfs_files;
	}

	return 0;

err_create_sysfs_files:
	mfd_remove_devices(&pdev->dev);
	ddata->mfd_devices_exist = false;
err_mfd_add_devices:
	writel(0, &asic->__reg5);
	writel(0, &asic->__reg6);
err_setup_pdata:
err_init_firmware_attr:
err_fpga_load_err:
	release_firmware(fw);
err_reload_firmware:
err_fpga_load_timeout:
err_map_windows:
	if (ddata->asic)
		pci_iounmap(ddata->pdev, ddata->asic);
	if (ddata->window)
		pci_iounmap(ddata->pdev, ddata->window);
	pci_release_regions(pdev);
err_request_regions:
	pci_disable_device(pdev);
err_pci_enable_device:
	return err;
}

static void nisync_remove(struct pci_dev *pdev)
{
	struct nisync_device_data *ddata;

	dev_dbg(&pdev->dev, "%s()\n", __func__);
	ddata = pci_get_drvdata(pdev);

	/* can't use device_destroy as we use MKDEV(0,0) for all instances */
	device_unregister(ddata->dev);

	if (ddata->mfd_devices_exist)
		mfd_remove_devices(&pdev->dev);

	pci_iounmap(ddata->pdev, ddata->asic);
	pci_iounmap(ddata->pdev, ddata->window);

	pci_release_regions(pdev);
	pci_disable_device(pdev);
}

static struct pci_driver nisync_pci_driver = {
	.name		= KBUILD_MODNAME,
	.id_table	= nisync_pci_tbl,
	.probe		= nisync_probe,
	.remove		= __devexit_p(nisync_remove),
};

static int __init nisync_init(void)
{
	int err = 0;

	pr_debug("%s()\n", __func__);

	nisync_class = class_create(THIS_MODULE, KBUILD_MODNAME);
	if (IS_ERR(nisync_class)) {
		err = PTR_ERR(nisync_class);
		pr_debug("Error %d creating nisync class.\n", err);
		goto err_create_class;
	}

	err = pci_register_driver(&nisync_pci_driver);
	if (err) {
		pr_debug("Error %d registering pci driver.\n", err);
		goto err_pci_register_driver;
	}

	return 0;

err_pci_register_driver:
	class_unregister(nisync_class);
	nisync_class = NULL;
err_create_class:
	return err;
}

static void __exit nisync_exit(void)
{
	pr_debug("%s()\n", __func__);
	pci_unregister_driver(&nisync_pci_driver);
	class_unregister(nisync_class);
	nisync_class = NULL;
}

module_init(nisync_init);
module_exit(nisync_exit);

MODULE_LICENSE("GPL v2");

