/*
 * nisync_tns_main.c: implements timing and triggering functionality of the
 * PXI-6683H, PXI-6683.
 *
 * Copyright (c) 2012-2018 National Instruments
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/slab.h>
#include <linux/device.h>
#include <linux/bug.h>
#include <linux/cdev.h>
#include <linux/interrupt.h>
#include <linux/list.h>
#include <linux/module.h>
#include <linux/mutex.h>
#include <linux/platform_device.h>
#include <linux/device.h>
#include <linux/poll.h>
#include <linux/rwsem.h>
#include <linux/sched.h>
#include <linux/time.h>
#include <linux/timer.h>
#include <linux/wait.h>
#include <linux/atomic.h>
#include <linux/delay.h>
#include <linux/serial.h>
#include <linux/serial_core.h>
#include <linux/tty_flip.h>
#include <stdbool.h>
#include "nisync.h"
#include "nisync_reg.h"
#include "nisync_pdata.h"
#include "nisync_buf.h"
#include "nisync_kern.h"
#include "fte_fifo.h"
#include "nisync_hardware_clock.h"

#define NISYNC_TRIGGER_TIMESTAMP_BUFFER_SIZE 0x1000

/* Frequency adjust safety margin where adjustments will begin */
#define FREQUENCY_ADJUST_SAFETY_BUFFER_IN_NS 1000000ULL

#define CLK10_PERIOD_IN_NS 100
#define PSEC_PER_NSEC 1000
#define PSEC_PER_SEC ((u64)NSEC_PER_SEC * PSEC_PER_NSEC)

/* Utility macros */
#define create_u64(_upper_u32_, _lower_u32_)				\
	((((u64)(_upper_u32_)) << 32) | ((u64)(_lower_u32_)))

#define container_of_array(ptr, type, member, index) ({			\
	const typeof(((type *)0)->member[index])*__mptr = (ptr);	\
	(type *)((char *)__mptr - offsetof(type, member[index])); })

#define tns_dev_dbg_ioctl(__IOCTL__) ({					\
	dev_dbg(&tns_dev->pdev->dev,					\
		"IOCTL: D: %lu, T: 0x%lx NR: %lu, S: %lu\n",		\
		(unsigned long) _IOC_DIR(__IOCTL__),			\
		(unsigned long) _IOC_TYPE(__IOCTL__),			\
		(unsigned long) _IOC_NR(__IOCTL__),			\
		(unsigned long) _IOC_SIZE(__IOCTL__)); })

#define tns_dev_dbg_ts(_label_, _ts_) ({				\
	dev_dbg(&tns_dev->pdev->dev, _label_ ": %10llds %9lldns\n",	\
		_ts_ / NSEC_PER_SEC,					\
		_ts_ % NSEC_PER_SEC); })

#define tns_dev_dbg_ts_now(...) ({					\
	u64 _curr_time_;						\
									\
	tns_dev->hardware_clock->_get_time(tns_dev->hardware_clock,	\
						&_curr_time_);		\
	tns_dev_dbg_ts("Time", _curr_time_);				\
	dev_dbg(&tns_dev->pdev->dev, ##__VA_ARGS__); })			\

#define tns_dev_dbg_pkt(_msg_, _pkt_) ({				\
	struct nisync_package_timestamp *ts_pkt = (_pkt_);		\
	u64 pkt_timestamp = create_u64(ts_pkt->parts[0], ts_pkt->parts[1]);\
	tns_dev_dbg_ts_now(						\
		"%s: time: %10llds %9lldns, clkID: %x-%x-%x-%x-%x-%x-%x-%x, "\
		"seqID: %02x%02x ctrlID: %x\n",				\
		(_msg_),						\
		pkt_timestamp / NSEC_PER_SEC,				\
		pkt_timestamp % NSEC_PER_SEC,				\
		(ts_pkt->parts[2] >> 8) & 0xFF,				\
		(ts_pkt->parts[2] >> 16) & 0xFF,			\
		(ts_pkt->parts[2] >> 24) & 0xFF,			\
		(ts_pkt->parts[3] >> 0) & 0xFF,				\
		(ts_pkt->parts[3] >> 8) & 0xFF,				\
		(ts_pkt->parts[3] >> 16) & 0xFF,			\
		(ts_pkt->parts[3] >> 24) & 0xFF,			\
		(ts_pkt->parts[4] >> 0) & 0xFF,				\
		(ts_pkt->parts[4] >> 8) & 0xFF,				\
		(ts_pkt->parts[4] >> 16) & 0xFF,			\
		(ts_pkt->parts[4] >> 24) & 0xFF); })

#define tns_dev_dbg_fte(_fte_, _term_name_) ({				\
	switch (_fte_.mode) {						\
	case NISYNC_WRITE_FTE:						\
		dev_dbg(&tns_dev->pdev->dev,				\
			"FTE( %s ): time: %10llds %9lldns, level: %s\n",\
			_term_name_,					\
			_fte_.data.fte.time / NSEC_PER_SEC,		\
			_fte_.data.fte.time % NSEC_PER_SEC,		\
			_fte_.data.fte.level == NISYNC_LEVEL_LOW ?	\
					"low" : "high");		\
		break;							\
	case NISYNC_WRITE_CLOCK:					\
		dev_dbg(&tns_dev->pdev->dev,				\
				"Clock( %s ): start: %10llds %9lldns, "	\
				"end: %10llds %9lldns "			\
				"high_ticks: %10u, "			\
				"low_ticks: %10u\n",			\
			_term_name_,					\
			_fte_.data.clock.start / NSEC_PER_SEC,		\
			_fte_.data.clock.start % NSEC_PER_SEC,		\
			_fte_.data.clock.end / NSEC_PER_SEC,		\
			_fte_.data.clock.end % NSEC_PER_SEC,		\
			_fte_.data.clock.high_ticks,			\
			_fte_.data.clock.low_ticks);			\
		break;							\
	case NISYNC_WRITE_PULSE:					\
		dev_dbg(&tns_dev->pdev->dev,				\
				"Pulse( %s ): time: %10llds %9lldns, "	\
				"width: %10llds %9lldns "		\
				"level: %s\n",				\
			_term_name_,					\
			_fte_.data.pulse.time / NSEC_PER_SEC,		\
			_fte_.data.pulse.time % NSEC_PER_SEC,		\
			_fte_.data.pulse.width / NSEC_PER_SEC,		\
			_fte_.data.pulse.width % NSEC_PER_SEC,		\
			_fte_.data.pulse.level == NISYNC_LEVEL_LOW ?	\
					"low" : "high");		\
		break;							\
	case NISYNC_ABORT_FTE:						\
		dev_dbg(&tns_dev->pdev->dev,				\
			"Abort FTE ( %s ): time: %10llds %9lldns\n",	\
			_term_name_,					\
			_fte_.data.fte.time / NSEC_PER_SEC,		\
			_fte_.data.fte.time % NSEC_PER_SEC);		\
		break;							\
	case NISYNC_ABORT_CLOCK:					\
		dev_dbg(&tns_dev->pdev->dev, "Abort Clock ( %s )\n",	\
			_term_name_);					\
		break;							\
	default:							\
		{							\
			int i;						\
			u32 *tmp_buff = (u32 *)user_buf;		\
			dev_dbg(&tns_dev->pdev->dev,			\
				"FTE Command is unknown...dump buffer\n");\
			/* divide by 4 to support 32-bit mode */\
			for (i = 0; i < (count >> 2); i++)		\
				dev_dbg(&tns_dev->pdev->dev, "   0x%08x\n",\
					tmp_buff[i]);\
		}							\
		break;							\
	} })

#define tns_copy_to_user(to, from, n) ({				\
		int _err_ = copy_to_user(to, from, n);			\
		if (_err_) {						\
			dev_dbg(&tns_dev->pdev->dev,			\
				"Error %d in copy_to_user\n", _err_);	\
			_err_ = -EFAULT;				\
		}							\
		_err_; })

#define tns_copy_from_user(to, from, n) ({				\
		int _err_ = copy_from_user(to, from, n);		\
		if (_err_) {						\
			dev_dbg(&tns_dev->pdev->dev,			\
				"Error %d in copy_from_user\n", _err_);	\
			_err_ = -EFAULT;				\
		}							\
		_err_; })

#define start_timer(_timer_, _nanoseconds_) ({				\
	getrawmonotonic(&(_timer_));					\
	timespec_add_ns(&(_timer_), ((s64)_nanoseconds_)); })

#define is_timer_expired(_timeout_) ({					\
	bool timer_expired = false;					\
	struct timespec curr_time;					\
									\
	getrawmonotonic(&curr_time);					\
	if (timespec_compare(&curr_time, &(_timeout_)) > 0)		\
		timer_expired = true;					\
	timer_expired; })

#define nisync_spin_event_timeout(_condition_, _timeout_in_ns_) ({	\
	bool timeout = true;						\
	struct timespec timer;						\
	start_timer(timer, (_timeout_in_ns_));				\
	do {								\
		if (_condition_) {					\
			timeout = false;				\
			break;						\
		}							\
	} while (!is_timer_expired(timer));				\
	timeout; })

enum nisync_eth_link_speed {
	eth_link_speed_100	= 0,
	eth_link_speed_10	= 1,
	eth_link_speed_1000	= 2
};

enum nisync_irq_state_bit_pos {
	irq_state_hw_rdy,	/* HW is ready for interrupts */
	irq_state_sw_rdy,	/* SW is ready for interrupts */
	irq_state_enabled,	/* Interrupts are enabled */
};

enum nisync_terminal_state {
	nisync_term_ts_enabled,
	nisync_term_ts_sw_overflow,
	nisync_term_ts_hw_overflow,
	nisync_term_ts_aborted,
	nisync_term_fte_enabled,
	nisync_term_fte_clk_enabled,
	nisync_term_irig_enabled,
};

enum nisync_led_power {
	nisync_led_power_off,
	nisync_led_power_on,
	nisync_led_power_blink_1_sec,
	nisync_led_power_blink_2_sec,
};

enum nisync_led_color {
	nisync_led_color_none,
	nisync_led_color_red,
	nisync_led_color_green,
	nisync_led_color_orange,
};

/* niSync device context */
struct nisync_tns_dev {

	struct platform_device *pdev;
	struct nisync_platform_data_tns *pdata;

	s32 product_support_idx;
	u32 driver_instance_id;

	struct nisync_hardware_clock *hardware_clock;
	/*
	 * TODO: should we allow concurrent read/get calls?
	 * Or at least make this lock tighter around accesses to HW?
	 */
	/* prevent concurrent calls to the device */
	struct mutex *dev_lock;

	/* prevent device reset with down_read */
	struct rw_semaphore *dev_reset_lock;

	u32 signature;
	u32 revision;
	u32 oldest_compatible_revision;
	u32 hardware_revision;

	struct tns_regs __iomem *regs;

	spinlock_t irq_state_spinlock;
	unsigned long irq_state;

	struct tasklet_struct irq_tasklet;

	u32 system_time_latency;
	u32 eth_tx_latency_10mbps;
	u32 eth_tx_latency_100mbps;
	u32 eth_tx_latency_1000mbps;
	u32 eth_rx_latency_10mbps;
	u32 eth_rx_latency_100mbps;
	u32 eth_rx_latency_1000mbps;
	u32 input_timestamp_latency;
	u32 partial_clear_operation_latency;
	u32 future_time_latency;
	u32 gps_pps_latency;
	u32 irig_dc_latency;
	u32 irig_am_latency;

	/* HW variables for adjusting rate */
	u32 oscillator_frequency;
	u32 nominal_frequency_comp;
	u64 nco_precision;

	u32 nominal_frequency;
	u32 nominal_clock_period_in_ns;

	enum nisync_ptpd_state ptp_state;
	enum nisync_status status;

	/* Prevent concurrent accesses to the HW & SW FTE FIFO */
	struct mutex fte_lock;
	struct fte_fifo *fte_fifo;

	struct nisync_irig_packet irig_packet;
	enum nisync_irig_mode irig_mode;
	seqlock_t irig_seqlock;

	struct nisync_sysfs_dev {

		struct cdev char_dev;
		struct device *dev;

		u32 term;

		/* Protect against state changes on each terminal */
		/* TODO: are there places where term_lock could be tighter? */
		struct mutex term_lock;

		unsigned long term_state;
		unsigned long open_count;

		union {
			struct {
				u8 edge_detection;
				u8 last_edge;

				bool initial_state_set;

				struct nisync_buf timestamps;
			};

			struct nisync_buf pkts;
		};

		bool wake_up;
		wait_queue_head_t wait_queue;

		struct list_head list_node;

	} sysfs_devs[NISYNC_DEVICE_MINOR_STEP];

	struct uart_driver nisync_gps_uart_driver;
	struct uart_port nisync_gps_uart_port;
	seqlock_t gps_seqlock;
	u64 gps_pps_ts;
	u16 gps_tx_buffer_size;
	u16 gps_rx_buffer_size;

	spinlock_t active_ts_terms_spinlock;
	struct list_head active_ts_terms;
};

#define irq_test_state(_tns_dev_, _state_) ({				\
	test_bit((_state_), &(_tns_dev_)->irq_state); })

#define irq_set_state(_tns_dev_, _state_) ({				\
	set_bit((_state_), &(_tns_dev_)->irq_state); })

#define irq_clear_state(_tns_dev_, _state_) ({				\
	clear_bit((_state_), &(_tns_dev_)->irq_state); })

#define terminal_test_state(_sysfs_dev_, _state_) ({			\
dev_dbg(&tns_dev->pdev->dev,						\
	"terminal_test_state(%s, state bit = %d) = 0x%lx\n",		\
	terminal_name((_sysfs_dev_)->term), _state_,			\
			(_sysfs_dev_)->term_state);			\
	test_bit((_state_), &(_sysfs_dev_)->term_state); })

#define terminal_set_state(_sysfs_dev_, _state_) ({			\
dev_dbg(&tns_dev->pdev->dev,						\
	"terminal_set_state(%s, state bit = %d) = 0x%lx\n",		\
	terminal_name(_sysfs_dev_->term), _state_, _sysfs_dev_->term_state);\
	set_bit((_state_), &(_sysfs_dev_)->term_state); })

#define terminal_clear_state(_sysfs_dev_, _state_) ({			\
dev_dbg(&tns_dev->pdev->dev,						\
	"terminal_clear_state(%s, state bit = %d) = 0x%lx\n",		\
	terminal_name(_sysfs_dev_->term), _state_, _sysfs_dev_->term_state);\
	clear_bit((_state_), &(_sysfs_dev_)->term_state); })

/* Prototypes */
static int _get_trig_ts_pxi668x(struct nisync_tns_dev *tns_dev,
				u64 *timestamp, u64 *trig_lines,
				bool *initial_state);

static int _get_trig_ts_status_pxi668x(struct nisync_tns_dev *tns_dev,
				u32 *ts_available, bool *ts_fifo_overflow);

static int _nisync_reset_pxi668x(struct nisync_tns_dev *tns_dev);

static void _nisync_set_1588_status_led_pxi668x(struct nisync_tns_dev *tns_dev,
				enum nisync_led_power power,
				enum nisync_led_color color);

static void _nisync_adjust_pkt_ts_pxi668x(struct nisync_tns_dev *tns_dev,
				u64 *pkt_ts);

static int _write_fte_fifo_pxi668x(struct nisync_tns_dev *tns_dev,
				const u64 fte_time,
				const u64 terminal_bitmask);

static int _nisync_abort_fte(struct nisync_tns_dev *tns_dev,
				struct nisync_write_buf *buf,
				const u64 term_bitmask, const u64 curr_time,
				size_t *earliest_modified_fte_index);
static int _nisync_disable_future_time_events(struct nisync_tns_dev *tns_dev,
				struct nisync_sysfs_dev *sysfs_dev);
static int _nisync_disable_timestamp_trigger(struct nisync_tns_dev *tns_dev,
				struct nisync_sysfs_dev *sysfs_dev);
static int _nisync_disable_irig(struct nisync_tns_dev *tns_dev,
				struct nisync_sysfs_dev *sysfs_dev);

/* nisync terminals */
struct nisync_tns_dev_terminal {
	const char *name;
	const struct file_operations *fops;

	union {
		struct {
			u32 future_time_clock_period_offset;
			u32 future_time_clock_duty_cycle_offset;
			u32 control_enable_offset;
			u32 control_disable_offset;
			u32 decimation_count_offset;
			u32 terminal_bit_position;
		} io;
		struct {
			u32 packet_buffer_control_enable_offset;
			u32 packet_buffer_control_disable_offset;
			u32 packet_buffer_status_offset;
			u32 packet_instr_fifo_offset;
			u32 packet_fifo_read_offset;
		} eth;
	} u;
};

struct nisync_tns_dev_terminal tns_dev_terminals_pxi[];

#define round_to_nearest(_value_, _interval_) \
	((_interval_) == 0 ? (_value_) : \
	(((_value_) + (_interval_) / (s64)2) / (_interval_)) * (_interval_))

/*
 * VCTCXO adjustment rate max and min based on the input characteristics of the
 * control voltage (Vc): 0.6V min, 1.65V nom, 3.0V max
 * and tuning frequency of +- 17.5 ppm
 * adj_rate = ((nVc - nVc_nom) * PSEC_PER_NSEC * PULLABILITY_IN_PPB) /
 *						(nVc_max - nVc_min)
 * Rates are in picoseconds per second
 */
#define VCTCXO_MAX_ADJUSTMENT_RATE (19687500)
#define VCTCXO_MIN_ADJUSTMENT_RATE (-15312500)

#define PXI668x_FTE_REARM_TIME_TICKS 4
#define PXI668x_FTE_MIN_DUTY_CYCLE_TICKS 1
#define PXI668x_FTE_MIN_CLK_PERIOD_TICKS 2
#define PXI668x_MAX_DECIMATION_COUNT ((u32)(BIT(20) - 1))

/* Product support */
struct nisync_product_support {
	u32 product_id;
	const char *model_name;
	u32 compatible_hardware_revision;
	u32 newest_compatible_fpga;
	u32 oldest_compatible_fpga;

	bool clkin_supported;
	u32 terminals[NISYNC_MAX_TERMINALS];
};

struct nisync_product_support product_support_table[] = {
	{.product_id			= 0x75E5,
	 .model_name			= "PXI-6683",
	 .compatible_hardware_revision	= 0x07052020,
	 .newest_compatible_fpga	= 0x16120116,
	 .oldest_compatible_fpga	= 0x14121920,


	 .clkin_supported		= true,
	 .terminals			= {
	 NISYNC_PFI0, NISYNC_PFI1, NISYNC_PFI2,
	 /* Trigger lines are known as PXI_TRIG on PXI devices */
	 NISYNC_PXI_TRIG0, NISYNC_PXI_TRIG1, NISYNC_PXI_TRIG2,
	 NISYNC_PXI_TRIG3, NISYNC_PXI_TRIG4, NISYNC_PXI_TRIG5,
	 NISYNC_PXI_TRIG6, NISYNC_PXI_TRIG7,
	 /*
	  * 6683 _can_ go in system timing slot, where it would have access to
	  * all STAR lines. Since we don't have an easy way to tell whether we
	  * are in the system timing slot, we always expose the lines and leave
	  * it up to the customer to decide whether using them is appropriate
	  * in their setup.
	  */
	 NISYNC_PXI_STAR0, NISYNC_PXI_STAR1, NISYNC_PXI_STAR2,
	 NISYNC_PXI_STAR3, NISYNC_PXI_STAR4, NISYNC_PXI_STAR5,
	 NISYNC_PXI_STAR6, NISYNC_PXI_STAR7, NISYNC_PXI_STAR8,
	 NISYNC_PXI_STAR9, NISYNC_PXI_STAR10, NISYNC_PXI_STAR11,
	 NISYNC_PXI_STAR12,
	 NISYNC_CLK10, NISYNC_GND,
	 NISYNC_ETHIN, NISYNC_ETHOUT,
	 NISYNC_LAST_TERMINAL} },

	{.product_id			= 0x75E6,
	 .model_name			= "PXI-6683H",
	 .compatible_hardware_revision	= 0x07052020,
	 .newest_compatible_fpga	= 0x16120116,
	 .oldest_compatible_fpga	= 0x14121920,

	 .clkin_supported		= false,
	 .terminals			= {
	 NISYNC_PFI0, NISYNC_PFI1, NISYNC_PFI2,
	 /* Trigger lines are known as PXI_TRIG on PXI devices */
	 NISYNC_PXI_TRIG0, NISYNC_PXI_TRIG1, NISYNC_PXI_TRIG2,
	 NISYNC_PXI_TRIG3, NISYNC_PXI_TRIG4, NISYNC_PXI_TRIG5,
	 NISYNC_PXI_TRIG6, NISYNC_PXI_TRIG7,
	 /* 6683H can only go in peripheral slot, no STAR lines */
	 NISYNC_CLK10, NISYNC_GND,
	 NISYNC_ETHIN, NISYNC_ETHOUT,
	 NISYNC_LAST_TERMINAL} }
};

#define tns_dev_product_id()({						\
	product_support_table[tns_dev->product_support_idx].product_id; })

#define tns_dev_product_model_name()({					\
	product_support_table[tns_dev->product_support_idx].model_name; })

#define tns_dev_product_compatible_hardware_revision()({		\
	product_support_table[tns_dev->product_support_idx].		\
	compatible_hardware_revision; })

#define tns_dev_product_newest_compatible_fpga()({			\
	product_support_table[tns_dev->product_support_idx].		\
	newest_compatible_fpga; })

#define tns_dev_product_oldest_compatible_fpga()({			\
	product_support_table[tns_dev->product_support_idx].		\
	oldest_compatible_fpga; })

#define tns_dev_product_feature_supported(_feature_)({			\
	product_support_table[tns_dev->product_support_idx].		\
		_feature_##_supported; })

#define tns_dev_product_regmap(_reg_)({					\
	void __iomem *register_address;					\
	register_address = (void __iomem *)			\
		&tns_dev->regs->_reg_;				\
	if ((void __iomem *) &tns_dev->regs->			\
					_register_address_invalid\
		<= register_address) {				\
		dev_err(&tns_dev->pdev->dev,			\
			"register %s invalid for PXI\n", #_reg_);\
		register_address = (void __iomem *)(0xDEADBEEF);\
	}							\
	register_address; })

#define tns_dev_product_terminal(_idx_)({				\
	product_support_table[tns_dev->product_support_idx].		\
		terminals[(_idx_)]; })

/* Forward declarations of supported file operations */
static const struct file_operations nisync_fops;
static const struct file_operations nisync_terminal_fops;
static const struct file_operations nisync_eth_fops;

/* nisync input timestamp terminal bit positions */
enum nisync_terminal_time_event_bit_pos {
	pfi0_time_event_bit_pos		= 20,
	pfi1_time_event_bit_pos		= 21,
	pfi2_time_event_bit_pos		= 22,
	rtsi0_time_event_bit_pos	= 24,
	rtsi1_time_event_bit_pos	= 25,
	rtsi2_time_event_bit_pos	= 26,
	rtsi3_time_event_bit_pos	= 27,
	rtsi4_time_event_bit_pos	= 28,
	rtsi5_time_event_bit_pos	= 29,
	rtsi6_time_event_bit_pos	= 30,
	rtsi7_time_event_bit_pos	= 31,
	star0_time_event_bit_pos	= 32,
	star1_time_event_bit_pos	= 33,
	star2_time_event_bit_pos	= 34,
	star3_time_event_bit_pos	= 35,
	star4_time_event_bit_pos	= 36,
	star5_time_event_bit_pos	= 37,
	star6_time_event_bit_pos	= 38,
	star7_time_event_bit_pos	= 39,
	star8_time_event_bit_pos	= 40,
	star9_time_event_bit_pos	= 41,
	star10_time_event_bit_pos	= 42,
	star11_time_event_bit_pos	= 43,
	star12_time_event_bit_pos	= 44
};

#define pci_decimation_offset(_prefix_) 0xDEADBEEF
#define pxi_decimation_offset(_prefix_) offsetof(struct tns_regs,	\
				_prefix_##_decimation_count)

#define tns_dev_io_term_ex(_name_, _prefix_, _bus_) {			\
	.name = #_name_,						\
	.fops = &nisync_terminal_fops,					\
	.u.io = {							\
		offsetof(struct tns_regs,				\
			_prefix_##_future_time_clock_period),\
		offsetof(struct tns_regs,				\
			_prefix_##_future_time_clock_duty_cycle),\
		offsetof(struct tns_regs,				\
			_prefix_##_control_enable),		\
		offsetof(struct tns_regs,				\
			_prefix_##_control_disable),		\
		_bus_##_decimation_offset(_prefix_),			\
		_prefix_##_time_event_bit_pos} }

#define tns_dev_io_term(_name_, _bus_) tns_dev_io_term_ex(_name_, _name_, _bus_)

#define tns_dev_eth_term(_name_, _prefix_, _bus_) {			\
	.name = #_name_,						\
	.fops = &nisync_eth_fops,					\
	.u.eth = {							\
		offsetof(struct tns_regs,				\
			_prefix_##_packet_buffer_control_enable),\
		offsetof(struct tns_regs,				\
			_prefix_##_packet_buffer_control_disable),\
		offsetof(struct tns_regs,				\
			_prefix_##_packet_buffer_status),	\
		offsetof(struct tns_regs,				\
			_prefix_##_instr_fifo),		\
		offsetof(struct tns_regs,				\
			_prefix_##_packet_fifo_read)} }

#define tns_dev_gen_term(_name_) {.name = #_name_}

struct nisync_tns_dev_terminal tns_dev_terminals_pxi[] = {

	/* PFI terminals */
	tns_dev_io_term(pfi0, pxi), tns_dev_io_term(pfi1, pxi),
	tns_dev_io_term(pfi2, pxi),

	/*
	 * FIXME: we don't support a device with RTSI lines.
	 * Do we want to remove all RTSI references from the source base?
	 */
	/* RTSI terminals */
	tns_dev_io_term(rtsi0, pxi), tns_dev_io_term(rtsi1, pxi),
	tns_dev_io_term(rtsi2, pxi), tns_dev_io_term(rtsi3, pxi),
	tns_dev_io_term(rtsi4, pxi), tns_dev_io_term(rtsi5, pxi),
	tns_dev_io_term(rtsi6, pxi), tns_dev_io_term(rtsi7, pxi),

	/* PXI_Trig terminals */
	tns_dev_io_term_ex(trig0, rtsi0, pxi),
	tns_dev_io_term_ex(trig1, rtsi1, pxi),
	tns_dev_io_term_ex(trig2, rtsi2, pxi),
	tns_dev_io_term_ex(trig3, rtsi3, pxi),
	tns_dev_io_term_ex(trig4, rtsi4, pxi),
	tns_dev_io_term_ex(trig5, rtsi5, pxi),
	tns_dev_io_term_ex(trig6, rtsi6, pxi),
	tns_dev_io_term_ex(trig7, rtsi7, pxi),

	/* PXI_Star terminals */
	tns_dev_io_term(star0, pxi), tns_dev_io_term(star1, pxi),
	tns_dev_io_term(star2, pxi), tns_dev_io_term(star3, pxi),
	tns_dev_io_term(star4, pxi), tns_dev_io_term(star5, pxi),
	tns_dev_io_term(star6, pxi), tns_dev_io_term(star7, pxi),
	tns_dev_io_term(star8, pxi), tns_dev_io_term(star9, pxi),
	tns_dev_io_term(star10, pxi), tns_dev_io_term(star11, pxi),
	tns_dev_io_term(star12, pxi),
	/* Supported devices do not use star13 - star16 */
	tns_dev_gen_term(rsvd), tns_dev_gen_term(rsvd),
	tns_dev_gen_term(rsvd), tns_dev_gen_term(rsvd),

	/* other terminals */
	tns_dev_gen_term(clk10), tns_dev_gen_term(gnd),
	tns_dev_gen_term(rsvd), tns_dev_gen_term(rsvd), tns_dev_gen_term(sw),

	/* Ethernet terminals */
	tns_dev_eth_term(ethin, incoming, pxi),
	tns_dev_eth_term(ethout, outgoing, pxi)
};

#define terminal_name(_terminal_) \
	({tns_dev_terminals_pxi[_terminal_].name; })

#define terminal_fops(_terminal_) \
	({tns_dev_terminals_pxi[_terminal_].fops; })

/* cast to void __iomem * to get pointer arithmetic right */
#define terminal_future_time_clock_period(_tns_dev_, _terminal_) ({	\
	(void __iomem *)_tns_dev_->regs +				\
	tns_dev_terminals_pxi[_terminal_].u.io.		\
				future_time_clock_period_offset; })

#define terminal_future_time_clock_duty_cycle(_tns_dev_, _terminal_) ({	\
	(void __iomem *)_tns_dev_->regs +				\
	tns_dev_terminals_pxi[_terminal_].u.io.		\
				future_time_clock_duty_cycle_offset; })

#define terminal_control_enable(_tns_dev_, _terminal_) ({		\
	(void __iomem *)_tns_dev_->regs +				\
	tns_dev_terminals_pxi[_terminal_].u.io.		\
				control_enable_offset; })

#define terminal_control_disable(_tns_dev_, _terminal_) ({		\
	(void __iomem *)_tns_dev_->regs +				\
	tns_dev_terminals_pxi[_terminal_].u.io.		\
				control_disable_offset; })

#define terminal_decimation_count(_tns_dev_, _terminal_) ({		\
	(void __iomem *)_tns_dev_->regs +				\
	tns_dev_terminals_pxi[_terminal_].u.io.		\
				decimation_count_offset; })

#define terminal_time_event_bit(_terminal_) ({				\
	((u64)(1)) << tns_dev_terminals_pxi[_terminal_].u.io.	\
				terminal_bit_position; })

#define terminal_pkt_control_enable(_tns_dev_, _terminal_) ({		\
	(void __iomem *)(_tns_dev_)->regs +				\
	tns_dev_terminals_pxi[_terminal_].u.eth.		\
				packet_buffer_control_enable_offset; })

#define terminal_pkt_control_disable(_tns_dev_, _terminal_) ({		\
	(void __iomem *)(_tns_dev_)->regs +				\
	tns_dev_terminals_pxi[_terminal_].u.eth.		\
				packet_buffer_control_disable_offset; })

#define terminal_pkt_buffer_status(_tns_dev_, _terminal_) ({		\
	(void __iomem *)(_tns_dev_)->regs +				\
	tns_dev_terminals_pxi[_terminal_].u.eth.		\
				packet_buffer_status_offset; })

#define terminal_pkt_instr_fifo(_tns_dev_, _terminal_) ({		\
	(void __iomem *)(_tns_dev_)->regs +				\
	tns_dev_terminals_pxi[_terminal_].u.eth.		\
				packet_instr_fifo_offset; })

#define terminal_pkt_fifo_read(_tns_dev_, _terminal_) ({		\
	(void __iomem *)(_tns_dev_)->regs +				\
	tns_dev_terminals_pxi[_terminal_].u.eth.		\
				packet_fifo_read_offset; })

/* niSync functions */
static int _nisync_1588_start_capturing(struct nisync_tns_dev *tns_dev)
{
	static u32 ptp_filter_inst[] = {
		/*
		 * Ignore 12 bytes between SFD and Protocol field in the
		 * Ethernet Frame
		 */
		0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
		0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
		0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
		/*
		 * Verify that Ethernet Protocol field (two bytes) is
		 * Protocol IPv4
		 */
		0x2 << 4 | 0x8, 0x2 << 4 | 0x0,
		0x2 << 4 | 0x0, 0x2 << 4 | 0x0,
		/* Ignore 9 bytes between Ethertype and IP Protocol Field */
		0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
		0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
		/* Verify that IP Protocol field (one byte) is UDP */
		0x2 << 4 | 0x1, 0x2 << 4 | 0x1,
		/*
		 * Ignore 6 bytes between Protocol field and Destination IP
		 * address
		 */
		0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
		0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
		/*
		 * Verify that Destination address (4 bytes) is
		 * multicast address
		 */
		0x2 << 4 | 0x0, 0x2 << 4 | 0xe,
		/* 0x2 << 4 | 0x0, 0x2 << 4 | 0x0, */
		/* 0x2 << 4 | 0x1, 0x2 << 4 | 0x0, */
		/* 0x2 << 4 | 0x1, 0x2 << 4 | 0x8, */
		0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
		/* Ignore 2 bytes */
		0x10, 0x10, 0x10, 0x10,
		/* Verify that destination port (2 bytes) is Event Port */
		0x2 << 4 | 0x1, 0x2 << 4 | 0x0,
		0x2 << 4 | 0xF, 0x2 << 4 | 0x3,
		/* Ignore 24 bytes */
		0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
		0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
		0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
		0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
		0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
		0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
		/* Capture Clock ID (8 bytes) */
		0x3 << 4, 0x3 << 4, 0x3 << 4, 0x3 << 4,
		0x3 << 4, 0x3 << 4, 0x3 << 4, 0x3 << 4,
		0x3 << 4, 0x3 << 4, 0x3 << 4, 0x3 << 4,
		0x3 << 4, 0x3 << 4, 0x3 << 4, 0x3 << 4,
		/* Ignore 2 bytes */
		0x10, 0x10, 0x10, 0x10,
		/* Capture Sequence ID (2 bytes) */
		0x3 << 4, 0x3 << 4, 0x3 << 4, 0x3 << 4,
		/* Capture Control field (1 byte) */
		0x3 << 4, 0x3 << 4,
		/* Two stops */
		0x0, 0x0
	};

	int err;
	int i;

	dev_dbg(&tns_dev->pdev->dev, "%s(ptp_filter_inst: %zu)\n",
		__func__, sizeof(ptp_filter_inst) >> 2);

	err = nisync_buf_init(tns_dev->sysfs_devs[NISYNC_ETHIN].pkts,
		struct nisync_package_timestamp, NISYNC_ETH_PKT_BUFFER_SIZE);
	if (err) {
		dev_err(&tns_dev->pdev->dev, "Failed allocating NISYNC_ETHIN pkt buffer.\n  Buffer size: %d\n",
			NISYNC_ETH_PKT_BUFFER_SIZE);
		goto err_cb_init_ethin;
	}

	err = nisync_buf_init(tns_dev->sysfs_devs[NISYNC_ETHOUT].pkts,
		struct nisync_package_timestamp, NISYNC_ETH_PKT_BUFFER_SIZE);
	if (err) {
		dev_err(&tns_dev->pdev->dev, "Failed allocating NISYNC_ETHOUT pkt buffer.\n  Buffer size: %d\n",
			NISYNC_ETH_PKT_BUFFER_SIZE);
		goto err_cb_init_ethout;
	}

	/* Clear snooping instruction FIFO and packet FIFO for TX and RX*/
	writel(BIT(5) | BIT(4),
		terminal_pkt_control_enable(tns_dev, NISYNC_ETHIN));
	writel(BIT(5) | BIT(4),
		terminal_pkt_control_enable(tns_dev, NISYNC_ETHOUT));

	/* Program both incoming and outgoing packet filters */
	for (i = 0; i < (sizeof(ptp_filter_inst) >> 2); i++) {
		writel(ptp_filter_inst[i],
			terminal_pkt_instr_fifo(tns_dev, NISYNC_ETHIN));
		writel(ptp_filter_inst[i],
			terminal_pkt_instr_fifo(tns_dev, NISYNC_ETHOUT));
	}

	/* Enable packet snooping and FIFO not empty interrupts */
	writel(BIT(2) | BIT(0),
		terminal_pkt_control_enable(tns_dev, NISYNC_ETHIN));
	writel(BIT(2) | BIT(0),
		terminal_pkt_control_enable(tns_dev, NISYNC_ETHOUT));

	return 0;

err_cb_init_ethout:
	nisync_buf_destroy(tns_dev->sysfs_devs[NISYNC_ETHIN].pkts);
err_cb_init_ethin:

	return err;
}

static void _nisync_1588_stop_capturing(struct nisync_tns_dev *tns_dev)
{
	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	/* Disable packet snooping and FIFO not empty interrupts */
	writel(BIT(2) | BIT(0),
		terminal_pkt_control_disable(tns_dev, NISYNC_ETHIN));
	writel(BIT(2) | BIT(0),
		terminal_pkt_control_disable(tns_dev, NISYNC_ETHOUT));

	/* Clear snooping instruction FIFO and packet FIFO for TX and RX*/
	writel(BIT(5) | BIT(4),
		terminal_pkt_control_enable(tns_dev, NISYNC_ETHIN));
	writel(BIT(5) | BIT(4),
		terminal_pkt_control_enable(tns_dev, NISYNC_ETHOUT));

	nisync_buf_destroy(tns_dev->sysfs_devs[NISYNC_ETHIN].pkts);
	nisync_buf_destroy(tns_dev->sysfs_devs[NISYNC_ETHOUT].pkts);
}

static int _nisync_reset_pxi668x(struct nisync_tns_dev *tns_dev)
{
	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	writel(1, tns_dev_product_regmap(timing_and_sync_reset_reg));

	return 0;
}

#define NISYNC_BOARD_PLL_NOT_LOCKED_BIT BIT(0)

static bool _nisync_is_pll_locked(struct nisync_tns_dev *tns_dev)
{
	bool pll_locked;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	pll_locked = !(readl(tns_dev_product_regmap(board_status)) &
					NISYNC_BOARD_PLL_NOT_LOCKED_BIT);

	dev_dbg(&tns_dev->pdev->dev, "() => %s\n",
		pll_locked ? "true" : "false");
	return pll_locked;
}

static int _nisync_wait_for_pll_lock(struct nisync_tns_dev *tns_dev)
{
	int err = 0;
	bool timeout;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	timeout = nisync_spin_event_timeout(_nisync_is_pll_locked(tns_dev),
					    NSEC_PER_SEC);
	if (timeout) {
		dev_err(&tns_dev->pdev->dev, "Timeout waiting to get PLL lock\n");
		tns_dev->status = NISYNC_STATUS_HW_CLK_NOT_LOCKED;
		err = -ETIMEDOUT;
	}

	return err;
}

static int nisync_reset(struct nisync_tns_dev *tns_dev)
{
	int err;
	int i;
	unsigned long irq_flags;
	u32 terminal;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	/* Ensure no other operations are outstanding before resetting HW */
	down_write(tns_dev->dev_reset_lock);
	if (tns_dev->status != NISYNC_STATUS_INITIALIZING)
		tns_dev->status = NISYNC_STATUS_RESETTING;

	/* Ensure HW interrupts are disabled */
	spin_lock_irqsave(&tns_dev->irq_state_spinlock, irq_flags);
	writel(1, tns_dev_product_regmap(global_board_settings_disable));
	irq_clear_state(tns_dev, irq_state_enabled);
	irq_clear_state(tns_dev, irq_state_hw_rdy);
	spin_unlock_irqrestore(&tns_dev->irq_state_spinlock, irq_flags);

	tasklet_kill(&tns_dev->irq_tasklet);

	_nisync_1588_stop_capturing(tns_dev);

	/*
	 * Abort all outstanding timestamp reads
	 * ...Note that since the IRQ tasklet has been disabled, it is not
	 *    necessary to acquire the active_ts_terms_spinlock.
	 */
	while (!list_empty(&tns_dev->active_ts_terms)) {
		struct nisync_sysfs_dev *sysfs_dev =
			list_first_entry(&tns_dev->active_ts_terms,
					struct nisync_sysfs_dev, list_node);

		wake_up_all(&sysfs_dev->wait_queue);
		list_del_init(&sysfs_dev->list_node);
	}

	/*
	 * For each terminal that has timestamping or future time events
	 * or irig enabled, update state of terminal to reflect reset
	 */
	i = 0;
	while ((terminal = tns_dev_product_terminal(i)) !=
						NISYNC_LAST_TERMINAL) {
		struct nisync_sysfs_dev *sysfs_dev =
				&tns_dev->sysfs_devs[terminal];
		if (terminal_test_state(sysfs_dev, nisync_term_ts_enabled)) {
			terminal_set_state(sysfs_dev, nisync_term_ts_aborted);
			_nisync_disable_timestamp_trigger(tns_dev, sysfs_dev);
		}

		if (terminal_test_state(sysfs_dev, nisync_term_fte_enabled))
			_nisync_disable_future_time_events(tns_dev, sysfs_dev);

		if (terminal_test_state(sysfs_dev, nisync_term_irig_enabled))
			_nisync_disable_irig(tns_dev, sysfs_dev);

		i++;
	}

	tns_dev->irig_packet.time_and_days = 0;
	tns_dev->irig_packet.year_and_seconds = 0;
	tns_dev->irig_packet.arrival_hw_ts = 0;
	tns_dev->irig_mode = NISYNC_IRIG_INVALID;

	tns_dev->gps_pps_ts = 0;

	fte_fifo_reset(tns_dev->fte_fifo);

	err = _nisync_reset_pxi668x(tns_dev);
	if (err)
		goto err_reset_hw;

	err = _nisync_wait_for_pll_lock(tns_dev);
	if (err)
		goto err_wait_for_pll_lock;

	err = tns_dev->hardware_clock->_reset_time(tns_dev->hardware_clock);
	if (err) {
		tns_dev->status = NISYNC_STATUS_HW_ERROR;
		goto err_reset_hardware_time;
	}

	err = tns_dev->hardware_clock->_set_adj_rate(tns_dev->hardware_clock, 0);
	if (err) {
		/* HACK: Check if EIO to set status to h/w error */
		if (-EIO == err)
			tns_dev->status = NISYNC_STATUS_HW_ERROR;
		goto err_set_adj_rate;
	}

	err = tns_dev->hardware_clock->_set_time(tns_dev->hardware_clock, 0);
	if (err) {
		tns_dev->status = NISYNC_STATUS_HW_ERROR;
		goto err_set_time;
	}

	/* Enable HW interrupts */
	spin_lock_irqsave(&tns_dev->irq_state_spinlock, irq_flags);
	irq_set_state(tns_dev, irq_state_hw_rdy);
	if (irq_test_state(tns_dev, irq_state_sw_rdy)) {
		writel(1,
			tns_dev_product_regmap(global_board_settings_enable));
		irq_set_state(tns_dev, irq_state_enabled);
	}
	spin_unlock_irqrestore(&tns_dev->irq_state_spinlock, irq_flags);

	err = _nisync_1588_start_capturing(tns_dev);
	if (err)
		goto err_1588_start_capturing;

	tns_dev->ptp_state = NISYNC_PTPD_DISABLED;

	tns_dev->status = NISYNC_STATUS_OK;

	up_write(tns_dev->dev_reset_lock);

	return 0;

err_1588_start_capturing:
err_set_time:
err_set_adj_rate:
err_reset_hardware_time:
err_wait_for_pll_lock:
err_reset_hw:
	up_write(tns_dev->dev_reset_lock);

	dev_dbg(&tns_dev->pdev->dev, "() => err: %d\n", err);
	return err;
}

#define NISYNC_OUTPUT_CLOCK_ENABLE_BIT_OFFSET	30
#define NISYNC_OUTPUT_ENABLE_BIT_OFFSET		29
#define NISYNC_INPUT_CAPTURE_BIT_OFFSET		28
#define NISYNC_INPUT_CAPTURE_EVENT_BIT_OFFSET	24
#define NISYNC_ROUTE_ENABLE_BIT_OFFSET		22
#define NISYNC_INVERT_ENABLE_BIT_OFFSET		21
#define NISYNC_SYNCHRONIZE_BIT_OFFSET		20
#define NISYNC_SOURCE_BIT_OFFSET		8

#define NISYNC_OUTPUT_CLOCK_ENABLE_BIT					\
				(BIT(NISYNC_OUTPUT_CLOCK_ENABLE_BIT_OFFSET))
#define NISYNC_OUTPUT_ENABLE_BIT	(BIT(NISYNC_OUTPUT_ENABLE_BIT_OFFSET))
#define NISYNC_ROUTE_ENABLE_BIT		(BIT(NISYNC_ROUTE_ENABLE_BIT_OFFSET))
#define NISYNC_INVERT_ENABLE_BIT	(BIT(NISYNC_INVERT_ENABLE_BIT_OFFSET))
#define NISYNC_SYNCHRONIZE_BIT		(BIT(NISYNC_SYNCHRONIZE_BIT_OFFSET))
#define NISYNC_SOURCE_MASK		(0x3F << NISYNC_SOURCE_BIT_OFFSET)
#define NISYNC_INPUT_CAPTURE_BIT	BIT(NISYNC_INPUT_CAPTURE_BIT_OFFSET)
#define NISYNC_ROUTE_DISABLE_MASK	(NISYNC_ROUTE_ENABLE_BIT |	\
					NISYNC_INVERT_ENABLE_BIT |	\
					NISYNC_SYNCHRONIZE_BIT |	\
					NISYNC_SOURCE_MASK)
#define NISYNC_INPUT_CAPTURE_EVENT_MASK					\
				(0x3 << NISYNC_INPUT_CAPTURE_EVENT_BIT_OFFSET)
#define NISYNC_CONTROL_RESET_MASK	(NISYNC_SYNCHRONIZE_BIT |		\
					NISYNC_INVERT_ENABLE_BIT |		\
					NISYNC_ROUTE_ENABLE_BIT |		\
					NISYNC_INPUT_CAPTURE_EVENT_MASK |	\
					NISYNC_INPUT_CAPTURE_BIT |		\
					NISYNC_OUTPUT_ENABLE_BIT |		\
					NISYNC_OUTPUT_CLOCK_ENABLE_BIT_OFFSET)

/* trigger routing functions */
static bool nisync_is_valid_terminal(struct nisync_tns_dev *tns_dev,
				const enum nisync_terminal termid)
{
	u32 i = 0;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	while (NISYNC_TERMINAL_INVALID != tns_dev_product_terminal(i)) {
		/* This print is particularly frequent, so use vdbg */
		dev_vdbg(&tns_dev->pdev->dev, "checking termid %u against %u\n",
				termid, tns_dev_product_terminal(i));
		if (termid == tns_dev_product_terminal(i))
			return true;
		++i;
	};
	return false;
}

static int nisync_validate_route_destination(
		struct nisync_tns_dev *tns_dev,
		struct nisync_trigger_connection *tconn)
{
	if (!nisync_is_valid_terminal(tns_dev, tconn->dest_terminal)
			|| NISYNC_CLK10 == tconn->dest_terminal
			|| NISYNC_GND == tconn->dest_terminal) {
		dev_dbg(&tns_dev->pdev->dev, "Invalid dest terminal %u\n",
			tconn->dest_terminal);
		return -EINVAL;
	}

	return 0;
}

static int nisync_validate_route_source(
		struct nisync_tns_dev *tns_dev,
		struct nisync_trigger_connection *tconn)
{
	if (!nisync_is_valid_terminal(tns_dev, tconn->src_terminal)) {
		dev_dbg(&tns_dev->pdev->dev, "Invalid src terminal %u\n",
			tconn->src_terminal);
		return -EINVAL;
	}

	if (tconn->dest_terminal == tconn->src_terminal) {
		dev_dbg(&tns_dev->pdev->dev, "Invalid to route back to the same terminal, src == dest == %u\n",
			    tconn->dest_terminal);
		return -EINVAL;
	}

	return 0;
}

static int nisync_get_trigger_route(struct nisync_tns_dev *tns_dev,
				struct nisync_trigger_connection *tconn)
{
	u32 regval = 0;
	int err = 0;

	dev_dbg(&tns_dev->pdev->dev, "%s(dst: %d)\n", __func__,
		tconn->dest_terminal);

	err = nisync_validate_route_destination(tns_dev, tconn);
	if (err)
		/* appropriate error message already printed by check */
		/* data in tconn remains unmodified */
		return err;

	down_read(tns_dev->dev_reset_lock);
	mutex_lock(&tns_dev->sysfs_devs[tconn->dest_terminal].term_lock);

	regval = readl(terminal_control_enable(tns_dev,
						tconn->dest_terminal));
	dev_dbg(&tns_dev->pdev->dev, "current route settings for terminal %u: 0x%08X\n",
						tconn->dest_terminal, regval);

	if (regval & NISYNC_ROUTE_ENABLE_BIT
		&& regval & NISYNC_OUTPUT_ENABLE_BIT) {
		tconn->src_terminal = (regval & NISYNC_SOURCE_MASK)
					>> NISYNC_SOURCE_BIT_OFFSET;
		tconn->synchronous = (regval & NISYNC_SYNCHRONIZE_BIT)
					>> NISYNC_SYNCHRONIZE_BIT_OFFSET;
		tconn->invert = (regval & NISYNC_INVERT_ENABLE_BIT)
					>> NISYNC_INVERT_ENABLE_BIT_OFFSET;
	} else {
		tconn->src_terminal = NISYNC_TERMINAL_INVALID;
		tconn->synchronous = NISYNC_ASYNCHRONOUS;
		tconn->invert = NISYNC_DONT_INVERT;
	}

	mutex_unlock(&tns_dev->sysfs_devs[tconn->dest_terminal].term_lock);
	up_read(tns_dev->dev_reset_lock);

	dev_dbg(&tns_dev->pdev->dev, "(src: %d, dst: %d, sync: %d, inv: %d) =>\n",
			tconn->src_terminal, tconn->dest_terminal,
			tconn->synchronous, tconn->invert);

	return 0;
}

static int nisync_disconnect_trigger_route(struct nisync_tns_dev *tns_dev,
				struct nisync_trigger_connection *tconn)
{
	u32 regval = 0;
	int err = 0;

	dev_dbg(&tns_dev->pdev->dev, "%s(src: %d, dst: %d, sync: %d, inv: %d)\n",
			__func__, tconn->src_terminal, tconn->dest_terminal,
			tconn->synchronous, tconn->invert);

	err = nisync_validate_route_destination(tns_dev, tconn);
	if (err)
		/* appropriate error message already printed by check */
		return err;

	err = nisync_validate_route_source(tns_dev, tconn);
	if (err)
		/* appropriate error message already printed by check */
		return err;

	down_read(tns_dev->dev_reset_lock);
	mutex_lock(&tns_dev->sysfs_devs[tconn->dest_terminal].term_lock);

	regval = readl(terminal_control_enable(tns_dev,
						tconn->dest_terminal));

	if (tconn->src_terminal != (regval & NISYNC_SOURCE_MASK) >> 8) {
		dev_dbg(&tns_dev->pdev->dev, "route from %u to %u doesn't currently exist\n",
			tconn->src_terminal, tconn->dest_terminal);
		err = -EINVAL;
		goto err_route_doesnt_exist;
	}

	/*
	 * only disconnect the route if the RouteEnable bit is set;
	 * otherwise, we would be undoing a FTE
	 */
	if (regval & NISYNC_ROUTE_ENABLE_BIT) {
		writel(NISYNC_OUTPUT_ENABLE_BIT,
				terminal_control_disable(tns_dev,
						tconn->dest_terminal));
		writel(NISYNC_ROUTE_DISABLE_MASK,
				terminal_control_disable(tns_dev,
						tconn->dest_terminal));
		dev_dbg(&tns_dev->pdev->dev, "disabling route for terminal %u\n",
				tconn->dest_terminal);
	} else {
		/*
		 * We prevent routing to a terminal with FTE enabled, and
		 * we prevent enabling FTE on a terminal that is currently
		 * routed to. We then also check if a route even exists
		 * before getting to this point. So we should never get here.
		 */
		dev_dbg(&tns_dev->pdev->dev, "invalid state; fte enabled and routing to the same terminal!\n");
		err = -EBUSY;
	}

err_route_doesnt_exist:
	mutex_unlock(&tns_dev->sysfs_devs[tconn->dest_terminal].term_lock);
	up_read(tns_dev->dev_reset_lock);

	return err;
}

static int nisync_connect_trigger_route(struct nisync_tns_dev *tns_dev,
				struct nisync_trigger_connection *tconn)
{
	u32 newRouteSettings = 0x00000000;
	int err = 0;
	u32 regval = 0;
	struct nisync_sysfs_dev *dest_sysfs_dev;

	dev_dbg(&tns_dev->pdev->dev, "%s(src: %d, dst: %d, sync: %d, inv: %d)\n",
			__func__, tconn->src_terminal, tconn->dest_terminal,
			tconn->synchronous, tconn->invert);
	if (tconn->synchronous != NISYNC_ASYNCHRONOUS
			&& tconn->synchronous != NISYNC_SYNCHRONOUS) {
		dev_dbg(&tns_dev->pdev->dev, "invalid synchronous flag %d; defaulting to NISYNC_ASYNCHRONOUS\n",
			    tconn->synchronous);
		tconn->synchronous = NISYNC_ASYNCHRONOUS;
	}

	if (tconn->invert != NISYNC_DONT_INVERT
			&& tconn->invert != NISYNC_INVERT) {
		dev_dbg(&tns_dev->pdev->dev, "invalid inversion flag %d; defaulting to NISYNC_DONT_INVERT\n",
			    tconn->invert);
		tconn->invert = NISYNC_DONT_INVERT;
	}

	err = nisync_validate_route_destination(tns_dev, tconn);
	if (err)
		/* appropriate error message already printed by check */
		return err;

	err = nisync_validate_route_source(tns_dev, tconn);
	if (err)
		/* appropriate error message already printed by check */
		return err;

	if (NISYNC_CLK10 == tconn->src_terminal
			&& NISYNC_SYNCHRONOUS == tconn->synchronous) {
		dev_dbg(&tns_dev->pdev->dev, "Undefined behavior when routing from CLK10 synchronously\n");
		return -EINVAL;
	}

	down_read(tns_dev->dev_reset_lock);
	mutex_lock(&tns_dev->sysfs_devs[tconn->dest_terminal].term_lock);

	dest_sysfs_dev = &tns_dev->sysfs_devs[tconn->dest_terminal];
	if (terminal_test_state(dest_sysfs_dev, nisync_term_fte_enabled)) {
		dev_dbg(&tns_dev->pdev->dev, "destination terminal (%u) already in use for FTEs\n",
					tconn->dest_terminal);
		err = -EBUSY;
		goto unlock_terminal;
	}

	if (terminal_test_state(dest_sysfs_dev, nisync_term_irig_enabled)) {
		dev_dbg(&tns_dev->pdev->dev, "destination terminal (%u) already in use for IRIG\n",
					tconn->dest_terminal);
		err = -EBUSY;
		goto unlock_terminal;
	}

	regval = readl(terminal_control_enable(tns_dev, tconn->dest_terminal));

	if ((regval & NISYNC_ROUTE_ENABLE_BIT)) {
		dev_dbg(&tns_dev->pdev->dev, "destination terminal (%u) already in use for routing\n",
					tconn->dest_terminal);
		err = -EBUSY;
		goto unlock_terminal;
	}

	if (NISYNC_SYNCHRONOUS == tconn->synchronous)
		newRouteSettings |= NISYNC_SYNCHRONIZE_BIT;

	if (NISYNC_INVERT == tconn->invert)
		newRouteSettings |= NISYNC_INVERT_ENABLE_BIT;

	newRouteSettings |= (tconn->src_terminal << 8) & NISYNC_SOURCE_MASK;
	newRouteSettings |= NISYNC_ROUTE_ENABLE_BIT;

	/*
	 * Hardware docs say to set route settings first,
	 * then set OutputEnable
	 */
	dev_dbg(&tns_dev->pdev->dev, "newRouteSettings for dest %u: 0x%08X\n",
				tconn->dest_terminal, newRouteSettings);
	writel(newRouteSettings,
			terminal_control_enable(tns_dev, tconn->dest_terminal));
	writel(NISYNC_OUTPUT_ENABLE_BIT,
			terminal_control_enable(tns_dev, tconn->dest_terminal));

unlock_terminal:
	mutex_unlock(&tns_dev->sysfs_devs[tconn->dest_terminal].term_lock);
	up_read(tns_dev->dev_reset_lock);

	return err;
}

/* clock routing functions */
const u32 NISYNC_CLKOUT_SOURCE_IS_BACKPLANE	= BIT(13);
const u32 NISYNC_CLKOUT_SOURCE_IS_BOARD_CLK	= BIT(12);
const u32 NISYNC_BACKPLANE_SOURCE_IS_BOARD_CLK	= BIT(9);
const u32 NISYNC_BACKPLANE_SOURCE_IS_CLKIN	= BIT(8);
const u32 NISYNC_CLOCK_PLL_RESET		= BIT(4);
const u32 NISYNC_TIMEKEEPER_SOURCE		= BIT(0);

/*
 * nisync_get_clock_route expects the caller to set cconn->dest_clock;
 * this function will fill in cconn->src_clock
 */
static int nisync_get_clock_route(struct nisync_tns_dev *tns_dev,
				struct nisync_clock_connection *cconn)
{
	int err = 0;
	u32 regval = 0;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	down_read(tns_dev->dev_reset_lock);
	mutex_lock(tns_dev->dev_lock);

	regval = readl(tns_dev_product_regmap(clocking_settings_enable));
	dev_dbg(&tns_dev->pdev->dev, "read ClockingSettings: 0x%08X\n", regval);

	switch (cconn->dest_clock) {
	case NISYNC_CLK_NONE:
	case NISYNC_BOARD_CLK:
	case NISYNC_CLKIN:
		dev_dbg(&tns_dev->pdev->dev, "dest_clock (%u) is output-only; it never has a source\n",
			    cconn->dest_clock);
		err = -EINVAL;
		cconn->src_clock = NISYNC_CLK_NONE;
		break;
	case NISYNC_BACKPLANE_CLK:
		if (regval & NISYNC_BACKPLANE_SOURCE_IS_BOARD_CLK)
			cconn->src_clock = NISYNC_BOARD_CLK;
		else if (regval & NISYNC_BACKPLANE_SOURCE_IS_CLKIN)
			cconn->src_clock = NISYNC_CLKIN;
		else
			cconn->src_clock = NISYNC_CLK_NONE;
		break;
	case NISYNC_TIMEKEEPER_CLK:
		if (regval & NISYNC_TIMEKEEPER_SOURCE)
			cconn->src_clock = NISYNC_BACKPLANE_CLK;
		else
			cconn->src_clock = NISYNC_BOARD_CLK;
		break;
	case NISYNC_CLKOUT:
		if (regval & NISYNC_CLKOUT_SOURCE_IS_BACKPLANE)
			cconn->src_clock = NISYNC_BACKPLANE_CLK;
		else if (regval & NISYNC_CLKOUT_SOURCE_IS_BOARD_CLK)
			cconn->src_clock = NISYNC_BOARD_CLK;
		else
			cconn->src_clock = NISYNC_CLK_NONE;
		break;
	default:
		dev_dbg(&tns_dev->pdev->dev, "dest clock type %u invalid\n",
			cconn->dest_clock);
		err = -EINVAL;
		break;
	}

	mutex_unlock(tns_dev->dev_lock);
	up_read(tns_dev->dev_reset_lock);

	return err;
}

#define NISYNC_FORCE_BOARD_PLL_RESET BIT(4)

static int nisync_connect_clock_route(struct nisync_tns_dev *tns_dev,
				struct nisync_clock_connection *cconn)
{
	int err = 0;
	bool board_is_running_from_backplane = false;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	down_read(tns_dev->dev_reset_lock);
	mutex_lock(tns_dev->dev_lock);

	switch (cconn->dest_clock) {
	case NISYNC_CLK_NONE:
	case NISYNC_BOARD_CLK:
	case NISYNC_CLKIN:
		dev_dbg(&tns_dev->pdev->dev, "dest_clock (%u) is output-only; it never has a source\n",
			    cconn->dest_clock);
		err = -EINVAL;
		break;
	case NISYNC_BACKPLANE_CLK:
		board_is_running_from_backplane = NISYNC_TIMEKEEPER_SOURCE
				& readl(tns_dev_product_regmap(
						clocking_settings_enable));
		if (NISYNC_CLK_NONE == cconn->src_clock) {
			writel(NISYNC_BACKPLANE_SOURCE_IS_BOARD_CLK
				| NISYNC_BACKPLANE_SOURCE_IS_CLKIN,
				tns_dev_product_regmap(
					clocking_settings_disable));
		} else if (NISYNC_CLKIN == cconn->src_clock &&
				tns_dev_product_feature_supported(clkin)) {
			if (board_is_running_from_backplane)
				writel(NISYNC_FORCE_BOARD_PLL_RESET,
					tns_dev_product_regmap(
						clocking_settings_enable));

			writel(NISYNC_BACKPLANE_SOURCE_IS_BOARD_CLK,
					tns_dev_product_regmap(
						clocking_settings_disable));
			writel(NISYNC_BACKPLANE_SOURCE_IS_CLKIN,
					tns_dev_product_regmap(
						clocking_settings_enable));

			if (board_is_running_from_backplane) {
				writel(NISYNC_FORCE_BOARD_PLL_RESET,
					tns_dev_product_regmap(
						clocking_settings_disable));
				err = _nisync_wait_for_pll_lock(tns_dev);
			}
		} else if (NISYNC_BOARD_CLK == cconn->src_clock) {
			if (board_is_running_from_backplane)
				writel(NISYNC_FORCE_BOARD_PLL_RESET,
					tns_dev_product_regmap(
						clocking_settings_enable));

			writel(NISYNC_BACKPLANE_SOURCE_IS_CLKIN,
					tns_dev_product_regmap(
						clocking_settings_disable));
			writel(NISYNC_BACKPLANE_SOURCE_IS_BOARD_CLK,
					tns_dev_product_regmap(
						clocking_settings_enable));

			if (board_is_running_from_backplane) {
				writel(NISYNC_FORCE_BOARD_PLL_RESET,
					tns_dev_product_regmap(
						clocking_settings_disable));
				err = _nisync_wait_for_pll_lock(tns_dev);
			}
		} else {
			dev_dbg(&tns_dev->pdev->dev, "invalid src_clock %u to route to backplane_clk\n",
				    cconn->src_clock);
			err = -EINVAL;
		}
		break;
	case NISYNC_TIMEKEEPER_CLK:
		/* TODO: avoid glitches if src is already routed to dest? */
		if (NISYNC_BOARD_CLK == cconn->src_clock) {
			writel(NISYNC_FORCE_BOARD_PLL_RESET,
					tns_dev_product_regmap(
						clocking_settings_enable));
			writel(NISYNC_TIMEKEEPER_SOURCE,
					tns_dev_product_regmap(
						clocking_settings_disable));
			writel(NISYNC_FORCE_BOARD_PLL_RESET,
					tns_dev_product_regmap(
						clocking_settings_disable));
			err = _nisync_wait_for_pll_lock(tns_dev);
		} else if (NISYNC_BACKPLANE_CLK == cconn->src_clock) {
			writel(NISYNC_FORCE_BOARD_PLL_RESET,
					tns_dev_product_regmap(
						clocking_settings_enable));
			writel(NISYNC_TIMEKEEPER_SOURCE,
					tns_dev_product_regmap(
						clocking_settings_enable));
			writel(NISYNC_FORCE_BOARD_PLL_RESET,
					tns_dev_product_regmap(
						clocking_settings_disable));
			err = _nisync_wait_for_pll_lock(tns_dev);
		} else {
			dev_dbg(&tns_dev->pdev->dev, "invalid src_clock %u to route to timekeeper_clk\n",
				    cconn->src_clock);
			err = -EINVAL;
		}
		break;
	case NISYNC_CLKOUT:
		if (NISYNC_CLK_NONE == cconn->src_clock) {
			writel(NISYNC_CLKOUT_SOURCE_IS_BACKPLANE
				| NISYNC_CLKOUT_SOURCE_IS_BOARD_CLK,
				tns_dev_product_regmap(
					clocking_settings_disable));
		} else if (NISYNC_BOARD_CLK == cconn->src_clock) {
			writel(NISYNC_CLKOUT_SOURCE_IS_BACKPLANE,
				tns_dev_product_regmap(
					clocking_settings_disable));
			writel(NISYNC_CLKOUT_SOURCE_IS_BOARD_CLK,
				tns_dev_product_regmap(
					clocking_settings_enable));
		} else if (NISYNC_BACKPLANE_CLK == cconn->src_clock) {
			writel(NISYNC_CLKOUT_SOURCE_IS_BOARD_CLK,
				tns_dev_product_regmap(
					clocking_settings_disable));
			writel(NISYNC_CLKOUT_SOURCE_IS_BACKPLANE,
				tns_dev_product_regmap(
					clocking_settings_enable));
		} else {
			dev_dbg(&tns_dev->pdev->dev, "invalid src_clock %u to route to clkout\n",
					cconn->src_clock);
			err = -EINVAL;
		}
		break;
	default:
		dev_dbg(&tns_dev->pdev->dev, "clock type %u invalid\n",
			cconn->dest_clock);
		err = -EINVAL;
		break;
	}

	mutex_unlock(tns_dev->dev_lock);
	up_read(tns_dev->dev_reset_lock);

	return err;
}

static void _nisync_set_1588_status_led_pxi668x(struct nisync_tns_dev *tns_dev,
				enum nisync_led_power power,
				enum nisync_led_color color)
{
	static const u8 led_power_val[] = {
		[nisync_led_power_off]		= 0x00,
		[nisync_led_power_on]		= 0x00,
		[nisync_led_power_blink_1_sec]	= 0x04,
		[nisync_led_power_blink_2_sec]	= 0x08,
	};

	static const u8 led_color_val[] = {
		[nisync_led_color_none]		= 0x00,
		[nisync_led_color_red]		= 0x01,
		[nisync_led_color_green]	= 0x02,
		[nisync_led_color_orange]	= 0x03,
	};

	dev_dbg(&tns_dev->pdev->dev, "%s(power: %d, color: %d)\n",
		__func__, power, color);

	writel(0x1, tns_dev_product_regmap(led_disable));
	if ((power != nisync_led_power_off) &&
			(color != nisync_led_color_none))
		writel(led_power_val[power] | led_color_val[color],
				tns_dev_product_regmap(led_enable));
}

static void _nisync_set_1588_status_led(struct nisync_tns_dev *tns_dev,
				enum nisync_ptpd_state ptp_state)
{
	static enum nisync_led_power ptp_state_led_power[] = {
		[NISYNC_PTPD_DISABLED]		= nisync_led_power_off,
		[NISYNC_PTPD_INITIALIZING]	= nisync_led_power_on,
		[NISYNC_PTPD_SLAVE]		= nisync_led_power_on,
		[NISYNC_PTPD_LISTENING]		= nisync_led_power_blink_2_sec,
		[NISYNC_PTPD_MASTER]		= nisync_led_power_blink_2_sec,
		[NISYNC_PTPD_FAULTY]		= nisync_led_power_on,
	};
	static enum nisync_led_color ptp_state_led_color[] = {
		[NISYNC_PTPD_DISABLED]		= nisync_led_color_none,
		[NISYNC_PTPD_INITIALIZING]	= nisync_led_color_orange,
		[NISYNC_PTPD_SLAVE]		= nisync_led_color_green,
		[NISYNC_PTPD_LISTENING]		= nisync_led_color_orange,
		[NISYNC_PTPD_MASTER]		= nisync_led_color_green,
		[NISYNC_PTPD_FAULTY]		= nisync_led_color_red,
	};

	dev_dbg(&tns_dev->pdev->dev, "%s(ptp_state = %u)\n",
		__func__, ptp_state);

	_nisync_set_1588_status_led_pxi668x(tns_dev,
			ptp_state_led_power[ptp_state],
			ptp_state_led_color[ptp_state]);
}

static int nisync_set_ptp_state(struct nisync_tns_dev *tns_dev,
				enum nisync_ptpd_state new_ptp_state)
{
	int err = 0;

	dev_dbg(&tns_dev->pdev->dev, "%s(new_ptp_state = %u)\n",
		__func__, new_ptp_state);

	if (new_ptp_state < NISYNC_PTPD_DISABLED ||
		new_ptp_state > NISYNC_PTPD_FAULTY) {
		dev_dbg(&tns_dev->pdev->dev, "invalid PTP state %u; staying in state %u\n",
				new_ptp_state, tns_dev->ptp_state);
		return -EINVAL;
	}

	down_read(tns_dev->dev_reset_lock);
	mutex_lock(tns_dev->dev_lock);

	tns_dev->ptp_state = new_ptp_state;
	_nisync_set_1588_status_led(tns_dev, new_ptp_state);

	mutex_unlock(tns_dev->dev_lock);
	up_read(tns_dev->dev_reset_lock);
	dev_dbg(&tns_dev->pdev->dev, "() => tns_dev->ptp_state = %u, err = %d\n",
			tns_dev->ptp_state, err);
	return err;
}

static int nisync_get_ptp_state(struct nisync_tns_dev *tns_dev,
				enum nisync_ptpd_state *ptp_state)
{
	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	down_read(tns_dev->dev_reset_lock);

	*ptp_state = tns_dev->ptp_state;

	up_read(tns_dev->dev_reset_lock);

	dev_dbg(&tns_dev->pdev->dev, "() => ptp_state = %u\n", *ptp_state);

	return 0;
}

static int nisync_get_device_info(struct nisync_tns_dev *tns_dev,
				struct nisync_device_info *dev_info)
{
	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	down_read(tns_dev->dev_reset_lock);
	mutex_lock(tns_dev->dev_lock);

	snprintf(dev_info->driver_version,
			FIELD_SIZEOF(struct nisync_device_info, driver_version),
			__stringify(NISYNC_DRIVER_VERSION));
	snprintf(dev_info->model,
			FIELD_SIZEOF(struct nisync_device_info, model),
			tns_dev_product_model_name());
	snprintf(dev_info->name,
			FIELD_SIZEOF(struct nisync_device_info, name),
			tns_dev->pdata->product_name);

	dev_info->serial_number = tns_dev->pdata->serial_num;
	dev_info->signature = tns_dev->signature;
	dev_info->revision = tns_dev->revision;
	dev_info->oldest_compatible_revision =
		tns_dev->oldest_compatible_revision;
	dev_info->hardware_revision = tns_dev->hardware_revision;

	mutex_unlock(tns_dev->dev_lock);
	up_read(tns_dev->dev_reset_lock);

	return 0;
}

static int __nisync_write_fte_fifo_to_hw(struct nisync_tns_dev *tns_dev,
				size_t earliest_modified_fte_index,
				size_t hw_fifo_size) {
	fte_fifo_itr_t itr;
	fte_fifo_itr_t earliest_modified_fte = NULL;
	size_t i;
	u64 event_time;
	int err = 0;

	dev_dbg(&tns_dev->pdev->dev, "%s(hw_fifo_size = %zu, earliest_modified_fte_index = %zu)\n",
			__func__, hw_fifo_size, earliest_modified_fte_index);

	/*
	 * Clear from HW FIFO any events that have been modified in the SW FIFO
	 */
	if (0 == earliest_modified_fte_index) {
		dev_dbg(&tns_dev->pdev->dev, "complete clear of HW FTE FIFO\n");
		writel(BIT(31), tns_dev_product_regmap(
					future_time_buffer_control_enable));
	} else if (earliest_modified_fte_index < hw_fifo_size) {
		size_t trailing_ftes_to_clear = hw_fifo_size -
						earliest_modified_fte_index;
		dev_dbg(&tns_dev->pdev->dev, "partial clear of %zu events at end of HW FTE FIFO\n",
			trailing_ftes_to_clear);
		writel(trailing_ftes_to_clear, tns_dev_product_regmap(
					future_time_partial_clear));
		if (nisync_spin_event_timeout((readl(tns_dev_product_regmap(
			future_time_partial_clear_status)) & 0x3) == 0x3,
			NSEC_PER_SEC)) {
			dev_err(&tns_dev->pdev->dev, "Failed to re-order FTEs in HW FIFO.\n");
			tns_dev->status = NISYNC_STATUS_HW_ERROR;
			return -ETIMEDOUT;
		}
	} else if (earliest_modified_fte_index == hw_fifo_size) {
		/* append to existing fifo; no FTEs to clear */
	}

	/*
	 * Update HW FIFO with the new events, starting at
	 * earliest_modified_fte_index
	 */
	for (itr = fte_fifo_begin(tns_dev->fte_fifo), i = 0;
			itr != fte_fifo_end(tns_dev->fte_fifo);
			itr = fte_fifo_next(tns_dev->fte_fifo, itr), i++) {
		/*
		 * cache the first event to use below when checking for late
		 * events
		 */
		if (i == earliest_modified_fte_index)
			earliest_modified_fte = itr;

		if (i >= earliest_modified_fte_index) {
			u64 linestate;

			fte_fifo_peek(tns_dev->fte_fifo, itr, &event_time, NULL,
					&linestate);
			_write_fte_fifo_pxi668x(tns_dev,
						 event_time, linestate);
		}
	}

	if (earliest_modified_fte_index >= fte_fifo_size(tns_dev->fte_fifo)) {
		/*
		 * We didn't add anything to the queue, we only deleted events
		 * at the end of the queue. All of the events remaining in the
		 * queue were there before, so none of them were written late.
		 */
	} else {
		/*
		 * We had to overwrite (or write new) events in the hardware
		 * queue. Since the FTE safety margin isn't a 100% guarantee,
		 * check whether it is possible that some of those events were
		 * written late.
		 */
		u64 finish_time;

		tns_dev->hardware_clock->_get_time(tns_dev->hardware_clock,
						   &finish_time);

		itr = earliest_modified_fte;
		fte_fifo_peek(tns_dev->fte_fifo, itr, &event_time, NULL, NULL);
		/*
		 * When event_time == finish_time, software is guaranteed to
		 * have written this FTE before calling _get_time, so no
		 * need to report an error.
		 */
		while (event_time < finish_time) {
			if (NISYNC_TIME_IMMEDIATE_NANOS != event_time) {
				dev_err(&tns_dev->pdev->dev, "FTE at %lluns may have occurred later than requested\n",
						event_time);
				err = -EIO;
			}
			itr = fte_fifo_next(tns_dev->fte_fifo, itr);
			if (itr == fte_fifo_end(tns_dev->fte_fifo))
				break;
			fte_fifo_peek(tns_dev->fte_fifo, itr, &event_time, NULL,
					NULL);
		}
	}

	return err;
}

static int nisync_terminal_abort_all_ftes(struct nisync_tns_dev *tns_dev,
				struct nisync_sysfs_dev *sysfs_dev)
{
	int err = 0;
	size_t hw_fifo_size;
	size_t earliest_modified_fte_index = fte_fifo_size(tns_dev->fte_fifo);
	u64 curr_time;
	const u64 term_bitmask = terminal_time_event_bit(sysfs_dev->term);

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	down_read(tns_dev->dev_reset_lock);
	mutex_lock(&tns_dev->fte_lock);
	get_cpu();

	tns_dev->hardware_clock->_get_time(tns_dev->hardware_clock,
					   &curr_time);

	/*
	 * First synchronize FIFO cache with onboard FIFO by removing
	 * FTEs that have already been executed.
	 */
	fte_fifo_del_older_than(tns_dev->fte_fifo, curr_time);

	hw_fifo_size = fte_fifo_size(tns_dev->fte_fifo);

	err = fte_fifo_clear_terminal(tns_dev->fte_fifo, term_bitmask,
				      curr_time, &earliest_modified_fte_index);
	if (err) {
		dev_dbg(&tns_dev->pdev->dev, "Failed to clear FTEs for this terminal\n");
		goto err_fte_fifo_clear_terminal;
	}

	err = __nisync_write_fte_fifo_to_hw(tns_dev,
				earliest_modified_fte_index, hw_fifo_size);

err_fte_fifo_clear_terminal:
	put_cpu();
	mutex_unlock(&tns_dev->fte_lock);
	up_read(tns_dev->dev_reset_lock);

	return err;
}

static int nisync_abort_all_ftes(struct nisync_tns_dev *tns_dev)
{
	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	down_read(tns_dev->dev_reset_lock);
	mutex_lock(&tns_dev->fte_lock);

	/* Clear HW FTE FIFO */
	writel(BIT(31),
		tns_dev_product_regmap(future_time_buffer_control_enable));

	/* Clear FTEs from SW FIFO */
	fte_fifo_clear(tns_dev->fte_fifo);

	mutex_unlock(&tns_dev->fte_lock);
	up_read(tns_dev->dev_reset_lock);

	return 0;
}

static int nisync_count_pending_ftes(struct nisync_tns_dev *tns_dev,
				struct nisync_sysfs_dev *sysfs_dev,
				u32 *pending_ftes)
{
	u64 curr_time;
	/* -1 = sentinel for FTEs on all terminals */
	const u64 term_bitmask = sysfs_dev ?
			terminal_time_event_bit(sysfs_dev->term) :
			(u64) -1;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	down_read(tns_dev->dev_reset_lock);
	mutex_lock(&tns_dev->fte_lock);

	tns_dev->hardware_clock->_get_time(tns_dev->hardware_clock,
					   &curr_time);

	/*
	 * First synchronize FIFO cache with onboard FIFO by removing
	 * FTEs that have already been executed.
	 */
	/*
	 * TODO: why make tns del_older_than when it should be clear in
	 * fte_fifo calls that the caller doesn't care about events that have
	 * passed?
	 */
	fte_fifo_del_older_than(tns_dev->fte_fifo, curr_time);
	fte_fifo_terminal_events(tns_dev->fte_fifo, term_bitmask, pending_ftes);

	mutex_unlock(&tns_dev->fte_lock);
	up_read(tns_dev->dev_reset_lock);

	return 0;
}

static int nisync_get_fte_queue_size(struct nisync_tns_dev *tns_dev,
				u32 *fte_queue_size)
{
	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	*fte_queue_size = fte_fifo_capacity(tns_dev->fte_fifo);

	dev_dbg(&tns_dev->pdev->dev, "() => %u\n", *fte_queue_size);
	return 0;
}

static void _nisync_check_fte_levels(struct nisync_tns_dev *tns_dev,
				u64 term_bitmask,
				u32 *fte_errors)
{
	u64 curr_time;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	tns_dev->hardware_clock->_get_time(tns_dev->hardware_clock,
					   &curr_time);

	mutex_lock(&tns_dev->fte_lock);
	fte_fifo_del_older_than(tns_dev->fte_fifo, curr_time);
	fte_fifo_check_errors(tns_dev->fte_fifo, term_bitmask, fte_errors);
	mutex_unlock(&tns_dev->fte_lock);
}

static int nisync_check_fte_levels(struct nisync_tns_dev *tns_dev,
				struct nisync_sysfs_dev *sysfs_dev,
				u32 *fte_errors)
{
	int err = 0;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	down_read(tns_dev->dev_reset_lock);
	mutex_lock(&sysfs_dev->term_lock);

	if (!terminal_test_state(sysfs_dev, nisync_term_fte_enabled)) {
		dev_dbg(&tns_dev->pdev->dev, "FTEs not enabled for terminal %s\n",
			terminal_name(sysfs_dev->term));
		err = -EINVAL;
		goto err_fte_not_enabled;
	}

	_nisync_check_fte_levels(tns_dev,
			terminal_time_event_bit(sysfs_dev->term), fte_errors);

err_fte_not_enabled:
	mutex_unlock(&sysfs_dev->term_lock);
	up_read(tns_dev->dev_reset_lock);
	return err;
}

#define GPS_RX_LISTEN_BIT			BIT(0)
#define GPS_RX_INT_OVER_WM_BIT			BIT(1)
#define GPS_RX_CLEAR_BIT			BIT(3)
#define GPS_RX_IDLE_INT_BIT			BIT(4)
#define GPS_TX_INT_OVER_WM_BIT			BIT(8)
#define GPS_TX_CLEAR_BIT			BIT(9)
#define GPS_PPS_TS_BIT				BIT(16)
#define GPS_RESET_RECEIVER_BIT			BIT(24)

#define GPS_ENABLE_MASK			(GPS_RX_LISTEN_BIT | \
					 GPS_RX_INT_OVER_WM_BIT | \
					 GPS_RX_CLEAR_BIT | \
					 GPS_RX_IDLE_INT_BIT | \
					 GPS_TX_CLEAR_BIT | \
					 GPS_PPS_TS_BIT)

#define GPS_DISABLE_MASK		GPS_ENABLE_MASK

static inline struct nisync_tns_dev *
to_tns_from_uart_port(struct uart_port *port) {
	return container_of(port, struct nisync_tns_dev, nisync_gps_uart_port);
}

static int nisync_gps_tx_empty_count(struct nisync_tns_dev *tns_dev)
{
	u32 data;
	data = readl(tns_dev_product_regmap(gps_fifo_status));

	return (data >> 16);
}

static unsigned int nisync_gps_uart_tx_empty(struct uart_port *port)
{
	struct nisync_tns_dev *tns_dev = to_tns_from_uart_port(port);

	return nisync_gps_tx_empty_count(tns_dev);
}

static void nisync_gps_uart_set_mctrl(struct uart_port *port,
							unsigned int mctrl)
{
	/* Not supported */
}

static unsigned int nisync_gps_uart_get_mctrl(struct uart_port *port)
{
	return 0;
}

static void nisync_gps_uart_stop_tx(struct uart_port *port)
{
	struct nisync_tns_dev *tns_dev = to_tns_from_uart_port(port);

	writel(GPS_TX_INT_OVER_WM_BIT,
				tns_dev_product_regmap(gps_control_disable));
}

static void nisync_gps_uart_start_tx(struct uart_port *port)
{
	struct nisync_tns_dev *tns_dev = to_tns_from_uart_port(port);

	writel(GPS_TX_INT_OVER_WM_BIT,
				tns_dev_product_regmap(gps_control_enable));
}

static void nisync_gps_uart_stop_rx(struct uart_port *port)
{
	struct nisync_tns_dev *tns_dev = to_tns_from_uart_port(port);

	writel(GPS_RX_LISTEN_BIT | GPS_RX_INT_OVER_WM_BIT | GPS_RX_IDLE_INT_BIT,
				tns_dev_product_regmap(gps_control_disable));
}

static void nisync_gps_uart_enable_ms(struct uart_port *port)
{
	/* Not supported */
}

static void nisync_gps_uart_break_ctl(struct uart_port *port, int break_state)
{
	/* We may not need this - gpsd doesn't use it */
}

static void nisync_gps_uart_set_termios(struct uart_port *port,
			struct ktermios *termios, struct ktermios *old)
{
	/* TODO: set parity, change baud rate, etc */
}

static int nisync_gps_uart_startup(struct uart_port *port)
{
	return 0;
}

static void nisync_gps_uart_shutdown(struct uart_port *port)
{
}

static const char *nisync_gps_uart_type(struct uart_port *port)
{
	return "NIGPS";
}

static int nisync_gps_uart_request_port(struct uart_port *port)
{
	return 0;
}

static void nisync_gps_uart_release_port(struct uart_port *port)
{
}

static struct uart_ops nisync_gps_uart_ops = {
	.tx_empty = nisync_gps_uart_tx_empty,
	.set_mctrl = nisync_gps_uart_set_mctrl,
	.get_mctrl = nisync_gps_uart_get_mctrl,
	.stop_tx = nisync_gps_uart_stop_tx,
	.start_tx = nisync_gps_uart_start_tx,
	.stop_rx = nisync_gps_uart_stop_rx,
	.enable_ms = nisync_gps_uart_enable_ms,
	.break_ctl = nisync_gps_uart_break_ctl,
	.startup = nisync_gps_uart_startup,
	.shutdown = nisync_gps_uart_shutdown,
	.set_termios = nisync_gps_uart_set_termios,
	.type = nisync_gps_uart_type,
	.request_port = nisync_gps_uart_request_port,
	.release_port = nisync_gps_uart_release_port,
};

static int nisync_gps_register_uart(struct nisync_tns_dev *tns_dev)
{
	struct uart_driver *ud = &tns_dev->nisync_gps_uart_driver;
	char *gps_dev_name;

	ud->owner = THIS_MODULE;

	gps_dev_name = devm_kasprintf(&tns_dev->pdev->dev, GFP_KERNEL, "%s.%u.gps",
					tns_dev->pdata->product_name,
					tns_dev->pdata->instance_id);
	ud->driver_name = gps_dev_name;
	ud->dev_name = gps_dev_name;
	ud->nr = 1;

	return uart_register_driver(ud);
}

static void nisync_gps_unregister_uart(struct nisync_tns_dev *tns_dev)
{
	uart_unregister_driver(&tns_dev->nisync_gps_uart_driver);
}

static int nisync_gps_register_port(struct nisync_tns_dev *tns_dev)
{
	struct uart_port *up = &tns_dev->nisync_gps_uart_port;
	struct uart_driver *ud = &tns_dev->nisync_gps_uart_driver;

	up->fifosize = tns_dev->gps_tx_buffer_size;
	up->ops = &nisync_gps_uart_ops;
	up->flags = UPF_SKIP_TEST;
	up->line = 0;
	/* FIXME: Not sure what value to put here */
	up->type = 124;
	up->dev = &tns_dev->pdev->dev;

	return uart_add_one_port(ud, up);
}

static void nisync_gps_unregister_port(struct nisync_tns_dev *tns_dev)
{
	uart_remove_one_port(&tns_dev->nisync_gps_uart_driver,
					&tns_dev->nisync_gps_uart_port);
}

static void nisync_gps_enable(struct nisync_tns_dev *tns_dev)
{
	down_read(tns_dev->dev_reset_lock);
	mutex_lock(tns_dev->dev_lock);

	/* Pulse the reset bit */
	writel(GPS_RESET_RECEIVER_BIT,
				tns_dev_product_regmap(gps_control_disable));
	usleep_range(1000, 2000);
	writel(GPS_RESET_RECEIVER_BIT,
				tns_dev_product_regmap(gps_control_enable));
	usleep_range(1000, 2000);
	writel(GPS_RESET_RECEIVER_BIT,
				tns_dev_product_regmap(gps_control_disable));

	/* Watermark values as used in the Windows driver. */
	writel(8, tns_dev_product_regmap(gps_rx_buffer_watermark));
	writel(1, tns_dev_product_regmap(gps_tx_buffer_watermark));

	writel(GPS_ENABLE_MASK, tns_dev_product_regmap(gps_control_enable));

	mutex_unlock(tns_dev->dev_lock);
	up_read(tns_dev->dev_reset_lock);
}

static void nisync_gps_disable(struct nisync_tns_dev *tns_dev)
{
	down_read(tns_dev->dev_reset_lock);
	mutex_lock(tns_dev->dev_lock);

	writel(GPS_DISABLE_MASK, tns_dev_product_regmap(gps_control_disable));

	mutex_unlock(tns_dev->dev_lock);
	up_read(tns_dev->dev_reset_lock);
}

static void nisync_gps_get_ts(struct nisync_tns_dev *tns_dev, u64 *ts)
{
	unsigned int seq;

	down_read(tns_dev->dev_reset_lock);

	do {
		seq = read_seqbegin(&tns_dev->gps_seqlock);
		*ts = tns_dev->gps_pps_ts;
	} while (read_seqretry(&tns_dev->gps_seqlock, seq));

	up_read(tns_dev->dev_reset_lock);
}

#define ANTENNA_SHORTED_BIT	BIT(4)
#define ANTENNA_OPEN_BIT	BIT(0)
static void nisync_gps_is_antenna_connected(struct nisync_tns_dev *tns_dev, bool *status)
{
	u32 regval;

	down_read(tns_dev->dev_reset_lock);
	mutex_lock(tns_dev->dev_lock);

	regval = readl(tns_dev_product_regmap(gps_antenna));

	mutex_unlock(tns_dev->dev_lock);

	*status = !((regval & ANTENNA_OPEN_BIT) || (regval & ANTENNA_SHORTED_BIT));

	up_read(tns_dev->dev_reset_lock);
}

static void nisync_check_all_fte_levels(struct nisync_tns_dev *tns_dev,
					u32 *fte_errors)
{
	struct nisync_sysfs_dev *sysfs_dev;
	int i = 0;
	u32 term;
	u64 term_bitmask = 0;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	down_read(tns_dev->dev_reset_lock);

	while ((term = tns_dev_product_terminal(i)) != NISYNC_LAST_TERMINAL) {
		sysfs_dev = &tns_dev->sysfs_devs[term];
		mutex_lock(&sysfs_dev->term_lock);
		if (terminal_test_state(sysfs_dev, nisync_term_fte_enabled))
			term_bitmask |= terminal_time_event_bit(term);
		i++;
	}

	_nisync_check_fte_levels(tns_dev, term_bitmask, fte_errors);

	while (--i >= 0) {
		term = tns_dev_product_terminal(i);
		sysfs_dev = &tns_dev->sysfs_devs[term];
		mutex_unlock(&sysfs_dev->term_lock);
	}

	up_read(tns_dev->dev_reset_lock);
}

static void nisync_get_terminal_status(struct nisync_tns_dev *tns_dev,
				struct nisync_sysfs_dev *sysfs_dev,
				enum nisync_status *status)
{
	enum nisync_status _status;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	down_read(tns_dev->dev_reset_lock);
	/*
	 * tns_dev->status is only directly assigned to something other than
	 * NISYNC_STATUS_OK when there is something terribly wrong with the
	 * device.
	 */
	if (NISYNC_STATUS_OK != tns_dev->status) {
		*status = tns_dev->status;
		goto dev_reset_unlock;
	}
	mutex_lock(&sysfs_dev->term_lock);

	_status = NISYNC_STATUS_OK;
	if (terminal_test_state(sysfs_dev, nisync_term_ts_enabled)) {
		/*
		 * HW overflow is higher priority, so return that status even
		 * if both HW and SW overflow bits are set
		 */
		if (terminal_test_state(sysfs_dev, nisync_term_ts_hw_overflow))
			_status = NISYNC_STATUS_HW_FIFO_OVERFLOW;
		else if (terminal_test_state(sysfs_dev,
						nisync_term_ts_sw_overflow))
			_status = NISYNC_STATUS_DRV_BUFFER_OVERFLOW;
	}
	*status = _status;

	mutex_unlock(&sysfs_dev->term_lock);
dev_reset_unlock:
	up_read(tns_dev->dev_reset_lock);
}

static int nisync_get_attribute(struct nisync_tns_dev *tns_dev,
				enum nisync_attribute attr, u32 *val)
{
	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	switch (attr) {
	case NISYNC_FTE_LATENCY:
		*val = tns_dev->future_time_latency;
		break;

	case NISYNC_INPUT_TIMESTAMP_LATENCY:
		*val = tns_dev->input_timestamp_latency;
		break;

	case NISYNC_SYS_TIME_LATENCY:
		*val = tns_dev->system_time_latency;
		break;

	case NISYNC_ENET_RECEIVE_LATENCY:
		{
			u32 eth_link_speed;

			eth_link_speed = (readl(
				tns_dev_product_regmap(board_status)) >> 5)
				& 0x3;

			switch (eth_link_speed) {
			case eth_link_speed_10:
				*val = tns_dev->eth_rx_latency_10mbps;
				break;
			case eth_link_speed_100:
				*val = tns_dev->eth_rx_latency_100mbps;
				break;
			case eth_link_speed_1000:
				*val = tns_dev->eth_rx_latency_1000mbps;
				break;
			default:
				dev_dbg(&tns_dev->pdev->dev, "Invalid eth_link_speed %u reported by HW\n",
						eth_link_speed);
				return -EIO;
			}
		}
		break;

	case NISYNC_ENET_TRANSMIT_LATENCY:
		{
			u32 eth_link_speed;

			eth_link_speed = (readl(
				tns_dev_product_regmap(board_status)) >> 5)
				& 0x3;

			switch (eth_link_speed) {
			case eth_link_speed_10:
				*val = tns_dev->eth_tx_latency_10mbps;
				break;
			case eth_link_speed_100:
				*val = tns_dev->eth_tx_latency_100mbps;
				break;
			case eth_link_speed_1000:
				*val = tns_dev->eth_tx_latency_1000mbps;
				break;
			default:
				dev_dbg(&tns_dev->pdev->dev, "Invalid eth_link_speed %u reported by HW\n",
						eth_link_speed);
				return -EIO;
			}
		}
		break;

	default:
		dev_dbg(&tns_dev->pdev->dev, "Invalid attribute requested: %d\n",
			attr);
		return -EINVAL;
	}

	return 0;
}

static void nisync_get_status(struct nisync_tns_dev *tns_dev,
				enum nisync_status *status)
{
	enum nisync_status _status;

	down_read(tns_dev->dev_reset_lock);
	_status = tns_dev->status;

	/*
	 * If all is ok with the device, we still need to check all
	 * active terminals on the device to see if ts_hw_overflow has been
	 * set. If so, we give that status to the user so she knows she needs
	 * to check her active terminals.
	 */
	if (NISYNC_STATUS_OK == _status) {
		struct nisync_sysfs_dev *sysfs_dev;

		spin_lock_bh(&tns_dev->active_ts_terms_spinlock);
		list_for_each_entry(sysfs_dev, &tns_dev->active_ts_terms,
						list_node) {
			if (terminal_test_state(sysfs_dev,
						nisync_term_ts_hw_overflow))
				_status = NISYNC_STATUS_HW_FIFO_OVERFLOW;
		}
		spin_unlock_bh(&tns_dev->active_ts_terms_spinlock);
	}

	*status = _status;

	up_read(tns_dev->dev_reset_lock);
}

static int nisync_open(struct inode *inode, struct file *filp)
{
	int dev_index = MINOR(inode->i_cdev->dev) % NISYNC_DEVICE_MINOR_STEP;
	struct nisync_sysfs_dev *sysfs_dev =
		container_of(inode->i_cdev, struct nisync_sysfs_dev, char_dev);
	struct nisync_tns_dev *tns_dev =
		container_of_array(sysfs_dev, struct nisync_tns_dev,
				sysfs_devs, dev_index);

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	filp->private_data = tns_dev;

	return 0;
}

static int nisync_release(struct inode *inode, struct file *filp)
{
	dev_dbg(&((struct nisync_tns_dev *)filp->private_data)->pdev->dev,
		"%s (inode: %p, filp: %p) <=\n", __func__, inode, filp);

	filp->private_data = NULL;

	return 0;
}

static long nisync_ioctl(struct file *filp, unsigned int cmd,
				unsigned long arg)
{
	int err = 0;
	struct nisync_tns_dev *tns_dev = filp->private_data;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	switch (cmd) {
	case NISYNC_IOC_RESET:
		err = nisync_reset(tns_dev);
		break;
	case NISYNC_IOC_CONNECT_TRIG_TERMS:
		{
			struct nisync_trigger_connection tconn;

			err = tns_copy_from_user((void *)&tconn, (void *)arg,
								sizeof(tconn));

			if (!err)
				err = nisync_connect_trigger_route(tns_dev,
								&tconn);
		}
		break;
	case NISYNC_IOC_DISCONNECT_TRIG_TERMS:
		{
			struct nisync_trigger_connection tconn;

			err = tns_copy_from_user((void *)&tconn, (void *)arg,
								sizeof(tconn));
			if (!err)
				err = nisync_disconnect_trigger_route(tns_dev,
								&tconn);
		}
		break;
	case NISYNC_IOC_GET_TRIG_CONN:
		{
			struct nisync_trigger_connection tconn;

			err = tns_copy_from_user((void *)&tconn, (void *)arg,
								sizeof(tconn));
			if (!err)
				err = nisync_get_trigger_route(tns_dev,
								&tconn);

			if (!err)
				err = tns_copy_to_user((void *)arg,
							(void *)&tconn,
							sizeof(tconn));
		}
		break;
	case NISYNC_IOC_CONNECT_CLOCK_TERM:
		{
			struct nisync_clock_connection cconn;

			err = tns_copy_from_user((void *)&cconn, (void *)arg,
								sizeof(cconn));
			if (!err)
				err = nisync_connect_clock_route(tns_dev,
								&cconn);
		}
		break;
	case NISYNC_IOC_GET_CLOCK_CONN:
		{
			struct nisync_clock_connection cconn;

			err = tns_copy_from_user((void *)&cconn, (void *)arg,
								sizeof(cconn));
			if (!err)
				err = nisync_get_clock_route(tns_dev, &cconn);

			if (!err)
				err = tns_copy_to_user((void *)arg,
							(void *)&cconn,
							sizeof(cconn));
		}
		break;
	case NISYNC_IOC_GET_TIME:
		{
			u64 curr_time;

			err = tns_dev->hardware_clock->get_time(
					tns_dev->hardware_clock, &curr_time);
			if (!err)
				err = tns_copy_to_user((void *)arg,
							(void *)&curr_time,
							sizeof(curr_time));
		}
		break;
	case NISYNC_IOC_SET_TIME:
		{
			u64 time;

			err = tns_copy_from_user((void *)&time, (void *)arg,
							sizeof(time));
			if (!err) {
				err = tns_dev->hardware_clock->set_time(
						tns_dev->hardware_clock, time);
				if (err)
					tns_dev->status = NISYNC_STATUS_HW_ERROR;
			}
		}
		break;
	case NISYNC_IOC_ADJUST_TIME:
		{
			s64 offset;

			err = tns_copy_from_user((void *)&offset, (void *)arg,
							sizeof(offset));
			if (!err) {
				err = tns_dev->hardware_clock->adjust_time(
					tns_dev->hardware_clock, offset);
				if (err)
					tns_dev->status = NISYNC_STATUS_HW_ERROR;
			}
		}
		break;
	case NISYNC_IOC_RESET_FREQ:
		err = tns_dev->hardware_clock->set_adj_rate(
					tns_dev->hardware_clock, 0);
		/* HACK: Check if EIO to set status to h/w error */
		if (-EIO == err)
			tns_dev->status = NISYNC_STATUS_HW_ERROR;
		break;
	case NISYNC_IOC_GET_FREQ:
		{
			u32 frequency;

			err = tns_dev->hardware_clock->get_frequency(
					tns_dev->hardware_clock, &frequency);
			if (!err)
				err = tns_copy_to_user((void *)arg,
							(void *)&frequency,
							sizeof(frequency));
		}
		break;
	case NISYNC_IOC_SET_FREQ:
		{
			u32 frequency;

			err = tns_copy_from_user((void *)&frequency,
							(void *)arg,
							sizeof(frequency));
			if (!err) {
				err = tns_dev->hardware_clock->set_frequency(
					tns_dev->hardware_clock, frequency);
				/* HACK: Check if EIO to set status to h/w error */
				if (-EIO == err)
					tns_dev->status = NISYNC_STATUS_HW_ERROR;
			}
		}
		break;
	case NISYNC_IOC_SET_PTPD_STATE:
		{
			enum nisync_ptpd_state ptp_state;

			err = tns_copy_from_user((void *)&ptp_state,
							(void *)arg,
							sizeof(ptp_state));
			if (!err)
				err = nisync_set_ptp_state(tns_dev, ptp_state);
		}
		break;
	case NISYNC_IOC_GET_PTPD_STATE:
		{
			enum nisync_ptpd_state ptp_state;

			err = nisync_get_ptp_state(tns_dev, &ptp_state);
			if (!err)
				err = tns_copy_to_user((void *)arg,
							(void *)&ptp_state,
							sizeof(ptp_state));
		}
		break;
	case NISYNC_IOC_GET_ATTRIBUTE:
		{
			struct nisync_hw_attribute attribute;

			err = tns_copy_from_user((void *) &attribute,
							(void *)arg,
							sizeof(attribute));
			if (!err)
				err = nisync_get_attribute(tns_dev,
							attribute.attr,
							&attribute.val);

			if (!err)
				err = tns_copy_to_user((void *)arg,
							(void *) &attribute,
							sizeof(attribute));
		}
		break;
	case NISYNC_IOC_GET_DEVICE_INFO:
		{
			struct nisync_device_info dev_info;

			err = nisync_get_device_info(tns_dev, &dev_info);
			if (!err)
				err = tns_copy_to_user((void *)arg,
							(void *)&dev_info,
							sizeof(dev_info));
		}
		break;
	case NISYNC_IOC_ABORT_ALL_FTES:
		err = nisync_abort_all_ftes(tns_dev);
		break;
	case NISYNC_IOC_GET_NUM_PENDING_FTES:
		{
			u32 pending_ftes;

			err = nisync_count_pending_ftes(tns_dev, NULL,
							&pending_ftes);
			if (!err)
				err = tns_copy_to_user((void *)arg,
							(void *)&pending_ftes,
							sizeof(pending_ftes));
		}
		break;
	case NISYNC_IOC_GET_MAX_QUEUED_FTES:
		{
			u32 fte_queue_size;

			err = nisync_get_fte_queue_size(tns_dev,
							&fte_queue_size);
			if (!err)
				err = tns_copy_to_user((void *)arg,
						(void *)&fte_queue_size,
						sizeof(fte_queue_size));
		}
		break;
	case NISYNC_IOC_CHECK_FTE_CONSISTENCY:
		{
			u32 fte_errors;

			/*
			 * TODO: move to new pattern of always returning an int
			 * error from top level function calls in the module.
			 * Look for other occurances.
			 */
			nisync_check_all_fte_levels(tns_dev, &fte_errors);
			err = tns_copy_to_user((void *)arg,
						(void *)&fte_errors,
						sizeof(fte_errors));
		}
		break;
	case NISYNC_IOC_GET_ADJ_RATE:
		{
			s32 rate_in_ps_per_sec;
			err = tns_dev->hardware_clock->get_adj_rate(
				tns_dev->hardware_clock, &rate_in_ps_per_sec);

			if (!err)
				err = tns_copy_to_user((void *)arg,
					(void *)&rate_in_ps_per_sec,
					sizeof(rate_in_ps_per_sec));
		}
		break;
	case NISYNC_IOC_SET_ADJ_RATE:
		{
			s32 rate;

			err = tns_copy_from_user((void *)&rate, (void *)arg,
							sizeof(rate));
			if (!err) {
				err = tns_dev->hardware_clock->set_adj_rate(
						tns_dev->hardware_clock, rate);
				/* HACK: Check if EIO to set status to h/w error */
				if (-EIO == err)
					tns_dev->status = NISYNC_STATUS_HW_ERROR;
			}
		}
		break;
	case NISYNC_IOC_GET_MAX_ADJ_RATE:
		{
			s32 max_adj_rate;

			max_adj_rate = VCTCXO_MAX_ADJUSTMENT_RATE;
			err = tns_copy_to_user((void *)arg,
					(void *)&max_adj_rate,
					sizeof(max_adj_rate));
		}
		break;
	case NISYNC_IOC_GET_MIN_ADJ_RATE:
		{
			s32 min_adj_rate;

			min_adj_rate = VCTCXO_MIN_ADJUSTMENT_RATE;
			err = tns_copy_to_user((void *)arg,
					(void *)&min_adj_rate,
					sizeof(min_adj_rate));
		}
		break;
	case NISYNC_IOC_GET_STATUS:
		{
			enum nisync_status status;

			nisync_get_status(tns_dev, &status);
			err = tns_copy_to_user((void *)arg,
						(void *)&status,
						sizeof(enum nisync_status));
		}
		break;
	case NISYNC_IOC_GPS_ENABLE:
		nisync_gps_enable(tns_dev);
		break;
	case NISYNC_IOC_GPS_DISABLE:
		nisync_gps_disable(tns_dev);
		break;
	case NISYNC_IOC_GPS_GET_TS:
		{
			u64 ts;
			nisync_gps_get_ts(tns_dev, &ts);

			err = tns_copy_to_user((void *)arg, (void *)&ts,
								sizeof(ts));
		}
		break;
	case NISYNC_IOC_GPS_IS_ANTENNA_CONNECTED:
		{
			bool status;
			nisync_gps_is_antenna_connected(tns_dev, &status);

			err = tns_copy_to_user((void *)arg, (void *)&status,
								sizeof(status));
		}
		break;

	default:
		dev_dbg(&tns_dev->pdev->dev, "Received invalid ioctl\n");
		tns_dev_dbg_ioctl(cmd);
		err = -ENOTTY;
	}

	dev_dbg(&tns_dev->pdev->dev, "() => err: %d\n", err);

	return err;
}

static const struct file_operations nisync_fops = {
	.owner		= THIS_MODULE,
	.open		= nisync_open,
	.release	= nisync_release,
	.unlocked_ioctl	= nisync_ioctl
};

#define NISYNC_INPUT_TIMESTAMP_BUFFER_CLEAR_BIT BIT(30)
#define NISYNC_GLOBAL_INPUT_TIMESTAMPS_BIT BIT(24)
#define NISYNC_INPUT_TIMESTAMP_BUFFER_INT_OVER_WATERMARK_BIT BIT(1)

static int nisync_enable_timestamp_trigger(struct nisync_tns_dev *tns_dev,
				struct nisync_sysfs_dev *sysfs_dev,
				struct nisync_timestamp_trigger *ts_conf)
{
	int err;

	dev_dbg(&tns_dev->pdev->dev, "%s(terminal %s)\n", __func__,
		terminal_name(sysfs_dev->term));
	dev_dbg(&tns_dev->pdev->dev, "  Active Edge: %s, Decimation Count: %u\n",
		ts_conf->active_edge == NISYNC_EDGE_RISING ? "Rising" :
		ts_conf->active_edge == NISYNC_EDGE_FALLING ? "Falling" :
		ts_conf->active_edge == NISYNC_EDGE_ANY ? "Any" :
		"INVALID", ts_conf->decimation_count);

	if (NISYNC_EDGE_INVALID <= ts_conf->active_edge) {
		dev_dbg(&tns_dev->pdev->dev, "Invalid active edge %u\n",
			ts_conf->active_edge);
		return -EINVAL;
	}

	if ((0 == ts_conf->decimation_count) ||
			(PXI668x_MAX_DECIMATION_COUNT <
			ts_conf->decimation_count)) {
		dev_dbg(&tns_dev->pdev->dev, "decimation_count %u out of range; max: %u\n",
			ts_conf->decimation_count,
			PXI668x_MAX_DECIMATION_COUNT);

		return -EINVAL;
	}

	down_read(tns_dev->dev_reset_lock);
	mutex_lock(&sysfs_dev->term_lock);

	if (terminal_test_state(sysfs_dev, nisync_term_ts_enabled)) {
		dev_dbg(&tns_dev->pdev->dev,
			"Timestamping is already enabled for terminal %s.\n",
			terminal_name(sysfs_dev->term));

		err = -EBUSY;
		goto err_ts_already_enabled;
	}

	if (terminal_test_state(sysfs_dev, nisync_term_irig_enabled)) {
		dev_dbg(&tns_dev->pdev->dev,
			"Can't enable timestamping for terminal %s because Irig is enabled\n",
			terminal_name(sysfs_dev->term));

		err = -EBUSY;
		goto err_irig_already_enabled;
	}

	terminal_clear_state(sysfs_dev, nisync_term_ts_sw_overflow);
	terminal_clear_state(sysfs_dev, nisync_term_ts_hw_overflow);
	terminal_clear_state(sysfs_dev, nisync_term_ts_aborted);

	err = nisync_buf_init(sysfs_dev->timestamps,
				struct nisync_timestamp_nanos,
				NISYNC_TRIGGER_TIMESTAMP_BUFFER_SIZE);
	if (err) {
		dev_err(&tns_dev->pdev->dev, "Error %d allocating terminal timestamp buffer.\n  Terminal: %d buffer size: %d\n",
			err, sysfs_dev->term,
			NISYNC_TRIGGER_TIMESTAMP_BUFFER_SIZE);

		goto err_cb_init;
	}

	/* Configure edge sensitivity and remember this configuration */
	writel(3 << 24, terminal_control_disable(tns_dev, sysfs_dev->term));
	writel(ts_conf->active_edge << 24,
		terminal_control_enable(tns_dev, sysfs_dev->term));
	sysfs_dev->edge_detection = ts_conf->active_edge;

	/* Prepare to receive the initial state of this terminal */
	sysfs_dev->initial_state_set = false;

	writel(ts_conf->decimation_count,
		terminal_decimation_count(tns_dev, sysfs_dev->term));

	/* if HW timestamping has not been enabled */
	if (list_empty(&tns_dev->active_ts_terms)) {
		/* Clear input timestamp buffer */
		writel(NISYNC_INPUT_TIMESTAMP_BUFFER_CLEAR_BIT,
			tns_dev_product_regmap(
				input_timestamp_buffer_control_enable));

		/* Set watermark to one */
		writel(1, tns_dev_product_regmap(
				input_timestamp_buffer_watermark));

		/* Enable timestamping and over watermark interrupts */
		writel(NISYNC_GLOBAL_INPUT_TIMESTAMPS_BIT
			| NISYNC_INPUT_TIMESTAMP_BUFFER_INT_OVER_WATERMARK_BIT,
			tns_dev_product_regmap(
				input_timestamp_buffer_control_enable));
	}

	/*
	 * Add the newly allocated timestamp buffer to the active terminal
	 * timestamp list.
	 */
	spin_lock_bh(&tns_dev->active_ts_terms_spinlock);
	list_add(&sysfs_dev->list_node, &tns_dev->active_ts_terms);
	spin_unlock_bh(&tns_dev->active_ts_terms_spinlock);

	/* Enable timestamping for this terminal and update terminal state */
	writel(BIT(28), terminal_control_enable(tns_dev, sysfs_dev->term));
	terminal_set_state(sysfs_dev, nisync_term_ts_enabled);

err_cb_init:
err_irig_already_enabled:
err_ts_already_enabled:
	mutex_unlock(&sysfs_dev->term_lock);
	up_read(tns_dev->dev_reset_lock);

	return err;
}

static int _nisync_disable_timestamp_trigger(struct nisync_tns_dev *tns_dev,
				struct nisync_sysfs_dev *sysfs_dev)
{
	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	if (!terminal_test_state(sysfs_dev, nisync_term_ts_enabled)) {
		dev_dbg(&tns_dev->pdev->dev, "Can't disable ts for terminal %s that is not enabled\n",
			    terminal_name(sysfs_dev->term));
		return -EINVAL;
	}

	/* Disable HW timestamping for this terminal */
	writel(BIT(28), terminal_control_disable(tns_dev, sysfs_dev->term));

	/* Remove terminal from list of active timestamping terminals */
	spin_lock_bh(&tns_dev->active_ts_terms_spinlock);
	wake_up_all(&sysfs_dev->wait_queue);
	list_del_init(&sysfs_dev->list_node);
	spin_unlock_bh(&tns_dev->active_ts_terms_spinlock);

	nisync_buf_destroy(sysfs_dev->timestamps);

	terminal_clear_state(sysfs_dev, nisync_term_ts_enabled);

	/* if this was the only terminal being timestamped */
	if (list_empty(&tns_dev->active_ts_terms)) {
		/* Disable HW timestamping and over watermark interrupts */
		writel(NISYNC_GLOBAL_INPUT_TIMESTAMPS_BIT
			| NISYNC_INPUT_TIMESTAMP_BUFFER_INT_OVER_WATERMARK_BIT,
			tns_dev_product_regmap(
				input_timestamp_buffer_control_disable));
		writel(NISYNC_INPUT_TIMESTAMP_BUFFER_CLEAR_BIT,
			tns_dev_product_regmap(
				input_timestamp_buffer_control_enable));
	}

	return 0;
}

static int nisync_disable_timestamp_trigger(struct nisync_tns_dev *tns_dev,
				struct nisync_sysfs_dev *sysfs_dev)
{
	int err;

	dev_dbg(&tns_dev->pdev->dev, "%s(terminal %s)\n", __func__,
		terminal_name(sysfs_dev->term));

	down_read(tns_dev->dev_reset_lock);
	mutex_lock(&sysfs_dev->term_lock);

	err = _nisync_disable_timestamp_trigger(tns_dev, sysfs_dev);
	if (err)
		dev_dbg(&tns_dev->pdev->dev, "_nisync_disable_timestamp_trigger returned error %d\n",
			    err);

	mutex_unlock(&sysfs_dev->term_lock);
	up_read(tns_dev->dev_reset_lock);

	return err;
}

static int nisync_read_timestamp_trigger_conf(struct nisync_tns_dev *tns_dev,
				struct nisync_sysfs_dev *sysfs_dev,
				struct nisync_timestamp_trigger *ts_conf,
				bool *timestamp_trigger_enabled)
{
	u32 term_ts_conf;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	down_read(tns_dev->dev_reset_lock);
	mutex_lock(&sysfs_dev->term_lock);

	term_ts_conf = readl(terminal_control_enable(tns_dev,
						sysfs_dev->term));
	ts_conf->active_edge = (term_ts_conf >> 24) & 0x3;
	*timestamp_trigger_enabled = ((term_ts_conf & BIT(28)) != 0);

	ts_conf->decimation_count =
		readl(terminal_decimation_count(tns_dev,
					sysfs_dev->term));

	mutex_unlock(&sysfs_dev->term_lock);
	up_read(tns_dev->dev_reset_lock);

	return 0;
}

static int nisync_enable_future_time_events(struct nisync_tns_dev *tns_dev,
				struct nisync_sysfs_dev *sysfs_dev)
{
	int err = 0;
	u32 value;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	down_read(tns_dev->dev_reset_lock);
	mutex_lock(&sysfs_dev->term_lock);

	if (terminal_test_state(sysfs_dev, nisync_term_fte_enabled)) {
		dev_dbg(&tns_dev->pdev->dev, "FTEs are already enabled for terminal %s.\n",
			terminal_name(sysfs_dev->term));
		err = -EBUSY;
		goto err_fte_already_enabled;
	}

	value = readl(terminal_control_enable(tns_dev, sysfs_dev->term));
	if (value & NISYNC_ROUTE_ENABLE_BIT) {
		dev_dbg(&tns_dev->pdev->dev, "Can't enable FTEs for terminal %s because a route exists.\n",
			terminal_name(sysfs_dev->term));
		err = -EBUSY;
		goto err_route_enabled;
	}

	if (terminal_test_state(sysfs_dev, nisync_term_irig_enabled)) {
		dev_dbg(&tns_dev->pdev->dev, "Can't enable FTEs for terminal %s because IRIG is enabled.\n",
			terminal_name(sysfs_dev->term));
		err = -EBUSY;
		goto err_irig_enabled;
	}

	writel(NISYNC_OUTPUT_ENABLE_BIT,
			terminal_control_enable(tns_dev, sysfs_dev->term));
	terminal_set_state(sysfs_dev, nisync_term_fte_enabled);

err_fte_already_enabled:
err_route_enabled:
err_irig_enabled:
	mutex_unlock(&sysfs_dev->term_lock);
	up_read(tns_dev->dev_reset_lock);

	return err;
}

static int _nisync_disable_future_time_events(struct nisync_tns_dev *tns_dev,
				struct nisync_sysfs_dev *sysfs_dev)
{
	int err;
	size_t hw_fifo_size;
	size_t earliest_modified_fte_index = fte_fifo_size(tns_dev->fte_fifo);
	u64 curr_time;
	const u64 term_bitmask = terminal_time_event_bit(sysfs_dev->term);

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	if (!terminal_test_state(sysfs_dev, nisync_term_fte_enabled)) {
		dev_dbg(&tns_dev->pdev->dev, "FTEs not enabled for terminal %s\n",
			terminal_name(sysfs_dev->term));
		return -EINVAL;
	}

	mutex_lock(&tns_dev->fte_lock);
	get_cpu();

	tns_dev->hardware_clock->_get_time(tns_dev->hardware_clock,
					   &curr_time);

	/*
	 * First synchronize FIFO cache with onboard FIFO by removing FTEs that
	 * have already been executed.
	 */
	fte_fifo_del_older_than(tns_dev->fte_fifo, curr_time);

	hw_fifo_size = fte_fifo_size(tns_dev->fte_fifo);

	err = fte_fifo_clear_terminal(tns_dev->fte_fifo, term_bitmask,
				      curr_time, &earliest_modified_fte_index);
	if (err) {
		dev_dbg(&tns_dev->pdev->dev, "Failed to clear FTEs for this terminal\n");
		goto err_fte_fifo_clear_terminal;
	}

	terminal_clear_state(sysfs_dev, nisync_term_fte_clk_enabled);
	terminal_clear_state(sysfs_dev, nisync_term_fte_enabled);
	writel(NISYNC_OUTPUT_ENABLE_BIT | NISYNC_OUTPUT_CLOCK_ENABLE_BIT,
			terminal_control_disable(tns_dev, sysfs_dev->term));

	err = __nisync_write_fte_fifo_to_hw(tns_dev,
			earliest_modified_fte_index, hw_fifo_size);

err_fte_fifo_clear_terminal:
	put_cpu();
	mutex_unlock(&tns_dev->fte_lock);
	return err;
}

static int nisync_disable_future_time_events(struct nisync_tns_dev *tns_dev,
				struct nisync_sysfs_dev *sysfs_dev)
{
	int err;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	down_read(tns_dev->dev_reset_lock);
	mutex_lock(&sysfs_dev->term_lock);

	err = _nisync_disable_future_time_events(tns_dev, sysfs_dev);

	mutex_unlock(&sysfs_dev->term_lock);
	up_read(tns_dev->dev_reset_lock);

	return err;
}

#define IRIG_AM_DECODE_ENABLE_BIT		BIT(24)
#define IRIG_DECODE_ENABLE_CONTROL_BIT		BIT(28)

/*
 * The bit positions for IRIG_DECODE and AM_DECODE in the disable register
 * are not the  same as in the enable register
 */
#define IRIG_AM_DECODE_DISABLE_BIT		BIT(28)
#define IRIG_DECODE_DISABLE_CONTROL_BIT		BIT(24)

static int nisync_enable_irig(struct nisync_tns_dev *tns_dev,
				struct nisync_sysfs_dev *sysfs_dev,
				u8 irig_mode)
{
	u32 value;
	int err = 0;

	dev_dbg(&tns_dev->pdev->dev, "%s() mode %d\n", __func__, irig_mode);

	down_read(tns_dev->dev_reset_lock);
	mutex_lock(&sysfs_dev->term_lock);

	if (tns_dev->sysfs_devs[NISYNC_PFI0].term != sysfs_dev->term) {
		dev_dbg(&tns_dev->pdev->dev,
			"IRIG not supported on this terminal\n");
		err = -EINVAL;
		goto err_unsupported_terminal;
	}

	if (irig_mode != NISYNC_IRIGB_DC && irig_mode != NISYNC_IRIGB_AM) {
		dev_dbg(&tns_dev->pdev->dev, "invalid IRIG mode %u\n",
			irig_mode);
		err = -EINVAL;
		goto err_invalid_mode;
	}

	if (terminal_test_state(sysfs_dev, nisync_term_fte_enabled)) {
		dev_dbg(&tns_dev->pdev->dev,
			"Can't enable IRIG for terminal %s because FTEs are enabled\n",
			terminal_name(sysfs_dev->term));
		err = -EBUSY;
		goto err_term_busy;
	}

	if (terminal_test_state(sysfs_dev, nisync_term_ts_enabled)) {
		dev_dbg(&tns_dev->pdev->dev,
			"Can't enable IRIG for terminal %s because timestamping is enabled\n",
			terminal_name(sysfs_dev->term));
		err = -EBUSY;
		goto err_term_busy;
	}

	value = readl(terminal_control_enable(tns_dev, sysfs_dev->term));
	if (value & NISYNC_ROUTE_ENABLE_BIT) {
		dev_dbg(&tns_dev->pdev->dev,
			"Can't enable IRIG for terminal %s because a route exists\n",
			terminal_name(sysfs_dev->term));
		err = -EBUSY;
		goto err_term_busy;
	}

	if (terminal_test_state(sysfs_dev, nisync_term_irig_enabled)) {
		dev_dbg(&tns_dev->pdev->dev, "IRIG already enabled\n");
		err = -EBUSY;
		goto err_already_enabled;
	}

	writel(NISYNC_CONTROL_RESET_MASK,
			terminal_control_disable(tns_dev, sysfs_dev->term));

	if (irig_mode == NISYNC_IRIGB_AM)
		writel(IRIG_AM_DECODE_ENABLE_BIT,
				tns_dev_product_regmap(irig_control_enable));
	else
		writel(IRIG_AM_DECODE_DISABLE_BIT,
				tns_dev_product_regmap(irig_control_disable));

	tns_dev->irig_mode = irig_mode;
	terminal_set_state(sysfs_dev, nisync_term_irig_enabled);

	writel(IRIG_DECODE_ENABLE_CONTROL_BIT,
			tns_dev_product_regmap(irig_control_enable));

err_unsupported_terminal:
err_already_enabled:
err_term_busy:
err_invalid_mode:
	mutex_unlock(&sysfs_dev->term_lock);
	up_read(tns_dev->dev_reset_lock);

	return err;
}

static void nisync_irig_timestamp_handler(struct nisync_tns_dev *tns_dev)
{
	u32 time_and_days;
	u32 year_and_seconds;
	u32 ts_high;
	u32 ts_low;
	struct nisync_irig_packet *irig_packet = &tns_dev->irig_packet;

	time_and_days = readl(tns_dev_product_regmap(irig_timeofday));
	year_and_seconds = readl(tns_dev_product_regmap(
				irig_year_and_straight_binary_seconds));

	ts_high = readl(tns_dev_product_regmap(irig_timestamp_high));
	ts_low = readl(tns_dev_product_regmap(irig_timestamp_low));

	write_seqlock(&tns_dev->irig_seqlock);

	irig_packet->time_and_days = time_and_days;
	irig_packet->year_and_seconds = year_and_seconds;

	irig_packet->arrival_hw_ts = create_u64(ts_high, ts_low);

	/* Correct for hardware latency */
	irig_packet->arrival_hw_ts -= (tns_dev->irig_mode == NISYNC_IRIGB_AM)?
			tns_dev->irig_am_latency : tns_dev->irig_dc_latency;

	write_sequnlock(&tns_dev->irig_seqlock);
}

static int nisync_get_irig_packet(struct nisync_tns_dev *tns_dev,
				struct nisync_sysfs_dev *sysfs_dev,
				struct nisync_irig_packet *ts)
{
	int err = 0;
	unsigned int seq;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	down_read(tns_dev->dev_reset_lock);
	mutex_lock(&sysfs_dev->term_lock);

	if (!terminal_test_state(sysfs_dev, nisync_term_irig_enabled))
	{
		dev_dbg(&tns_dev->pdev->dev, "IRIG not enabled on terminal");
		err = -EINVAL;
		goto err_not_enabled;
	}

	do {
		seq = read_seqbegin(&tns_dev->irig_seqlock);
		memcpy(ts, &tns_dev->irig_packet, sizeof(*ts));
	} while (read_seqretry(&tns_dev->irig_seqlock, seq));

err_not_enabled:
	mutex_unlock(&sysfs_dev->term_lock);
	up_read(tns_dev->dev_reset_lock);

	return err;
}

static int _nisync_disable_irig(struct nisync_tns_dev *tns_dev,
				struct nisync_sysfs_dev *sysfs_dev)
{
	if (!terminal_test_state(sysfs_dev, nisync_term_irig_enabled))
	{
		dev_dbg(&tns_dev->pdev->dev, "IRIG not enabled on terminal");
		return -EINVAL;
	}

	terminal_clear_state(sysfs_dev, nisync_term_irig_enabled);
	writel(IRIG_DECODE_DISABLE_CONTROL_BIT | IRIG_AM_DECODE_DISABLE_BIT,
				tns_dev_product_regmap(irig_control_disable));

	return 0;
}

static int nisync_disable_irig(struct nisync_tns_dev *tns_dev,
				struct nisync_sysfs_dev * sysfs_dev)
{
	int err = 0;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	down_read(tns_dev->dev_reset_lock);
	mutex_lock(&sysfs_dev->term_lock);

	err = _nisync_disable_irig(tns_dev, sysfs_dev);

	mutex_unlock(&sysfs_dev->term_lock);
	up_read(tns_dev->dev_reset_lock);

	return err;
}

static int nisync_terminal_open(struct inode *inode, struct file *filp)
{
	int dev_index = MINOR(inode->i_cdev->dev) % NISYNC_DEVICE_MINOR_STEP;
	struct nisync_sysfs_dev *sysfs_dev =
		container_of(inode->i_cdev, struct nisync_sysfs_dev, char_dev);
	struct nisync_tns_dev *tns_dev =
		container_of_array(sysfs_dev, struct nisync_tns_dev,
				sysfs_devs, dev_index);
	int err = 0;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	down_read(tns_dev->dev_reset_lock);
	mutex_lock(&sysfs_dev->term_lock);

	sysfs_dev->open_count++;
	filp->private_data = sysfs_dev;

	mutex_unlock(&sysfs_dev->term_lock);
	up_read(tns_dev->dev_reset_lock);

	return err;
}

static int nisync_terminal_release(struct inode *inode, struct file *filp)
{
	struct nisync_sysfs_dev *sysfs_dev = filp->private_data;
	struct nisync_tns_dev *tns_dev =
		container_of_array(sysfs_dev, struct nisync_tns_dev,
				sysfs_devs, sysfs_dev->term);

	dev_dbg(&tns_dev->pdev->dev, "%s(inode: %p, filp: %p)\n", __func__,
		inode, filp);

	down_read(tns_dev->dev_reset_lock);
	mutex_lock(&sysfs_dev->term_lock);

	sysfs_dev->open_count--;

	if (!sysfs_dev->open_count) {
		if (terminal_test_state(sysfs_dev, nisync_term_ts_enabled))
			_nisync_disable_timestamp_trigger(tns_dev, sysfs_dev);

		if (terminal_test_state(sysfs_dev, nisync_term_irig_enabled))
			_nisync_disable_irig(tns_dev, sysfs_dev);
	}

	filp->private_data = NULL;

	mutex_unlock(&sysfs_dev->term_lock);
	up_read(tns_dev->dev_reset_lock);

	return 0;
}

static long nisync_terminal_ioctl(struct file *filp, unsigned int cmd,
				unsigned long arg)
{
	int err = 0;
	struct nisync_sysfs_dev *sysfs_dev = filp->private_data;
	struct nisync_tns_dev *tns_dev =
		container_of_array(sysfs_dev, struct nisync_tns_dev,
				sysfs_devs, sysfs_dev->term);

	dev_dbg(&tns_dev->pdev->dev, "%s(%s)\n", __func__,
		terminal_name(sysfs_dev->term));
	tns_dev_dbg_ioctl(cmd);

	switch (cmd) {
	case NISYNC_IOC_ENABLE_TIMESTAMP:
		{
			struct nisync_timestamp_trigger ts_conf;

			err = tns_copy_from_user((void *)&ts_conf, (void *)arg,
						sizeof(ts_conf));

			if (!err)
				err = nisync_enable_timestamp_trigger(tns_dev,
							sysfs_dev, &ts_conf);
		}
		break;
	case NISYNC_IOC_DISABLE_TIMESTAMP:
		err = nisync_disable_timestamp_trigger(tns_dev, sysfs_dev);
		break;
	case NISYNC_IOC_GET_TIMESTAMP_CONF:
		{
			struct nisync_timestamp_trigger ts_conf;
			bool ts_enabled;

			err = nisync_read_timestamp_trigger_conf(tns_dev,
							sysfs_dev, &ts_conf,
							&ts_enabled);

			if (!err) {
				err = tns_copy_to_user((void *)arg,
							(void *)&ts_conf,
							sizeof(ts_conf));
				if (!err)
					/*
					 * TODO: overloading return value,
					 * put in ts_conf instead?
					 */
					err = ts_enabled;
			}
		}
		break;
	case NISYNC_IOC_ENABLE_FTE:
		err = nisync_enable_future_time_events(tns_dev, sysfs_dev);
		break;
	case NISYNC_IOC_READ_FTE_ENABLED:
		err = terminal_test_state(sysfs_dev, nisync_term_fte_enabled);
		break;
	case NISYNC_IOC_ABORT_ALL_FTES:
		err = nisync_terminal_abort_all_ftes(tns_dev, sysfs_dev);
		break;
	case NISYNC_IOC_GET_NUM_PENDING_FTES:
		{
			u32 pending_ftes;

			err = nisync_count_pending_ftes(tns_dev, sysfs_dev,
							&pending_ftes);
			if (!err)
				err = tns_copy_to_user((void *)arg,
							(void *)&pending_ftes,
							sizeof(pending_ftes));
		}
		break;
	case NISYNC_IOC_DISABLE_FTE:
		err = nisync_disable_future_time_events(tns_dev, sysfs_dev);
		break;
	case NISYNC_IOC_GET_STATUS:
		{
			enum nisync_status status;

			nisync_get_terminal_status(tns_dev, sysfs_dev, &status);
			err = tns_copy_to_user((void *)arg,
						(void *)&status,
						sizeof(enum nisync_status));
		}
		break;
	case NISYNC_IOC_CHECK_FTE_CONSISTENCY:
		{
			u32 fte_errors;

			err = nisync_check_fte_levels(tns_dev, sysfs_dev,
							&fte_errors);
			if (!err)
				err = tns_copy_to_user((void *)arg,
							(void *)&fte_errors,
							sizeof(fte_errors));
		}
		break;
	case NISYNC_IOC_ENABLE_IRIG:
		{
			u8 irig_mode;
			err = tns_copy_from_user((void *)&irig_mode,
							(void *)arg,
							sizeof(irig_mode));
			if (!err)
				err = nisync_enable_irig(tns_dev, sysfs_dev,
							irig_mode);
		}
		break;
	case NISYNC_IOC_GET_IRIG_TS:
		{
			struct nisync_irig_packet irig_time;

			err = nisync_get_irig_packet(tns_dev, sysfs_dev,
							&irig_time);
			if (!err)
				err = tns_copy_to_user((void *)arg,
							(void *)&irig_time,
							sizeof(irig_time));
		}
		break;
	case NISYNC_IOC_DISABLE_IRIG:
		err = nisync_disable_irig(tns_dev, sysfs_dev);
		break;

	default:
		dev_dbg(&tns_dev->pdev->dev, "Received invalid ioctl\n");
		tns_dev_dbg_ioctl(cmd);
		err = -ENOTTY;
	}

	dev_dbg(&tns_dev->pdev->dev, "%s => err = %d\n",
		terminal_name(sysfs_dev->term), err);

	return err;
}

static int _nisync_test_and_clear_term_overflow(struct nisync_tns_dev *tns_dev,
				struct nisync_sysfs_dev *sysfs_dev)
{
	/* check HW overflow first because it is higher priority */
	if (terminal_test_state(sysfs_dev, nisync_term_ts_hw_overflow)) {
		dev_dbg(&tns_dev->pdev->dev, "Terminal %s in term_ts_hw_overflow state.\n",
					terminal_name(sysfs_dev->term));
		terminal_clear_state(sysfs_dev,
					nisync_term_ts_hw_overflow);
		return -EOVERFLOW;
	}

	if (terminal_test_state(sysfs_dev, nisync_term_ts_sw_overflow)) {
		dev_dbg(&tns_dev->pdev->dev, "Terminal %s in term_ts_sw_overflow state.\n",
					terminal_name(sysfs_dev->term));
		terminal_clear_state(sysfs_dev,
					nisync_term_ts_sw_overflow);
		return -EOVERFLOW;
	}

	return 0;
}

static ssize_t nisync_terminal_read(struct file *filp, char __user *buf,
				size_t count, loff_t *f_pos)
{
	int err;
	struct nisync_sysfs_dev *sysfs_dev = filp->private_data;
	struct nisync_tns_dev *tns_dev =
		container_of_array(sysfs_dev, struct nisync_tns_dev,
				sysfs_devs, sysfs_dev->term);

	dev_dbg(&tns_dev->pdev->dev, "%s(%s)\n", __func__,
		terminal_name(sysfs_dev->term));

	if (0 == count) {
		dev_dbg(&tns_dev->pdev->dev, "Can't read zero timestamps from terminal %s\n",
				terminal_name(sysfs_dev->term));
		return -EINVAL;
	}

	down_read(tns_dev->dev_reset_lock);
	mutex_lock(&sysfs_dev->term_lock);

	if (!terminal_test_state(sysfs_dev, nisync_term_ts_enabled)) {
		dev_dbg(&tns_dev->pdev->dev, "Input timestamping is not enabled for terminal %s.\n",
			    terminal_name(sysfs_dev->term));
		err = -EPERM;
		goto err_ts_not_enabled;
	}

	/*
	 * Note: check for and clear overflow states here and notify the user
	 * for them to decide if they still want to try and read any existing
	 * timestamps or abort.
	 * Care should be taken, because not all threads on a multi-threaded
	 * app will see the overflow state, or some other thread using the
	 * driver could clear overflow before their app sees the state.
	 *
	 * Overflow is also cleared on ts_disable/ts_enable and reset
	 */
	err = _nisync_test_and_clear_term_overflow(tns_dev, sysfs_dev);
	if (err)
		goto err_overflow;

	if (nisync_buf_empty(sysfs_dev->timestamps)) {
		if (filp->f_flags & O_NONBLOCK) {
			dev_dbg(&tns_dev->pdev->dev, "No timestamps to read in nonblocking call to terminal %s\n",
					terminal_name(sysfs_dev->term));
			err = -EAGAIN;
			goto err_try_again;
		}

		/*
		 * While waiting, allow device resets and other terminal
		 * related operations to occur.
		 */
		mutex_unlock(&sysfs_dev->term_lock);
		up_read(tns_dev->dev_reset_lock);

		err = wait_event_interruptible(sysfs_dev->wait_queue,
			terminal_test_state(sysfs_dev,
				nisync_term_ts_sw_overflow)
			|| terminal_test_state(sysfs_dev,
				nisync_term_ts_hw_overflow)
			|| terminal_test_state(sysfs_dev,
				nisync_term_ts_aborted)
			|| !terminal_test_state(sysfs_dev,
				nisync_term_ts_enabled)
			|| nisync_buf_size(sysfs_dev->timestamps));

		down_read(tns_dev->dev_reset_lock);
		mutex_lock(&sysfs_dev->term_lock);

		if (err)
			goto err_wait_event;

		if (terminal_test_state(sysfs_dev, nisync_term_ts_aborted)) {
			dev_dbg(&tns_dev->pdev->dev, "Input timestamping for terminal %s has been aborted by reset.\n",
					terminal_name(sysfs_dev->term));
			/*
			 * don't clear state so all threads will see
			 * nisync_term_ts_aborted until nisync_term_ts_enabled
			 * is set again
			 */
			err = -EINTR;
			goto err_term_ts_aborted;
		}

		/*
		 * TODO: does the distinction between
		 *	aborted and !enabled actually matter?
		 */
		if (!terminal_test_state(sysfs_dev, nisync_term_ts_enabled)) {
			dev_dbg(&tns_dev->pdev->dev, "Input timestamping for terminal %s has been disabled.\n",
					terminal_name(sysfs_dev->term));
			err = -EPERM;
			goto err_term_ts_disabled;
		}

		/* overflow could have happened while we were waiting */
		err = _nisync_test_and_clear_term_overflow(tns_dev, sysfs_dev);
		if (err)
			goto err_overflow_after_wait;

		if (nisync_buf_empty(sysfs_dev->timestamps)) {
			dev_dbg(&tns_dev->pdev->dev, "Timestamps were emptied before re-acquiring the lock for terminal %s\n",
					terminal_name(sysfs_dev->term));
			err = -ERESTART;
			goto err_restart;
		}
	}

	err = nisync_buf_read_block_to_user(sysfs_dev->timestamps,
				struct nisync_timestamp_nanos, buf, count);
	if (err) {
		dev_dbg(&tns_dev->pdev->dev, "Error %d reading timestamps from terminal %s to user buffer\n",
			    err, terminal_name(sysfs_dev->term));
		goto err_read_block;
	}

	mutex_unlock(&sysfs_dev->term_lock);
	up_read(tns_dev->dev_reset_lock);

	dev_dbg(&tns_dev->pdev->dev, "(%s) => ts read: %zu\n",
		terminal_name(sysfs_dev->term),
				count / sizeof(struct nisync_timestamp_nanos));

	*f_pos += count;
	return count;

err_read_block:
err_restart:
err_overflow_after_wait:
err_term_ts_disabled:
err_term_ts_aborted:
err_wait_event:
err_try_again:
err_overflow:
err_ts_not_enabled:
	mutex_unlock(&sysfs_dev->term_lock);
	up_read(tns_dev->dev_reset_lock);

	return err;
}

static unsigned int nisync_terminal_poll(struct file *filp, poll_table *wait)
{
	struct nisync_sysfs_dev *sysfs_dev = filp->private_data;
	struct nisync_tns_dev *tns_dev =
		container_of_array(sysfs_dev, struct nisync_tns_dev,
				sysfs_devs, sysfs_dev->term);
	unsigned int mask;

	dev_dbg(&tns_dev->pdev->dev, "%s(%s)\n", __func__,
		terminal_name(sysfs_dev->term));

	down_read(tns_dev->dev_reset_lock);
	mutex_lock(&sysfs_dev->term_lock);

	if (!terminal_test_state(sysfs_dev, nisync_term_ts_enabled)) {
		mask = POLLHUP | POLLERR;
		goto err_ts_not_enabled;
	}

	poll_wait(filp, &sysfs_dev->wait_queue, wait);

	if (terminal_test_state(sysfs_dev, nisync_term_ts_aborted) ||
		!terminal_test_state(sysfs_dev, nisync_term_ts_enabled)) {
		mask = POLLHUP | POLLERR;
	} else {
		mask = 0;

		if (nisync_buf_size(sysfs_dev->timestamps))
			mask |= POLLIN;

		if (terminal_test_state(sysfs_dev, nisync_term_ts_sw_overflow)
				|| terminal_test_state(sysfs_dev,
						nisync_term_ts_hw_overflow))
			mask |= POLLERR;
	}

err_ts_not_enabled:
	mutex_unlock(&sysfs_dev->term_lock);
	up_read(tns_dev->dev_reset_lock);

	dev_dbg(&tns_dev->pdev->dev, "(%s) => mask: %u\n",
		terminal_name(sysfs_dev->term), mask);

	return mask;
}

static u64 _nisync_input_time_latency_correction(
				struct nisync_tns_dev *tns_dev, u64 fte_time)
{
	if (fte_time > tns_dev->future_time_latency)
		fte_time -= tns_dev->future_time_latency;

	fte_time = round_to_nearest(fte_time,
					tns_dev->nominal_clock_period_in_ns);

	return fte_time;
}

static int _nisync_create_fte(struct nisync_tns_dev *tns_dev,
				struct nisync_write_buf *buf,
				const u64 term_bitmask, const u64 curr_time,
				size_t *earliest_modified_fte_index)
{
	u64 fte_time	= buf->data.fte.time;
	u8 level	= buf->data.fte.level;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	if (NISYNC_TIME_INDEFINITE_NANOS == fte_time) {
		dev_dbg(&tns_dev->pdev->dev, "Can't schedule event for NISYNC_TIME_INDEFINITE_NANOS\n");
		return -EINVAL;
	}

	if (level != NISYNC_LEVEL_LOW && level != NISYNC_LEVEL_HIGH) {
		dev_dbg(&tns_dev->pdev->dev, "invalid level %u, defaulting to NISYNC_LEVEL_HIGH\n",
			level);
		level = NISYNC_LEVEL_HIGH;
	}

	if (fte_time != NISYNC_TIME_IMMEDIATE_NANOS)
		fte_time = _nisync_input_time_latency_correction(tns_dev,
								fte_time);

	return fte_fifo_insert(tns_dev->fte_fifo, fte_time, term_bitmask, level,
			curr_time, earliest_modified_fte_index);
}

static int __nisync_create_clock(struct nisync_tns_dev *tns_dev,
					struct nisync_write_buf *buf,
					struct nisync_sysfs_dev *sysfs_dev,
					const u64 curr_time,
					size_t *earliest_modified_fte_index)
{
	const u64 term_bitmask	= terminal_time_event_bit(sysfs_dev->term);
	u64 start_time		= buf->data.clock.start;
	u64 stop_time		= buf->data.clock.end;
	u32 duty_ticks		= buf->data.clock.high_ticks;
	u32 period_ticks	= duty_ticks + buf->data.clock.low_ticks;
	int err;

	dev_dbg(&tns_dev->pdev->dev, "%s(start_time %llu, stop_time %llu, period_ticks %u, duty_ticks %u)\n",
		__func__, start_time, stop_time, period_ticks, duty_ticks);

	if (start_time == NISYNC_TIME_INDEFINITE_NANOS) {
		dev_dbg(&tns_dev->pdev->dev, "Can't schedule clock with start time of NISYNC_TIME_INDEFINITE_NANOS\n");
		return -EINVAL;
	}

	if (start_time != NISYNC_TIME_IMMEDIATE_NANOS)
		start_time = _nisync_input_time_latency_correction(tns_dev,
								start_time);

	stop_time = _nisync_input_time_latency_correction(tns_dev, stop_time);

	if (start_time >= stop_time ||
	    ((start_time == NISYNC_TIME_IMMEDIATE_NANOS) &&
	     fte_fifo_immediate_fte_max_delay(tns_dev->fte_fifo, start_time) >= stop_time)) {
		dev_dbg(&tns_dev->pdev->dev, "Stop time is too close to or less than start time. (%llu >= %llu)\n",
				start_time, stop_time);
		return -EINVAL;
	}

	if (period_ticks < PXI668x_FTE_MIN_CLK_PERIOD_TICKS
		|| duty_ticks < PXI668x_FTE_MIN_DUTY_CYCLE_TICKS) {
		dev_dbg(&tns_dev->pdev->dev, "period_ticks %u < min period ticks %u or duty_ticks %u < min duty ticks %u\n",
			period_ticks, PXI668x_FTE_MIN_CLK_PERIOD_TICKS,
			duty_ticks, PXI668x_FTE_MIN_DUTY_CYCLE_TICKS);
		return -EINVAL;
	}

	if (period_ticks <= duty_ticks) {
		dev_dbg(&tns_dev->pdev->dev, "period_ticks must be more than duty_ticks.\n   period_ticks: %u\n   duty_ticks: %u\n\n",
			period_ticks, duty_ticks);
		return -EINVAL;
	}

	err = fte_fifo_insert(tns_dev->fte_fifo, start_time, term_bitmask,
		NISYNC_LEVEL_HIGH, curr_time, earliest_modified_fte_index);
	if (err)
		return err;
	dev_dbg(&tns_dev->pdev->dev, "Inserted clock start event at %llu, period %u ticks, duty_cycle %u ticks\n",
			start_time, period_ticks,
			duty_ticks);

	err = fte_fifo_insert(tns_dev->fte_fifo, stop_time, term_bitmask,
			      NISYNC_LEVEL_LOW, curr_time, NULL);
	if (err) {
		/*
		 * don't need safety checking in abort_fte; call
		 * fte_fifo_del directly
		 */
		fte_fifo_del(tns_dev->fte_fifo, start_time, term_bitmask,
				curr_time, NULL);
		return err;
	}
	/* extra space in debug print to line up with start print */
	dev_dbg(&tns_dev->pdev->dev, "Inserted clock stop  event at %llu\n",
		stop_time);

	writel(period_ticks,
		terminal_future_time_clock_period(tns_dev, sysfs_dev->term));
	writel(duty_ticks,
		terminal_future_time_clock_duty_cycle(tns_dev,
						sysfs_dev->term));

	/* Enable clock output (but will not actually start until start FTE) */
	terminal_set_state(sysfs_dev, nisync_term_fte_clk_enabled);
	writel(NISYNC_OUTPUT_CLOCK_ENABLE_BIT,
		terminal_control_enable(tns_dev, sysfs_dev->term));

	return 0;
}

static int _nisync_create_clock(struct nisync_tns_dev *tns_dev,
				struct nisync_write_buf *buf,
				struct nisync_sysfs_dev *sysfs_dev,
				const u64 curr_time,
				size_t *earliest_modified_fte_index)
{
	const u64 term_bitmask = terminal_time_event_bit(sysfs_dev->term);

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	if (fte_fifo_is_terminal_scheduled(tns_dev->fte_fifo, term_bitmask)) {
		dev_dbg(&tns_dev->pdev->dev, "Starting a clock could overwrite FTEs\n");
		return -EBUSY;
	}

	return __nisync_create_clock(tns_dev, buf, sysfs_dev, curr_time,
					earliest_modified_fte_index);
}

static int _nisync_create_pulse(struct nisync_tns_dev *tns_dev,
				struct nisync_write_buf *buf,
				const u64 term_bitmask, const u64 curr_time,
				size_t *earliest_modified_fte_index)
{
	const u32 nom_period	= tns_dev->nominal_clock_period_in_ns;
	u64 start_time		= buf->data.pulse.time;
	size_t _earliest_index  = fte_fifo_size(tns_dev->fte_fifo);
	u64 stop_time;
	int err;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	if (NISYNC_TIME_INDEFINITE_NANOS == start_time) {
		dev_dbg(&tns_dev->pdev->dev, "Can't schedule pulse for NISYNC_TIME_INDEFINITE_NANOS\n");
		return -EINVAL;
	}

	/*
	 * Immediate pulse is scheduled at a specific time to maintain the
	 * pulse width. This is subject to the FTE safety margin and limits
	 * the latency of how quickly you can schedule a pulse.
	 */
	if (start_time == NISYNC_TIME_IMMEDIATE_NANOS) {
		u64 min_fte_time = curr_time + FTE_FIFO_EVENT_PROGRAMMING_TIME +
					fte_fifo_current_safety_margin(
						tns_dev->fte_fifo);
		start_time = round_to_nearest(min_fte_time, nom_period);
		if (start_time < min_fte_time)
			start_time += nom_period;
	} else {
		start_time = _nisync_input_time_latency_correction(tns_dev,
								start_time);
	}
	stop_time = round_to_nearest(start_time + buf->data.pulse.width,
					nom_period);

	if (start_time >= stop_time) {
		dev_dbg(&tns_dev->pdev->dev, "Pulse start_time is after stop_time. (%llu >= %llu)\n",
			    start_time, stop_time);
		return -EINVAL;
	}

	err = fte_fifo_insert(tns_dev->fte_fifo, start_time, term_bitmask,
			buf->data.pulse.level, curr_time, &_earliest_index);
	if (err) {
		dev_dbg(&tns_dev->pdev->dev, "error %d scheduling pulse start event\n",
			err);
		return err;
	}

	err = fte_fifo_insert(tns_dev->fte_fifo, stop_time, term_bitmask,
			      !buf->data.pulse.level, curr_time, NULL);
	if (err) {
		dev_dbg(&tns_dev->pdev->dev, "error %d scheduling pulse stop event\n",
			err);
		goto err_stop_event;
	}

	dev_dbg(&tns_dev->pdev->dev, "pulse scheduled successfully\n");
	if (earliest_modified_fte_index &&
			_earliest_index < *earliest_modified_fte_index)
		*earliest_modified_fte_index = _earliest_index;
	return 0;

err_stop_event:
	/*
	 * remove start event; don't need safety checking in abort_fte, so call
	 * fte_fifo_del directly
	 */
	fte_fifo_del(tns_dev->fte_fifo, start_time, term_bitmask,
			curr_time, NULL);
	return err;
}

static int _nisync_abort_fte(struct nisync_tns_dev *tns_dev,
				struct nisync_write_buf *buf,
				const u64 term_bitmask, const u64 curr_time,
				size_t *earliest_modified_fte_index)
{
	u64 fte_time = buf->data.fte.time;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	if (NISYNC_TIME_IMMEDIATE_NANOS == fte_time ||
			NISYNC_TIME_INDEFINITE_NANOS == fte_time) {
		dev_dbg(&tns_dev->pdev->dev, "Invalid abort time %lluns\n",
			fte_time);
		return -EINVAL;
	}

	if (fte_time <= curr_time) {
		dev_dbg(&tns_dev->pdev->dev, "Cannot abort FTE in the past: %lluns\n",
			fte_time);
		return -ETIME;
	}

	fte_time = _nisync_input_time_latency_correction(tns_dev, fte_time);

	return fte_fifo_del(tns_dev->fte_fifo, fte_time, term_bitmask,
			curr_time, earliest_modified_fte_index);
}

static int _nisync_abort_clock(struct nisync_tns_dev *tns_dev,
				struct nisync_sysfs_dev *sysfs_dev,
				const u64 curr_time,
				size_t *earliest_modified_fte_index)
{
	int err;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	if (!terminal_test_state(sysfs_dev, nisync_term_fte_clk_enabled)) {
		dev_dbg(&tns_dev->pdev->dev, "Can't abort clock that doesn't exist (already finished?)\n");
		return -EINVAL;
	}

	err = fte_fifo_clear_terminal(tns_dev->fte_fifo,
			terminal_time_event_bit(sysfs_dev->term),
			curr_time, earliest_modified_fte_index);
	if (err) {
		dev_dbg(&tns_dev->pdev->dev, "error %d clearing terminal FTEs\n",
			err);
		return err;
	}

	/* Disconnect clock output */
	terminal_clear_state(sysfs_dev, nisync_term_fte_clk_enabled);
	writel(NISYNC_OUTPUT_CLOCK_ENABLE_BIT,
			terminal_control_disable(tns_dev, sysfs_dev->term));
	return 0;
}

static ssize_t nisync_terminal_write(struct file *filp,
				const char __user *user_buf,
				size_t count, loff_t *f_pos)
{
	int err;
	struct nisync_sysfs_dev *sysfs_dev = filp->private_data;
	struct nisync_tns_dev *tns_dev =
		container_of_array(sysfs_dev, struct nisync_tns_dev,
					sysfs_devs, sysfs_dev->term);
	const u64 term_bitmask = terminal_time_event_bit(sysfs_dev->term);
	struct nisync_write_buf write_buf;
	u64 curr_time;
	size_t hw_fifo_size;
	size_t earliest_modified_fte_index;
	int retry_count = 1;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	down_read(tns_dev->dev_reset_lock);
	mutex_lock(&sysfs_dev->term_lock);

	if (!terminal_test_state(sysfs_dev, nisync_term_fte_enabled)) {
		dev_dbg(&tns_dev->pdev->dev, "FTE not enabled on terminal (%s).\n",
					terminal_name(sysfs_dev->term));
		err = -EPERM;
		goto err_fte_not_enabled;
	}

	err = tns_copy_from_user((void *)&write_buf, (void *)user_buf,
				 sizeof(write_buf));
	if (err)
		goto err_write_buf_copy;

	tns_dev_dbg_fte(write_buf, terminal_name(sysfs_dev->term));

	mutex_lock(&tns_dev->fte_lock);
	get_cpu();

	do {

	earliest_modified_fte_index = fte_fifo_size(tns_dev->fte_fifo);

	tns_dev->hardware_clock->_get_time(tns_dev->hardware_clock,
					   &curr_time);

	/*
	 * First synchronize FIFO cache with onboard FIFO by removing FTEs that
	 * have already been executed.
	 */
	fte_fifo_del_older_than(tns_dev->fte_fifo, curr_time);

	if (terminal_test_state(sysfs_dev, nisync_term_fte_clk_enabled) &&
			write_buf.mode != NISYNC_ABORT_CLOCK) {
		if (fte_fifo_is_terminal_scheduled(tns_dev->fte_fifo,
							term_bitmask)) {
			dev_dbg(&tns_dev->pdev->dev, "Previously scheduled clock has not finished\n");
			err = -EBUSY;
			goto err_clk_already_scheduled;
		}

		/*
		 * No FTEs left in the fifo for this terminal,
		 * so clock generation must be finished.
		 */
		terminal_clear_state(sysfs_dev, nisync_term_fte_clk_enabled);
		writel(NISYNC_OUTPUT_CLOCK_ENABLE_BIT,
				terminal_control_disable(tns_dev,
				sysfs_dev->term));
	}

	hw_fifo_size = fte_fifo_size(tns_dev->fte_fifo);

	switch (write_buf.mode) {
	case NISYNC_WRITE_FTE:
		err = _nisync_create_fte(tns_dev, &write_buf, term_bitmask,
						curr_time,
						&earliest_modified_fte_index);
		break;
	case NISYNC_WRITE_CLOCK:
		err = _nisync_create_clock(tns_dev, &write_buf, sysfs_dev,
						curr_time,
						&earliest_modified_fte_index);
		break;
	case NISYNC_WRITE_PULSE:
		err = _nisync_create_pulse(tns_dev, &write_buf, term_bitmask,
						curr_time,
						&earliest_modified_fte_index);
		break;
	case NISYNC_ABORT_FTE:
		err = _nisync_abort_fte(tns_dev, &write_buf, term_bitmask,
						curr_time,
						&earliest_modified_fte_index);
		break;
	case NISYNC_ABORT_CLOCK:
		err = _nisync_abort_clock(tns_dev, sysfs_dev, curr_time,
						&earliest_modified_fte_index);
		break;
	default:
		dev_notice(&tns_dev->pdev->dev, "unknown write request: 0x%x\n",
			   write_buf.mode);
		err = -EINVAL;
		goto err_invalid_write_request;
	}

	} while ((err == -EAGAIN) && retry_count-- > 0);

	if (err)
		goto err_fte_request_failed;

	err = __nisync_write_fte_fifo_to_hw(tns_dev,
				earliest_modified_fte_index, hw_fifo_size);
	if (err)
		goto err_write_fte_fifo_to_hw;

	put_cpu();
	mutex_unlock(&tns_dev->fte_lock);
	mutex_unlock(&sysfs_dev->term_lock);
	up_read(tns_dev->dev_reset_lock);

	*f_pos += count;

	dev_dbg(&tns_dev->pdev->dev, "() => (count = %zu)\n", count);

	return count;

err_write_fte_fifo_to_hw:
err_fte_request_failed:
err_invalid_write_request:
err_clk_already_scheduled:
	put_cpu();
	mutex_unlock(&tns_dev->fte_lock);
err_write_buf_copy:
err_fte_not_enabled:
	mutex_unlock(&sysfs_dev->term_lock);
	up_read(tns_dev->dev_reset_lock);

	dev_dbg(&tns_dev->pdev->dev, "() => (err = %d)\n", err);
	return err;
}

static const struct file_operations nisync_terminal_fops = {
	.owner		= THIS_MODULE,
	.open		= nisync_terminal_open,
	.release	= nisync_terminal_release,
	.unlocked_ioctl	= nisync_terminal_ioctl,
	.read		= nisync_terminal_read,
	.poll		= nisync_terminal_poll,
	.write		= nisync_terminal_write,
};

static int nisync_eth_open(struct inode *inode, struct file *filp)
{
	int dev_index = MINOR(inode->i_cdev->dev) % NISYNC_DEVICE_MINOR_STEP;
	struct nisync_sysfs_dev *sysfs_dev =
		container_of(inode->i_cdev, struct nisync_sysfs_dev, char_dev);
	struct nisync_tns_dev *tns_dev =
		container_of_array(sysfs_dev, struct nisync_tns_dev,
				sysfs_devs, dev_index);

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	filp->private_data = sysfs_dev;

	return 0;
}

static ssize_t nisync_eth_read(struct file *filp, char __user *buf,
				size_t count, loff_t *f_pos)
{
	int err;
	struct nisync_sysfs_dev *sysfs_dev = filp->private_data;
	struct nisync_tns_dev *tns_dev =
		container_of_array(sysfs_dev, struct nisync_tns_dev,
				sysfs_devs, sysfs_dev->term);

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	down_read(tns_dev->dev_reset_lock);
	mutex_lock(&sysfs_dev->term_lock);

	if (nisync_buf_empty(sysfs_dev->pkts)) {
		if (filp->f_flags & O_NONBLOCK) {
			dev_dbg(&tns_dev->pdev->dev, "No eth pkts to read in nonblocking call\n");
			err = -EAGAIN;
			goto err_try_again;
		}

		/*
		 * While waiting, allow device resets and other terminal
		 * related operations to occur.
		 */
		mutex_unlock(&sysfs_dev->term_lock);
		up_read(tns_dev->dev_reset_lock);

		err = wait_event_interruptible(sysfs_dev->wait_queue,
					nisync_buf_size(sysfs_dev->pkts));

		down_read(tns_dev->dev_reset_lock);
		mutex_lock(&sysfs_dev->term_lock);
		if (err)
			goto err_wait_event;

		if (nisync_buf_empty(sysfs_dev->pkts)) {
			dev_dbg(&tns_dev->pdev->dev, "eth pkts were emptied before re-acquiring the terminal lock\n");
			err = -ERESTARTSYS;
			goto err_restartsys;
		}
	}

	err = nisync_buf_read_block_to_user(sysfs_dev->pkts,
					struct nisync_package_timestamp,
					buf, count);
	if (err)
		goto err_read_block;

	tns_dev_dbg_pkt("Read", (struct nisync_package_timestamp *)buf);

	mutex_unlock(&sysfs_dev->term_lock);
	up_read(tns_dev->dev_reset_lock);

	dev_dbg(&tns_dev->pdev->dev, "() => pkts read: %zu, pkts remaining: %d\n",
			count / sizeof(struct nisync_package_timestamp),
			nisync_buf_size(sysfs_dev->pkts));

	*f_pos += count;
	return count;

err_read_block:
err_restartsys:
err_wait_event:
err_try_again:
	dev_dbg(&tns_dev->pdev->dev, "() => err: %d\n", err);
	mutex_unlock(&sysfs_dev->term_lock);
	up_read(tns_dev->dev_reset_lock);
	return err;
}

static const struct file_operations nisync_eth_fops = {
	.owner	= THIS_MODULE,
	.open	= nisync_eth_open,
	.read	= nisync_eth_read
};

static int _get_trig_ts_pxi668x(struct nisync_tns_dev *tns_dev,
				u64 *timestamp, u64 *trig_lines,
				bool *initial_state)
{
	u32 high_word;
	u32 mid_high_word;
	u32 mid_low_word;
	u32 low_word;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	high_word = readl(tns_dev_product_regmap(
				input_timestamp_buffer_read));
	mid_high_word = readl(tns_dev_product_regmap(
				input_timestamp_buffer_read));
	mid_low_word = readl(tns_dev_product_regmap(
				input_timestamp_buffer_read));
	low_word = readl(tns_dev_product_regmap(
				input_timestamp_buffer_read));

	*timestamp = create_u64(high_word, mid_high_word);
	*trig_lines = create_u64(mid_low_word, low_word) & ~((u64)1);
	*initial_state = ((low_word & 1) != 0);

	return 0;
}

static int _get_trig_ts_status_pxi668x(struct nisync_tns_dev *tns_dev,
				u32 *ts_available, bool *ts_fifo_overflow)
{
	u32 ts_buffer_status;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	ts_buffer_status = readl(tns_dev_product_regmap(
				input_timestamp_buffer_status));

	/*
	 * Note that the PXI-668x boards report the number of 32-bit elements
	 * that are in the input timestamp buffers. A timestamp trigger for
	 * these boards are 4 32-bit words.
	 */
	*ts_available = (ts_buffer_status & 0xFFFF) >> 2;
	*ts_fifo_overflow = ((ts_buffer_status & BIT(31)) != 0);

	return 0;
}

static int _write_fte_fifo_pxi668x(struct nisync_tns_dev *tns_dev,
				const u64 fte_time,
				const u64 terminal_bitmask)
{
	dev_dbg(&tns_dev->pdev->dev, "%s(time: %10llus %9lluns, mask: 0x%08X %08X)\n",
					__func__,
					fte_time / NSEC_PER_SEC,
					fte_time % NSEC_PER_SEC,
					upper_32_bits(terminal_bitmask),
					lower_32_bits(terminal_bitmask));

	writel(upper_32_bits(fte_time),
		tns_dev_product_regmap(future_time_buffer_high_word));
	writel(lower_32_bits(fte_time),
		tns_dev_product_regmap(future_time_buffer_mid_word));
	writel(upper_32_bits(terminal_bitmask),
		tns_dev_product_regmap(future_time_buffer_mid_low_word));
	writel(lower_32_bits(terminal_bitmask),
		tns_dev_product_regmap(future_time_buffer_low_word));

	return 0;
}

static void nisync_ts_fifo_not_empty_handler(struct nisync_tns_dev *tns_dev)
{
	int err;
	u32 ts_available;
	u32 ts_processed;
	bool ts_fifo_overflow;
	struct nisync_sysfs_dev *sysfs_dev;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	_get_trig_ts_status_pxi668x(tns_dev, &ts_available, &ts_fifo_overflow);

	for (ts_processed = 0; ts_processed < ts_available; ts_processed++) {
		u64 trig_lines;
		bool initial_state;
		struct nisync_timestamp_nanos timestamp;

		_get_trig_ts_pxi668x(tns_dev, &timestamp.nanos, &trig_lines,
					&initial_state);

		tns_dev_dbg_ts("raw timestamp", timestamp.nanos);
		/* Adjust timestamp to account for HW latency */
		timestamp.nanos -= tns_dev->input_timestamp_latency;
		tns_dev_dbg_ts("adjusted timestamp", timestamp.nanos);

		/* Copy timestamp to the buffer of each associated terminal */
		spin_lock(&tns_dev->active_ts_terms_spinlock);
		list_for_each_entry(sysfs_dev, &tns_dev->active_ts_terms,
								list_node) {
			/*
			 * if this terminal's initial input timestamp state
			 * has not been set
			 */
			if (initial_state && !sysfs_dev->initial_state_set) {
				initial_state = false;
				sysfs_dev->initial_state_set = true;

				if (terminal_time_event_bit(sysfs_dev->term) &
								trig_lines)
					sysfs_dev->last_edge =
						NISYNC_EDGE_RISING;
				else
					sysfs_dev->last_edge =
						NISYNC_EDGE_FALLING;

			/* else if this terminal's state has changed */
			} else if (terminal_time_event_bit(sysfs_dev->term) &
								trig_lines) {
				if (sysfs_dev->edge_detection ==
							NISYNC_EDGE_ANY) {
					timestamp.edge = (NISYNC_EDGE_RISING ==
							sysfs_dev->last_edge) ?
							NISYNC_EDGE_FALLING :
							NISYNC_EDGE_RISING;
					sysfs_dev->last_edge = timestamp.edge;
				} else {
					timestamp.edge =
						sysfs_dev->edge_detection;
				}

				err = nisync_bb_write_element(
							sysfs_dev->timestamps,
							timestamp);
				if (err == -EOVERFLOW)
					terminal_set_state(sysfs_dev,
						nisync_term_ts_sw_overflow);

				sysfs_dev->wake_up = true;
			}
		}
		spin_unlock(&tns_dev->active_ts_terms_spinlock);
	}

	if (ts_fifo_overflow) {
		/* Disable HW timestamps */
		writel(NISYNC_INPUT_TIMESTAMP_BUFFER_INT_OVER_WATERMARK_BIT
			| NISYNC_GLOBAL_INPUT_TIMESTAMPS_BIT,
			tns_dev_product_regmap(
				input_timestamp_buffer_control_disable));

		/* Clear input timestamp buffer */
		writel(NISYNC_INPUT_TIMESTAMP_BUFFER_CLEAR_BIT,
			tns_dev_product_regmap(
				input_timestamp_buffer_control_enable));

		/* Enable timestamping and over watermark interrupts */
		writel(NISYNC_INPUT_TIMESTAMP_BUFFER_INT_OVER_WATERMARK_BIT
			| NISYNC_GLOBAL_INPUT_TIMESTAMPS_BIT,
			tns_dev_product_regmap(
				input_timestamp_buffer_control_enable));
	}

	/*
	 * For each active terminal, wakeup the consumer of timestamps if
	 * either an error has occurred or if timestamps are available.
	 */
	spin_lock(&tns_dev->active_ts_terms_spinlock);
	list_for_each_entry(sysfs_dev, &tns_dev->active_ts_terms, list_node) {
		if (sysfs_dev->wake_up || ts_fifo_overflow) {
			sysfs_dev->wake_up = false;
			wake_up_all(&sysfs_dev->wait_queue);
		}

		if (ts_fifo_overflow) {
			dev_dbg(&tns_dev->pdev->dev, "nisync_term_ts_hw_overflow set for terminal %s\n",
					terminal_name(sysfs_dev->term));
			terminal_set_state(sysfs_dev,
					nisync_term_ts_hw_overflow);
		}
	}
	spin_unlock(&tns_dev->active_ts_terms_spinlock);
}

static void _nisync_adjust_pkt_ts_pxi668x(struct nisync_tns_dev *tns_dev,
				u64 *pkt_ts)
{
	u64 curr_time;
	u64 curr_time_msb;
	u64 ts_time_ns;
	u64 adj_time_ns;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	/*
	 * Populate the upper 6-bits of the given pkt_ts time with the
	 * current time.
	 * ...Note that the upper 7th bit of the given time and the current
	 *    time are tested to detect a roll over condition.
	 */
	tns_dev->hardware_clock->_get_time(tns_dev->hardware_clock,
					   &curr_time);

	curr_time_msb = curr_time >> (64 - 6);
	ts_time_ns = *pkt_ts & ~(((u64)0x3F) << (64 - 6));

	if (ts_time_ns & (((u64)1) << (64 - 7)) &&
				!(curr_time & (((u64)1) << (64 - 7))))
		curr_time_msb--;

	adj_time_ns = ((curr_time_msb << (64 - 6)) | ts_time_ns);

	*pkt_ts = adj_time_ns;
}

static void nisync_pkt_fifo_not_empty_handler(struct nisync_tns_dev *tns_dev,
				enum nisync_terminal terminal)
{
	struct nisync_sysfs_dev *sysfs_dev = &tns_dev->sysfs_devs[terminal];
	u32 eth_link_speed;
	s32 pkt_latency = 0;
	u32 pkts_available;
	u32 pkts_processed;

	/*
	 * bitshift by the size of u32 instead of dividing by 4
	 * to support 32-bit mode
	 */
	u32 pkt_size = sizeof(struct nisync_package_timestamp) >> 2;

	dev_dbg(&tns_dev->pdev->dev, "%s(%s)\n", __func__,
		terminal == NISYNC_ETHIN ? "Rx":"Tx");

	pkts_available = readl(terminal_pkt_buffer_status(tns_dev,
						terminal)) & 0xFFFF;
	pkts_available /= pkt_size;
	dev_dbg(&tns_dev->pdev->dev, "pkts_available = %u\n", pkts_available);

	if (pkts_available) {
		eth_link_speed = (readl(tns_dev_product_regmap(
						board_status)) >> 5) & 0x3;
		if (terminal == NISYNC_ETHIN) {
			switch (eth_link_speed) {
			case eth_link_speed_10:
				pkt_latency = -tns_dev->eth_rx_latency_10mbps;
				break;
			case eth_link_speed_100:
				pkt_latency = -tns_dev->eth_rx_latency_100mbps;
				break;
			case eth_link_speed_1000:
				pkt_latency = -tns_dev->eth_rx_latency_1000mbps;
				break;
			}
		} else {
			switch (eth_link_speed) {
			case eth_link_speed_10:
				pkt_latency = tns_dev->eth_tx_latency_10mbps;
				break;
			case eth_link_speed_100:
				pkt_latency = tns_dev->eth_tx_latency_100mbps;
				break;
			case eth_link_speed_1000:
				pkt_latency = tns_dev->eth_tx_latency_1000mbps;
				break;
			}
		}
	}

	for (pkts_processed = 0; pkts_processed < pkts_available;
							pkts_processed++) {
		struct nisync_package_timestamp pkt_ts;
		u64 timestamp;

		/* Read pkt timestamp from HW */
		pkt_ts.parts[0] = readl(terminal_pkt_fifo_read(tns_dev,
								terminal));
		pkt_ts.parts[1] = readl(terminal_pkt_fifo_read(tns_dev,
								terminal));
		pkt_ts.parts[2] = readl(terminal_pkt_fifo_read(tns_dev,
								terminal));
		pkt_ts.parts[3] = readl(terminal_pkt_fifo_read(tns_dev,
								terminal));
		pkt_ts.parts[4] = readl(terminal_pkt_fifo_read(tns_dev,
								terminal));

		/* Adjust timestamp of packet for HW latency */
		timestamp = create_u64(pkt_ts.parts[0], pkt_ts.parts[1]);

		timestamp += pkt_latency;

		_nisync_adjust_pkt_ts_pxi668x(tns_dev, &timestamp);

		/* Write adjusted packet timestamp back to package structure */
		pkt_ts.parts[0] = upper_32_bits(timestamp);
		pkt_ts.parts[1] = lower_32_bits(timestamp);

		/* Write pkt timestamp to associated SW buffer */
		nisync_cb_write_element(sysfs_dev->pkts, pkt_ts);

		tns_dev_dbg_pkt(terminal == NISYNC_ETHIN ? "Rx":"Tx", &pkt_ts);
	}

	/* if packets are ready to be read by the driver client */
	if (pkts_processed)
		wake_up_interruptible(&sysfs_dev->wait_queue);
}

static void nisync_pkt_fifo_overflow_handler(struct nisync_tns_dev *tns_dev,
				enum nisync_terminal terminal)
{
	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	/* Disable snooping */
	writel(BIT(2), terminal_pkt_control_disable(tns_dev, terminal));

	/* Clear packet HW fifo */
	writel(BIT(4), terminal_pkt_control_enable(tns_dev, terminal));

	/* Re-enable snooping */
	writel(BIT(2), terminal_pkt_control_enable(tns_dev, terminal));
}

static void _nisync_get_offset_from_clk10(struct nisync_tns_dev *tns_dev,
				s32 *offset)
{
	u64 curr_time;
	u32 clk10_timestamp;
	u64 computed_clk10_timestamp;
	s32 clk10_offset;
	/* TODO: This latency should be provided by hardware */
	const u64 clk10_timestamp_latency_ns = 20;
	const u64 upper_32_bit_mask = 0xFFFFFFFF00000000;
	const u64 rollover_addition = 0x100000000;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	tns_dev->hardware_clock->_get_time(tns_dev->hardware_clock, &curr_time);

	clk10_timestamp = readl(tns_dev_product_regmap(clk10_timestamp_reg));

	/*
	 * The clk10 timestamp is only 32-bits so we need to use the upper
	 * 32 bits of the current time
	 */
	computed_clk10_timestamp = (curr_time & upper_32_bit_mask) |
							clk10_timestamp;

	/*
	 * Check for the rollover condition where the clk10 32-bit timestamp
	 * is greater than the current time's lower 32-bits.
	 */
	if ((curr_time & 0xFFFFFFFF) > clk10_timestamp)
		computed_clk10_timestamp += rollover_addition;

	/* Subtract the signal latencies incurred through the FPGA and board */
	computed_clk10_timestamp -= clk10_timestamp_latency_ns;
	clk10_offset = CLK10_PERIOD_IN_NS -
			(computed_clk10_timestamp % CLK10_PERIOD_IN_NS);

	if ((CLK10_PERIOD_IN_NS >> 1) < clk10_offset)
		clk10_offset = clk10_offset - CLK10_PERIOD_IN_NS;

	dev_dbg(&tns_dev->pdev->dev, "() => clk10_offset: %dns\n",
		clk10_offset);

	*offset = clk10_offset;
}

static void nisync_pll_status_changed(struct nisync_tns_dev *tns_dev)
{
	s32 clk10_offset = 0;
	int err;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	if (!_nisync_is_pll_locked(tns_dev))
		return;

	_nisync_get_offset_from_clk10(tns_dev, &clk10_offset);

	err = tns_dev->hardware_clock->_adjust_time(tns_dev->hardware_clock,
						(s64)clk10_offset);
	if (err) {
		tns_dev->status = NISYNC_STATUS_HW_ERROR;
		dev_dbg(&tns_dev->pdev->dev, "Error %d occurred adjusting time\n",
			err);
		return;
	}
}

static void nisync_gps_read_rx(struct nisync_tns_dev *tns_dev)
{
	u32 data;
	int bytes;
	struct uart_port *port = &tns_dev->nisync_gps_uart_port;

	do {
		data = readl(tns_dev_product_regmap(gps_rx_buffer));

		bytes = (data & (BIT(24) | BIT(25))) >> 24;

		/* FIXME: TTY_NORMAL */
		if (bytes > 2) {
			uart_insert_char(port, data, BIT(30),
					(data & 0xFF0000)>>16, TTY_NORMAL);
		}
		if (bytes > 1) {
			uart_insert_char(port, data, BIT(30),
					(data & 0xFF00)>>8, TTY_NORMAL);
		}
		if (bytes > 0) {
			uart_insert_char(port, data, BIT(30), data & 0xFF,
								TTY_NORMAL);
		}
		if (bytes == 0)
			return;

	} while (data & BIT(31));
	tty_flip_buffer_push(&port->state->port);
}

static void nisync_gps_rx_pending_handler(struct nisync_tns_dev *tns_dev)
{
	nisync_gps_read_rx(tns_dev);
}

static void nisync_gps_pps_int_handler(struct nisync_tns_dev *tns_dev)
{
	u32 ts_high, ts_low;
	u32 bt_high, bt_low;

	ts_high = readl(tns_dev_product_regmap(gps_pps_ts_high_word));
	ts_low = readl(tns_dev_product_regmap(gps_pps_ts_low_word));

	bt_low = readl(tns_dev_product_regmap(system_time_low_word));
	bt_high = readl(tns_dev_product_regmap(system_time_high_word));

	/* replace top 6 bits (which represent TDC) with board time */
	ts_high = (ts_high & ~0xFC000000) | (bt_high & 0xFC000000);

	write_seqlock(&tns_dev->gps_seqlock);

	tns_dev->gps_pps_ts = create_u64(ts_high, ts_low);
	tns_dev->gps_pps_ts -= tns_dev->gps_pps_latency;

	write_sequnlock(&tns_dev->gps_seqlock);
}

static void nisync_gps_tx_overwatermark_handler(struct nisync_tns_dev *tns_dev)
{
	struct uart_port *port = &tns_dev->nisync_gps_uart_port;
	struct circ_buf *xmit = &port->state->xmit;
	int count = 128;


	if (port->x_char) {
		writel(port->x_char, tns_dev_product_regmap(gps_tx_buffer));
		port->icount.tx++;
		port->x_char = 0;
		return;
	}

	if (uart_circ_empty(xmit) || uart_tx_stopped(port)) {
		nisync_gps_uart_stop_tx(port);
		return;
	}

	do {
		writel(xmit->buf[xmit->tail],
					tns_dev_product_regmap(gps_tx_buffer));
		xmit->tail = (xmit->tail + 1) & (UART_XMIT_SIZE - 1);
		port->icount.tx++;
		if (uart_circ_empty(xmit))
			break;
	/* TODO: can this be done more efficiently? */
	} while (--count > 0 && nisync_gps_tx_empty_count(tns_dev) > 0);

	if (uart_circ_chars_pending(xmit) < WAKEUP_CHARS)
		uart_write_wakeup(port);

	if (uart_circ_empty(xmit))
		nisync_gps_uart_stop_tx(port);
}

static void nisync_gps_rx_overflow_handler(struct nisync_tns_dev *tns_dev)
{
	nisync_gps_read_rx(tns_dev);

	writel(GPS_RX_CLEAR_BIT, tns_dev_product_regmap(gps_control_enable));
}

static void nisync_irq_tasklet_handler(unsigned long data)
{
	struct nisync_tns_dev *tns_dev = (struct nisync_tns_dev *) data;
	unsigned long flags;
	u32 int_status;
	bool int_processed;
	unsigned int i;
	unsigned int loopLimit = 5; /* loopLimit arbitrarily chosen as 5 */

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	for (i = 0, int_processed = true; int_processed && i < loopLimit; ++i) {
		int_processed = false;
		int_status = readl(tns_dev_product_regmap(
						int_status_read_to_clear));

		/* if input timestamp buffer is not empty */
		if (int_status & BIT(0)) {
			nisync_ts_fifo_not_empty_handler (tns_dev);
			int_processed = true;
		}

		/* if rx eth fifo is not empty */
		if (int_status & BIT(7)) {
			nisync_pkt_fifo_not_empty_handler(tns_dev,
							NISYNC_ETHIN);
			int_processed = true;
		}

		/* if tx eth fifo is not empty */
		if (int_status & BIT(6)) {
			nisync_pkt_fifo_not_empty_handler(tns_dev,
							NISYNC_ETHOUT);
			int_processed = true;
		}

		/* if rx eth fifo overflowed */
		if (int_status & BIT(5)) {
			nisync_pkt_fifo_overflow_handler(tns_dev,
							NISYNC_ETHIN);
			int_processed = true;
		}

		/* if tx eth fifo overflowed */
		if (int_status & BIT(4)) {
			nisync_pkt_fifo_overflow_handler(tns_dev,
							NISYNC_ETHOUT);
			int_processed = true;
		}

		/* if PLL lost lock or locked */
		if (int_status & BIT(17)) {
			nisync_pll_status_changed(tns_dev);
			int_processed = true;
		}

		/* if GPS RX buffer has data */
		if (int_status & BIT(12)) {
			nisync_gps_rx_pending_handler(tns_dev);
			int_processed = true;
		}

		/* if GPS PPS Int pending */
		if (int_status & BIT(11)) {
			nisync_gps_pps_int_handler(tns_dev);
			int_processed = true;
		}

		/* if GPS TX is over watermark */
		if (int_status & BIT(10)) {
			nisync_gps_tx_overwatermark_handler(tns_dev);
			int_processed = true;
		}

		/* if GPS RX overflows or overwatermark */
		if (int_status & BIT(9)) {
			nisync_gps_rx_overflow_handler(tns_dev);
			int_processed = true;
		}

		/* if IRIG-B message is decoded and timestamped */
		if (int_status & BIT(18)) {
			nisync_irig_timestamp_handler(tns_dev);
			int_processed = true;
		}
	}

	if (int_processed && i == loopLimit)
		dev_dbg(&tns_dev->pdev->dev, "processed %u sequential interrupts; exiting tasklet\n",
			i);

	/* if HW and SW are ready for interrupts, re-enable HW interrupts. */
	spin_lock_irqsave(&tns_dev->irq_state_spinlock, flags);
	if (irq_test_state(tns_dev, irq_state_hw_rdy) &&
			irq_test_state(tns_dev, irq_state_sw_rdy)) {

		writel(1, tns_dev_product_regmap(
					global_board_settings_enable));
		irq_set_state(tns_dev, irq_state_enabled);
	}
	spin_unlock_irqrestore(&tns_dev->irq_state_spinlock, flags);
}

static irqreturn_t nisync_tns_irq_handler(int irq, void *dev)
{
	irqreturn_t ret = IRQ_NONE;
	struct nisync_tns_dev *tns_dev = (struct nisync_tns_dev *)dev;

	dev_dbg(&tns_dev->pdev->dev, "%s(irq: %d)\n", __func__, irq);

	/* if an interrupt condition exists and HW interrupts are enabled */
	spin_lock(&tns_dev->irq_state_spinlock);
	if (irq_test_state(tns_dev, irq_state_enabled) &&
		readl(tns_dev_product_regmap(int_status))) {

		/* Disable interrupts and update IRQ state */
		writel(1, tns_dev_product_regmap(
					global_board_settings_disable));
		irq_clear_state(tns_dev, irq_state_enabled);

		tasklet_schedule(&tns_dev->irq_tasklet);

		ret = IRQ_HANDLED;
	}
	spin_unlock(&tns_dev->irq_state_spinlock);

	return ret;
}

static struct nisync_driver_data {
	int device_major;
	struct class *nisync_class;
} driver_data;

static int nisync_delete_sysfs_devices(struct nisync_tns_dev *tns_dev)
{
	int minor = tns_dev->driver_instance_id * NISYNC_DEVICE_MINOR_STEP;
	int major = driver_data.device_major;
	int index;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	for (index = NISYNC_DEVICE_MINOR_STEP - 1; index >= 0; index--) {
		if (tns_dev->sysfs_devs[index].dev) {
			device_destroy(driver_data.nisync_class,
					MKDEV(major, minor + index));
			cdev_del(&tns_dev->sysfs_devs[index].char_dev);
			tns_dev->sysfs_devs[index].dev = NULL;
		}
	}

	return 0;
}

static int nisync_create_sysfs_device(struct nisync_tns_dev *tns_dev,
				const struct file_operations *fops,
				enum nisync_terminal terminal,
				struct device *parent,
				const char* sysfs_device_name)
{
	int err;
	int minor = tns_dev->driver_instance_id * NISYNC_DEVICE_MINOR_STEP +
		terminal;
	int major = driver_data.device_major;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	BUG_ON(terminal >= NISYNC_DEVICE_MINOR_STEP);

	cdev_init(&tns_dev->sysfs_devs[terminal].char_dev, fops);
	tns_dev->sysfs_devs[terminal].char_dev.owner = THIS_MODULE;
	err = cdev_add(&tns_dev->sysfs_devs[terminal].char_dev,
			MKDEV(major, minor), 1);
	if (err < 0) {
		dev_err(&tns_dev->pdev->dev, "Error %d adding nisync device to system.\n",
			err);
		goto err_cdev_add;
	}

	tns_dev->sysfs_devs[terminal].dev =
		device_create(driver_data.nisync_class,
				parent,
				MKDEV(major, minor), tns_dev,
				sysfs_device_name);

	if (IS_ERR(tns_dev->sysfs_devs[terminal].dev)) {
		err = PTR_ERR(tns_dev->sysfs_devs[terminal].dev);
		dev_err(&tns_dev->pdev->dev, "Error %d adding nisync device to system.\n",
			err);
		tns_dev->sysfs_devs[terminal].dev = NULL;
		goto err_device_create;
	}

	return 0;

 err_device_create:
	cdev_del(&tns_dev->sysfs_devs[terminal].char_dev);
 err_cdev_add:
	return err;
}

static int nisync_create_sysfs_devices(struct nisync_tns_dev *tns_dev)
{
	int err;
	int i;
	u32 terminal;
	char sysfs_device_name[100];

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	snprintf(sysfs_device_name, sizeof(sysfs_device_name), "%s.%u",
			tns_dev->pdata->product_name,
			tns_dev->pdata->instance_id);
	err = nisync_create_sysfs_device(tns_dev, &nisync_fops, NISYNC_DEVICE,
					&tns_dev->pdev->dev, sysfs_device_name);
	if (err)
		return err; /* error printed by function already */

	i = 0;
	while ((terminal = tns_dev_product_terminal(i)) !=
						NISYNC_LAST_TERMINAL) {
		if (terminal != NISYNC_GND && terminal != NISYNC_CLK10) {
			snprintf(sysfs_device_name, sizeof(sysfs_device_name),
					"%s.%u.%s",
					tns_dev->pdata->product_name,
					tns_dev->pdata->instance_id,
					terminal_name(terminal));
			err = nisync_create_sysfs_device(tns_dev,
					terminal_fops(terminal), terminal,
					tns_dev->sysfs_devs[NISYNC_DEVICE].dev,
					sysfs_device_name);
		}

		if (err)
			goto err_create_terminal_sysfs;

		i++;
	}

	return 0;

err_create_terminal_sysfs:
	nisync_delete_sysfs_devices(tns_dev);
	return err;
}

static int _nisync_validate_hw_sw_compatibility(struct nisync_tns_dev *tns_dev)
{
	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	/*
	 * Print device information - used for solving issues with FPGA
	 * and firmware
	 */
	dev_info(&tns_dev->pdev->dev, "Found device %s (0x%x): Signature=0x%08x Revision=0x%08x OldestCompatibleRevision=0x%08x HardwareRevision=0x%08x\n",
		tns_dev->pdata->product_name,
		tns_dev_product_id(),
		tns_dev->signature,
		tns_dev->revision,
		tns_dev->oldest_compatible_revision,
		tns_dev->hardware_revision);

	dev_info(&tns_dev->pdev->dev, "Driver compatibility for device %s (0x%x): NewestCompatibleFPGAVersion=0x%08x OldestCompatibleFPGAVersion=0x%08x HardwareRevision=0x%08x\n",
		tns_dev->pdata->product_name,
		tns_dev_product_id(),
		tns_dev_product_newest_compatible_fpga(),
		tns_dev_product_oldest_compatible_fpga(),
		tns_dev_product_compatible_hardware_revision());

	if (tns_dev->oldest_compatible_revision > tns_dev->revision) {
		dev_err(&tns_dev->pdev->dev, "FPGA specifies that the oldest compatible version (0x%08x) is newer than the current one (0x%08x).\n",
			tns_dev->oldest_compatible_revision,
			tns_dev->revision);
		return -EINVAL;
	}

	if (tns_dev_product_oldest_compatible_fpga() >
				tns_dev_product_newest_compatible_fpga()) {
		dev_err(&tns_dev->pdev->dev, "Driver specifies that the oldest compatible FPGA version (0x%08x) is newer than the current one (0x%08x).\n",
			tns_dev_product_oldest_compatible_fpga(),
			tns_dev_product_newest_compatible_fpga());
		return -EINVAL;
	}

	if (tns_dev_product_oldest_compatible_fpga() > tns_dev->revision) {
		dev_err(&tns_dev->pdev->dev, "FPGA incompatible (too old): HW revision 0x%08x, SW oldest_compatible_fpga 0x%08x\n",
			tns_dev->revision,
			tns_dev_product_oldest_compatible_fpga());
		return -EINVAL;
	}

	if (tns_dev->oldest_compatible_revision >
				tns_dev_product_newest_compatible_fpga()) {
		dev_err(&tns_dev->pdev->dev, "FPGA incompatible (too new): HW oldest_compatible_revision 0x%08x, SW newest_compatible_fpga 0x%08x\n",
			tns_dev->oldest_compatible_revision,
			tns_dev_product_newest_compatible_fpga());
		return -EINVAL;
	}

	return 0;
}

static int _nisync_cache_hw_attributes(struct nisync_tns_dev *tns_dev)
{
	u32 value;
	u32 nco_width;

	dev_dbg(&tns_dev->pdev->dev, "%s()\n", __func__);

	/* Cache hardware compatibility information */
	tns_dev->signature = readl(tns_dev_product_regmap(signature));
	tns_dev->revision = readl(tns_dev_product_regmap(revision_reg));
	tns_dev->oldest_compatible_revision =
		readl(tns_dev_product_regmap(oldest_compatible_revision));
	tns_dev->hardware_revision =
		readl(tns_dev_product_regmap(hardware_revision));

	/* Cache hardware latencies */
	tns_dev->system_time_latency =
		readl(tns_dev_product_regmap(system_time_latency));

	value = readl(tns_dev_product_regmap(enet_transmit_latency));
	tns_dev->eth_tx_latency_10mbps = (value >> 16) & 0xFFFF;
	tns_dev->eth_tx_latency_100mbps = value & 0xFFFF;
	value = readl(tns_dev_product_regmap(enet_receive_latency));
	tns_dev->eth_rx_latency_10mbps = (value >> 16) & 0xFFFF;
	tns_dev->eth_rx_latency_100mbps = value & 0xFFFF;

	tns_dev->input_timestamp_latency =
		readl(tns_dev_product_regmap(input_timestamp_latency));
	tns_dev->partial_clear_operation_latency =
		readl(tns_dev_product_regmap(
					partial_clear_operation_latency));
	tns_dev->future_time_latency =
		readl(tns_dev_product_regmap(future_time_latency));
	tns_dev->gps_pps_latency =
		readl(tns_dev_product_regmap(gps_pps_latency));

	value = readl(tns_dev_product_regmap(gps_serial_buffer_size));
	tns_dev->gps_tx_buffer_size = value >> 16;
	tns_dev->gps_rx_buffer_size = value & 0xFFFF;

	tns_dev->irig_dc_latency =
		readl(tns_dev_product_regmap(irig_decodedc_latency));
	tns_dev->irig_am_latency =
		readl(tns_dev_product_regmap(irig_decodeam_latency));

	value = readl(tns_dev_product_regmap(
					genet_transmit_latency));
	tns_dev->eth_tx_latency_1000mbps = (value >> 16) & 0xFFFF;
	value = readl(tns_dev_product_regmap(
					genet_receive_latency));
	tns_dev->eth_rx_latency_1000mbps = (value >> 16) & 0xFFFF;

	/* Cache hardware clock attributes */
	tns_dev->oscillator_frequency =
		readl(tns_dev_product_regmap(oscillator_freq));
	tns_dev->nominal_frequency_comp =
		readl(tns_dev_product_regmap(freq_comp_nominal));
	nco_width = readl(tns_dev_product_regmap(nco_width));
	tns_dev->nco_precision = (u64)(1) <<
		readl(tns_dev_product_regmap(nco_width));

	/* Calculate nominal clock period */
	{
		u64 nominal_frequency =
			(u64) tns_dev->nominal_frequency_comp *
			(u64) tns_dev->oscillator_frequency *
			(u64) 2;
		nominal_frequency /= (u64) tns_dev->nco_precision;
		nominal_frequency += 1;
		nominal_frequency /= 2;

		tns_dev->nominal_frequency = (u32) nominal_frequency;

		tns_dev->nominal_clock_period_in_ns =
			NSEC_PER_SEC / tns_dev->nominal_frequency;

		dev_dbg(&tns_dev->pdev->dev, "tns_dev->nominal_frequency: %u\n",
					tns_dev->nominal_frequency);
		dev_dbg(&tns_dev->pdev->dev, "tns_dev->nominal_clock_period_in_ns: %u\n",
					tns_dev->nominal_clock_period_in_ns);
	}

	if ((0 == tns_dev->nominal_frequency_comp)
				|| (0 == tns_dev->oscillator_frequency)) {
		dev_err(&tns_dev->pdev->dev, "Unable to communicate to hardware.\n");
		return -EIO;
	}

	return 0;
}

static int nisync_tns_probe(struct platform_device *pdev)
{
	int err;
	int i;
	int irq;
	u32 terminal;
	u32 fte_buffer_size;
	struct nisync_tns_dev *tns_dev;

	dev_dbg(&pdev->dev, "%s()\n", __func__);

	tns_dev = kzalloc(sizeof(struct nisync_tns_dev), GFP_KERNEL);
	if (!tns_dev) {
		err = -ENOMEM;
		goto err_alloc;
	}
	tns_dev->status = NISYNC_STATUS_INITIALIZING;
	platform_set_drvdata(pdev, tns_dev);

	tns_dev->pdev = pdev;
	tns_dev->pdata = dev_get_platdata(&pdev->dev);

	tns_dev->product_support_idx = -1;
	for (i = 0; i < sizeof(product_support_table) /
			sizeof(struct nisync_product_support);
			i++) {
		if (tns_dev->pdata->product_id ==
				product_support_table[i].product_id) {
			tns_dev->product_support_idx = i;
			break;
		}
	}

	if (-1 == tns_dev->product_support_idx) {
		dev_err(&tns_dev->pdev->dev, "Unsupported product ID: 0x%x.\n",
					tns_dev->pdata->product_id);
		err = -EINVAL;
		goto err_invalid_product_id;
	}

	tns_dev->driver_instance_id = pdev->id;
	seqlock_init(&tns_dev->irig_seqlock);

	tns_dev->hardware_clock = tns_dev->pdata->hardware_clock;

	tns_dev->dev_lock = &tns_dev->hardware_clock->dev_lock;
	tns_dev->dev_reset_lock = &tns_dev->hardware_clock->dev_reset_lock;
	seqlock_init(&tns_dev->gps_seqlock);

	tns_dev->regs = (struct tns_regs __iomem *)tns_dev->pdata->mappedio;

	err = _nisync_cache_hw_attributes(tns_dev);
	if (err)
		/* appropriate error already printed by function */
		goto err_hw_fault;

	err = _nisync_validate_hw_sw_compatibility(tns_dev);
	if (err)
		/* appropriate error already printed by function */
		goto err_hw_sw_incompatible;

	spin_lock_init(&tns_dev->irq_state_spinlock);
	tasklet_init(&tns_dev->irq_tasklet, &nisync_irq_tasklet_handler,
			(unsigned long)tns_dev);

	irq = platform_get_irq(pdev, 0);
	if (irq < 0) {
		dev_err(&tns_dev->pdev->dev, "Error %d getting device IRQ.\n",
			irq);
		err = -EINVAL;
		goto err_get_irq;
	}

	/* TODO: Use devm_request_threaded_irq */
	err = request_irq(irq, nisync_tns_irq_handler, IRQF_SHARED,
				pdev->name, tns_dev);
	if (err) {
		dev_err(&tns_dev->pdev->dev, "Error %d installing IRQ handler!\n",
			err);
		goto err_req_irq;
	}

	tns_dev->ptp_state = NISYNC_PTPD_DISABLED;

	mutex_init(&tns_dev->fte_lock);

	/* Allocate FTE SW FIFO that is the same size as the FTE HW FIFO */
	fte_buffer_size =
		readl(tns_dev_product_regmap(future_time_buffer_size));
	tns_dev->fte_fifo = fte_fifo_alloc(fte_buffer_size);
	if (NULL == tns_dev->fte_fifo) {
		dev_err(&tns_dev->pdev->dev, "Failed allocating Future Time Event buffer of size: %u.\n",
			fte_buffer_size);
		goto err_fte_fifo_alloc;
	}

	/*
	 * Configure FTE FIFO to reject FTEs that violate the FTE rearm time of
	 * this hardware.
	 */
	fte_fifo_set_min_time_delta(tns_dev->fte_fifo,
				PXI668x_FTE_REARM_TIME_TICKS *
				tns_dev->nominal_clock_period_in_ns);

	/* Initialize each terminal */
	i = 0;
	while ((terminal = tns_dev_product_terminal(i)) !=
						NISYNC_LAST_TERMINAL) {

		struct nisync_sysfs_dev *sysfs_dev =
			&tns_dev->sysfs_devs[terminal];

		sysfs_dev->term = terminal;
		sysfs_dev->open_count = 0;
		mutex_init(&sysfs_dev->term_lock);
		init_waitqueue_head(&sysfs_dev->wait_queue);
		INIT_LIST_HEAD(&sysfs_dev->list_node);

		i++;
	}

	spin_lock_init(&tns_dev->active_ts_terms_spinlock);
	INIT_LIST_HEAD(&tns_dev->active_ts_terms);

	irq_set_state(tns_dev, irq_state_sw_rdy);

	err = nisync_reset(tns_dev);
	if (err) {
		dev_err(&tns_dev->pdev->dev, "Error %d resetting hardware\n",
			err);
		goto err_reset_hardware;
	}

	err = nisync_create_sysfs_devices(tns_dev);
	if (err) {
		dev_err(&tns_dev->pdev->dev, "Error %d creating subdevices\n",
			err);
		goto err_create_devices;
	}

	err = nisync_gps_register_uart(tns_dev);
	if (err) {
		dev_err(&tns_dev->pdev->dev,
				"Error %d registering gps uart driver\n", err);
		goto err_register_gps_uart;
	}

	err = nisync_gps_register_port(tns_dev);
	if (err) {
		dev_err(&tns_dev->pdev->dev,
				"Error %d adding gps uart port\n", err);
		goto err_gps_register_port;
	}

	tns_dev->status = NISYNC_STATUS_OK;
	return 0;

 err_gps_register_port:
	nisync_gps_unregister_uart(tns_dev);
 err_register_gps_uart:
	nisync_delete_sysfs_devices(tns_dev);
 err_create_devices:
 err_reset_hardware:
	fte_fifo_destroy(tns_dev->fte_fifo);
 err_fte_fifo_alloc:
	free_irq(irq, tns_dev);
 err_req_irq:
 err_get_irq:
 err_hw_sw_incompatible:
 err_hw_fault:
 err_invalid_product_id:
	platform_set_drvdata(pdev, NULL);
	kfree(tns_dev);
 err_alloc:
	return err;
}

static int nisync_tns_remove(struct platform_device *pdev)
{
	struct nisync_tns_dev *tns_dev = NULL;
	unsigned long irq_flags;
	int irq = -1;

	/*
	 * TODO (rdg): Can remove be called with open file descriptors
	 * (features enabled)?
	 */
	dev_dbg(&pdev->dev, "%s()\n", __func__);

	tns_dev = platform_get_drvdata(pdev);

	/* Ensure HW interrupts are disabled */
	spin_lock_irqsave(&tns_dev->irq_state_spinlock, irq_flags);
	writel(1, tns_dev_product_regmap(global_board_settings_disable));
	irq_clear_state(tns_dev, irq_state_enabled);
	irq_clear_state(tns_dev, irq_state_hw_rdy);
	spin_unlock_irqrestore(&tns_dev->irq_state_spinlock, irq_flags);

	tasklet_kill(&tns_dev->irq_tasklet);

	tns_dev->status = NISYNC_STATUS_UNAVAILABLE;
	platform_set_drvdata(pdev, NULL);

	irq = platform_get_irq(pdev, 0);
	free_irq(irq, tns_dev);

	/* Put HW into a known state */
	_nisync_reset_pxi668x(tns_dev);

	nisync_delete_sysfs_devices(tns_dev);

	fte_fifo_destroy(tns_dev->fte_fifo);

	nisync_gps_unregister_port(tns_dev);
	nisync_gps_unregister_uart(tns_dev);

	kfree(tns_dev);

	return 0;
}

static struct platform_driver nisync_tns_driver = {
	.driver = {
		.name = KBUILD_MODNAME,
	},
	.probe	= nisync_tns_probe,
	.remove	= __devexit_p(nisync_tns_remove),
};

MODULE_ALIAS("platform:" KBUILD_MODNAME);

static int __init nisync_tns_init(void)
{
	int err = 0;
	dev_t devno = MKDEV(0, 0);

	pr_debug("%s()\n", __func__);

	err = alloc_chrdev_region(&devno, 0, NISYNC_MINOR_NUMBERS,
				  KBUILD_MODNAME);
	if (err < 0) {
		pr_err("Error %d allocating device region of driver %s\n",
			err, KBUILD_MODNAME);
		goto err_alloc_chrdev_region;
	}

	driver_data.device_major = MAJOR(devno);

	driver_data.nisync_class = class_create(THIS_MODULE, KBUILD_MODNAME);

	if (IS_ERR(driver_data.nisync_class)) {
		err = PTR_ERR(driver_data.nisync_class);
		pr_err("Error %d allocating nisync class\n", err);
		goto err_create_class;
	}

	err = platform_driver_register(&nisync_tns_driver);
	if (err) {
		pr_err("Error %d registering platform driver\n", err);
		goto err_platform_driver_register;
	}

	return 0;

 err_platform_driver_register:
	class_unregister(driver_data.nisync_class);
	driver_data.nisync_class = NULL;
 err_create_class:
	unregister_chrdev_region(devno, NISYNC_MINOR_NUMBERS);
	driver_data.device_major = 0;
 err_alloc_chrdev_region:
	return err;
}

static void __exit nisync_tns_exit(void)
{
	pr_debug("%s()\n", __func__);

	platform_driver_unregister(&nisync_tns_driver);

	class_unregister(driver_data.nisync_class);
	driver_data.nisync_class = NULL;

	unregister_chrdev_region(MKDEV(driver_data.device_major, 0),
					NISYNC_MINOR_NUMBERS);
	driver_data.device_major = 0;
}

module_init(nisync_tns_init);
module_exit(nisync_tns_exit);

MODULE_LICENSE("GPL v2");

